(*
 Sungsoon Jang, The LET Project, KAIST

 Copyright(c) 2001 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*)

(********************************************************************)
(*																	*)
(*	 	Change the ast tree appropriate to the comparison			*)
(*			 2001. 6. 15. Jang Seong Soon							*)
(*																	*)
(********************************************************************)
signature LANG_C =
  sig
	type absyn
    val abstract: (string * absyn) list -> (string * Compare.exp) list
	val parsel  : string list -> (string * absyn) list
  end

structure Lang_C:LANG_C =
  struct

open ParseTree
open Compare
open List
type absyn = externalDecl list

fun chng_qualifier p =
  case p of
  CONST			=> CON 12
| VOLATILE		=> CON 13

fun chng_expression p =
  case p of
  EmptyExpr		=> CON 7
| IntConst n	=> CON 8
| RealConst	s	=> CON 9
| String s		=> CON 10
| Id s			=> CON 11
| Unop (o,e)	=> NODE(24,[chng_expression e])
| Binop(o,e,e')	=> NODE(25,[chng_expression e,chng_expression e'])
| QuestionColon (e0,e1,e2) => NODE(26,[chng_expression e0,chng_expression e1,
				chng_expression e2])
| Call(e,l)		=> NODE(27,chng_expression e::map chng_expression l)
| Cast(t,e)		=> NODE(28,[chng_expression e])
| InitList l	=> NODE(29,map chng_expression l)
| MARKexpression e => NODE(30,[chng_expression e])

(*
and chng_declarator p =
  case p of
  EmptyDecr		=> CON 4
| EllipsesDecr	=> CON 5
| VarDecr s		=> CON 6
| ArrayDecr(d,e)=> NODE(19,[chng_declarator d,chng_expression e])
| PointerDecr d => NODE(20,[chng_declarator d])
| QualDecr(q,d)	=> NODE(21,[chng_qualifier q,chng_declarator d])
| FuncDecr(d,l)	=> NODE(22,chng_declarator d::
    fold_right (fn (_,d) ll => chng_declarator d:: ll) l [])
| MARKdeclarator d => NODE(23,[chng_declarator d])
*)
and chng_statement p =
  case p of
  Decl d 		=> NODE(5,[chng_declaration d])
| Expr e 		=> NODE(6,[chng_expression e])
| Compound l 	=> NODE(7,map chng_statement l)
| While (e,s) 	=> NODE(8,[chng_expression e,chng_statement s])
| Do (e,s)		=> NODE(9,[chng_expression e,chng_statement s])
| For (e0,e1,e2,s) => NODE(10,[chng_expression e0,chng_expression e1,
							   chng_expression e2,chng_statement s])
| Labeled (_,s)	=> NODE(11,[chng_statement s])
| CaseLabel (e,s)  => NODE(12,[chng_expression e,chng_statement s])
| DefaultLabel s=> NODE(13,[chng_statement s])
| Goto _		=> CON 1
| Break			=> CON 2
| Continue		=> CON 3
| Return e		=> NODE(14,[chng_expression e])
| IfThen (e,s)	=> NODE(15,[chng_expression e,chng_statement s])
| IfThenElse(e,s1,s2) => NODE(16,[chng_expression e,chng_statement s1,
								  chng_statement s2])
| Switch (e,s)	=> NODE(17,[chng_expression e,chng_statement s])
| MARKstatement s => NODE(18,[chng_statement s])

and chng_declaration p =
  case p of
  Declaration (d,l) => NODE(3,
    fold_right (fn (r,e) ll => [(*chng_declarator r,*)chng_expression e] @ ll) l [])
| MARKdeclaration d => NODE(4, [chng_declaration d])

fun chng_externalDecl p =
  case p of
  ExternalDecl d => CON 14
| FunctionDef r	 => NODE(2, // [(chng_declarator r.funDecr)] @
							map chng_declaration r.krParams @
						   [(chng_statement r.body)])
| MARKexternalDecl e => chng_externalDecl e

fun change l = NODE(0,map chng_externalDecl l)

fun abstract l = map (fn (n,e) => (n, change e)) l

(************** Parsing files *********************************)
(************** and triangular n(n-1)/2 times iteration *******)
fun parse_file filename =
  let val ic = open_in_bin filename
      val lexbuf = Lexing.from_channel ic handle
            x => close_in ic; raise x
      val ast = C_parser.goal C_lexer.token lexbuf
  in
    close_in ic;
    ast
  end

fun pre_process filename =
  if Unix.WEXITED 0 = 
  	 Unix.system("/lib/cpp "^filename^" /tmp/clck-temporary.c")
  then
  	 "/tmp/clck-temporary.c"
  else raise C_lexer.Error "Can't execute cpp"


fun parsel [] = []
  | parsel(h::l) = (h,parse_file (pre_process h))::parsel l
                   handle _ =>
                        Format.eprintf "%s is an invalid file\n" h;
						flush stderr;	// flush the buffer
                        parsel l    	// ignore parse error file

end
