(*
 Sungsoon Jang, The LET Project, KAIST

 Copyright(c) 2001 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*)

(********************************************************************)
(*																	*)
(*	 					Base module signatures						*)
(*			 		2001. 6. 8. Jang Seong Soon						*)
(*																	*)
(********************************************************************)
signature COMPARE =
(* language independent part *)
  sig
    type exp = CON of int 
			 | NODE of int * exp list 
			 | VACANT_NODE of int * exp list

	val compares: (string * exp) list -> int * (string * string * real) list
  end  

(****************************************************************)
(*																*)
(*					Compares Program							*)
(*				2001. 6. 6.  Jang Sung Soon						*)
(*																*)
(****************************************************************)
structure Compare:COMPARE =
  struct

open Hashtbl
type exp = CON of int | NODE of int * exp list | VACANT_NODE of int * exp list

fun explode p =
  case p of 
    NODE(_,l) => p :: List.fold_right (fn x n => explode x @ n) l []
  | VACANT_NODE(_,l) => List.fold_right (fn x n => explode x @ n) l []
  | _ => []

fun compare_list l l' =
  let
    (* tbl1 contains (hash value, a tree in p') *)
	val tbl1 = create (List.length l')
	(* search the table and find the candidate possibly equal one... *)
	fun eqn x = List.exists (fn y => x = y) (find_all tbl1 (hash x))
	fun addit n x = if mem tbl1 n && eqn x then () else add tbl1 n x
  in
	(* Make a hash table of the 2nd tree *)
	List.iter (fn x => addit (hash x) x) l';
	List.fold_right (fn x n => if eqn x then n + 1 else n) l 0
  end

(****************************************************************)
(*																*)
(*	       return the similarity of two program					*)
(*      2 x equal node size / total of program size				*)
(*				2001. 6. 6   Jang Sung Soon						*)
(****************************************************************)

fun compare' p p' =
  let
	val l = List.rev(explode p)
	val l' = List.rev(explode p')
	val eqnode1 = compare_list l l'
	val eqnode2 = compare_list l' l
  in
	if List.length l + List.length l' = 0 then 0.0
	else (float_of_int(eqnode1 + eqnode2):real)
		/ float_of_int(List.length l + List.length l')
  end

fun trfold f [] = []
  | trfold f (h::l) = List.fold_right (fn x l => f h x::l) l (trfold f l)

fun count100 n = n - 100 * (n / 100) = 0
fun print_dot n l =
    (if count100 n then (prerr_string "."; flush stderr);
    if n = l then (prerr_string "\n"; flush stderr))

fun compares astl =
    let
		val n = List.length astl
		val num = ref 1
        fun check (an,a) (bn,b) =(
			print_dot !num ((n-1)*n/2);
            num++;
            (an,bn,compare' a b))
    in
        (n, trfold check astl)
        // iter (fn (s,_) => print_string s;print_string "\n") astl
    end

end

