/*
 Sunae Seo, The LET Project, KAIST

 Copyright(c) 2001 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*/

/*
JDK 1.2 Features added:
  strictfp modifier.
  explicit_constructor_invocation : ...
        | primary DOT THIS LPAREN argument_list_opt RPAREN SEMICOLON ;
  field_access : ...
        |       name DOT SUPER DOT IDENTIFIER ;
  method_invocation : ...
        |       name DOT SUPER DOT IDENTIFIER LPAREN argument_list_opt RPAREN ;
*/

/*
Yacc file for Java Parser (JDK 1.2) based on CUP java parser
 Last modification date : 2001/7/17
 First version initialized in 2001/5 by saseo
*/

%{
open Ast_gen

%}

/* Tokens */
%token BOOLEAN; // primitive_type
%token BYTE  SHORT  INT  LONG  CHAR  // integral_type
%token FLOAT  DOUBLE  // floating_point_type
%token LBRACK  RBRACK  // array_type
%token <string> IDENTIFIER  // name
%token DOT  // qualified_name
%token SEMICOLON  MULT  COMMA  LBRACE  RBRACE  EQ  LPAREN  RPAREN  COLON 
%token PACKAGE  // package_declaration
%token IMPORT  // import_declaration
%token PUBLIC  PROTECTED  PRIVATE  // modifier
%token STATIC  // modifier
%token ABSTRACT  FINAL  NATIVE  SYNCHRONIZED  TRANSIENT  VOLATILE 
%token CLASS  // class_declaration
%token EXTENDS  // super
%token IMPLEMENTS  // interfaces
%token VOID  // method_header
%token THROWS  // throws
%token THIS  SUPER  // explicit_constructor_invocation
%token INTERFACE  // interface_declaration
%token IF  ELSE  // if_then_statement  if_then_else_statement
%token SWITCH  // switch_statement
%token CASE  DEFAULT  // switch_label
%token DO  WHILE  // while_statement  do_statement
%token FOR  // for_statement
%token BREAK  // break_statement
%token CONTINUE  // continue_statement
%token RETURN  // return_statement
%token THROW  // throw_statement
%token TRY  // try_statement
%token CATCH  // catch_clause
%token FINALLY  // finally
%token NEW  // class_instance_creation_expression
%token PLUSPLUS  // postincrement_expression
%token MINUSMINUS  // postdecrement_expression
%token PLUS  MINUS  COMP  NOT  DIV  MOD 
%token LSHIFT  RSHIFT  URSHIFT  // shift_expression
%token LT  GT  LTEQ  GTEQ  INSTANCEOF  // relational_expression
%token EQEQ  NOTEQ  // equality_expression
%token AND  // and_expression
%token XOR  // exclusive_or_expression
%token OR   // inclusive_or_expression
%token ANDAND  // conditional_and_expression
%token OROR  // conditional_or_expression
%token QUESTION  // conditional_expression
%token MULTEQ  DIVEQ  MODEQ  PLUSEQ  MINUSEQ  // assignment_operator
%token LSHIFTEQ  RSHIFTEQ  URSHIFTEQ  // assignment_operator
%token ANDEQ  XOREQ  OREQ  // assignment_operator
%token EOF

%token <int> INTEGER_LITERAL 
%token <string> FLOATING_POINT_LITERAL 
%token <string> LONG_INTEGER_LITERAL 
%token <bool> BOOLEAN_LITERAL 
%token <char> CHARACTER_LITERAL 
%token <string> STRING_LITERAL 
%token NULL_LITERAL 

// strictfp keyword  new in Java 1.2
%token STRICTFP 

// Reserved but unused:
%token CONST  GOTO 

%start goal
%type <Ast_gen.top_dec> goal 

%%

// 19.2) The Syntactic Grammar
goal: 
 compilation_unit {$1}
 ;

// 19.3) Lexical Structure.
literal : INTEGER_LITERAL  {PInt $1}
 | LONG_INTEGER_LITERAL {PLInt $1}
 | FLOATING_POINT_LITERAL {PFloat $1}
 | BOOLEAN_LITERAL   {PBool $1}
 | CHARACTER_LITERAL  {PChar $1}
 | STRING_LITERAL   {PString $1}
 | NULL_LITERAL   {PNull}
 ;

// 19.4) Types, Values, and Variables
type : primitive_type  {PrimTy($1)}
 | reference_type   {RefTy($1)}
 ;
primitive_type :
  numeric_type   {$1}
 | BOOLEAN     {Boolean}
 ;
numeric_type: integral_type {$1}
 | floating_point_type  {$1}
 ;
integral_type : 
  BYTE      {Byte}
 | SHORT      {Short}
 | INT      {Int}
 | LONG      {Long}
 | CHAR      {Char}
 ;
floating_point_type : 
  FLOAT      {Float}
 | DOUBLE     {Double}
 ;

reference_type :
  class_or_interface_type {ClassInterfTy ($1)}
 | array_type    {$1}
 ;
class_or_interface_type : name  {$1};

class_type : class_or_interface_type  {RefTy(ClassInterfTy $1)}; 
interface_type : class_or_interface_type {RefTy(ClassInterfTy $1)};  

array_type : primitive_type dims {ArrayTy1($1,$2)}
 | name dims     {ArrayTy2($1,$2)}
 ;

// 19.5) Names
name : simple_name   {SName ($1)}
 | qualified_name   {$1}
 ;
simple_name : IDENTIFIER   {$1}
 ;
qualified_name :
  name DOT IDENTIFIER  {LName($1, $3)}
 ;

// 19.6) Packages
compilation_unit :
  package_declaration_opt 
  import_declarations_opt
  type_declarations_opt  {CompileUnit ($1, $2, $3)}
  ;
package_declaration_opt :    {None}
 | package_declaration  {Some($1)}
 ;
import_declarations_opt :   {[]}
 | import_declarations   {List.rev $1} 
 ;
type_declarations_opt   :   {[]}
 | type_declarations   {List.rev $1} 
 ;

import_declarations : 
  import_declaration  {[$1]}
 | import_declarations import_declaration {$2::$1}
 ;
type_declarations : 
  type_declaration  {[$1]}
 | type_declarations type_declaration  {$2::$1}
 ;
package_declaration : 
  PACKAGE name SEMICOLON  {$2}
 ;
import_declaration :  
  single_type_import_declaration   {SImp($1)}
 | type_import_on_demand_declaration {MImp($1)}
 ;
single_type_import_declaration : 
  IMPORT name SEMICOLON    {$2}
 ;
type_import_on_demand_declaration :
  IMPORT name DOT MULT SEMICOLON  {$2}
 ;
type_declaration :    
  class_declaration  {TClassDec ($1)}
 | interface_declaration {TInterfDec ($1)}
 | SEMICOLON    {TNothing}
 ;

// 19.7) Productions used only in the LALR(1) grammar
modifiers_opt:     {[]}
 | modifiers    {List.rev $1}
 ;
modifiers :  modifier  {[$1]}
 | modifiers modifier  {$2::$1}
 ;
modifier : PUBLIC   {Public}
 |  PROTECTED   {Protected}
 |  PRIVATE   {Private}
 | STATIC   {Static}
 | ABSTRACT   {Abstract}
 |  FINAL    {Final}
 |  NATIVE    {Native}
 |  SYNCHRONIZED  {Synchronized}
 |  TRANSIENT   {Transient}
 |  VOLATILE  {Volatile}
 | STRICTFP   {Strictfp}  
     // note that semantic analysis must check that the
                  // context of the modifier allows strictfp.
 ;

// 19.8) Classes

// 19.8.1) Class Declaration:
class_declaration : 
 modifiers_opt CLASS IDENTIFIER super_opt interfaces_opt class_body
 {($1, SName ($3), $4, $5, $6)}
 ;
super : EXTENDS class_type {$2}
 ;
super_opt :   {None}
 | super  {Some($1)}
 ;
interfaces : IMPLEMENTS interface_type_list {List.rev $2}
 ;
interfaces_opt:  {[]}
 | interfaces  {$1}
 ;
interface_type_list : 
  interface_type        {[$1]} 
 | interface_type_list COMMA interface_type {$3::$1}
 ;
class_body : LBRACE class_body_declarations_opt RBRACE  {$2}
 ;
class_body_declarations_opt :    {[]}
 | class_body_declarations ;  {List.rev $1}
class_body_declarations : 
  class_body_declaration       {[$1]}
 | class_body_declarations class_body_declaration {$2::$1}
 ;
class_body_declaration :
  class_member_declaration {$1}
 | static_initializer   {StaticInit ($1)}
 | constructor_declaration  {ConstrDec ($1)}
 | block      {BlockDec ($1)}
 ;
class_member_declaration :
  field_declaration   {FieldDec($1)}
 | method_declaration   {MethodDec($1)}
 /* repeat the prod for 'class_declaration' here: */
 | modifiers_opt CLASS IDENTIFIER super_opt interfaces_opt class_body
    {ClassDec ($1,SName ($3),$4,$5,$6)}
 | interface_declaration  {InterfDec($1)}
 ;

// 19.8.2) Field Declarations
field_declaration : 
  modifiers_opt type variable_declarators SEMICOLON
  {($1, $2, List.rev $3)}
 ;
variable_declarators :
  variable_declarator  {[$1]}
 | variable_declarators COMMA variable_declarator {$3::$1}
 ;
variable_declarator :
  variable_declarator_id   {VarId($1)}
 | variable_declarator_id EQ variable_initializer 
    {let val (id, br) = $1 
     in
      VarInitId (id, br, $3)
     end}
 ;
variable_declarator_id :
  IDENTIFIER  {(SName ($1), [])}
 | variable_declarator_id LBRACK RBRACK 
     {let val (id,br) = $1
      in 
       (id, LRBrack::br)
      end}
 ;
variable_initializer :
  expression   {VarInitExpr ($1)}
 | array_initializer {VarInitArray ($1)}
 ;

// 19.8.3) Method Declarations
method_declaration :
  method_header method_body  {($1,$2)}
 ;
method_header :
  modifiers_opt type method_declarator throws_opt 
   {let val (id,param,br) = $3
    in
      ($1, Some($2), id, param, br, $4)
    end}
 | modifiers_opt VOID method_declarator throws_opt 
   {let val (id,param,br) = $3
    in
      ($1, None, id, param, br, $4)
    end}
 ;
method_declarator :
  IDENTIFIER LPAREN formal_parameter_list_opt RPAREN {(SName ($1),$3,[])}
 | method_declarator LBRACK RBRACK // deprecated
   {let val (id,param,br) = $1
    in
      (id,param,LRBrack::br)
    end}
 // be careful; the above production also allows 'void foo() []'
 ;
formal_parameter_list_opt :  {[]}
 | formal_parameter_list {List.rev $1}
 ;
formal_parameter_list :
  formal_parameter  {[$1]}
 | formal_parameter_list COMMA formal_parameter {$3::$1}
 ;
formal_parameter :
  type variable_declarator_id   
   {let val (na,br) = $2
    in
     FormalPara(FinalNo,$1,na,br)
    end}
 | FINAL type variable_declarator_id
   {let val (na,br) = $3
    in
     FormalPara(FinalYes,$2,na,br)
    end}
 ;
throws_opt :  {[]} 
 | throws  {List.rev $1}
 ;
throws : THROWS class_type_list {$2}
 ;
class_type_list :
  class_type  {[$1]}
 | class_type_list COMMA class_type {$3::$1}
 ;
method_body : 
  block  {$1}
 | SEMICOLON {[]}
 ;

// 19.8.4) Static Initializers
static_initializer :
  STATIC block {$2} 
 ;

// 19.8.5) Constructor Declarations
constructor_declaration :
  modifiers_opt constructor_declarator throws_opt 
   constructor_body
     {let val (nm,para) = $2 
      in
      ($1,nm,para,$3,$4)
      end}
 ;
constructor_declarator :
  simple_name LPAREN formal_parameter_list_opt RPAREN  {(SName $1,$3)}
 ;
constructor_body :
  LBRACE explicit_constructor_invocation
   block_statements RBRACE  {(Some ($2), List.rev $3)}
 | LBRACE explicit_constructor_invocation RBRACE {(Some ($2), [])}
 | LBRACE block_statements RBRACE {(None, List.rev $2)}
 | LBRACE RBRACE     {(None, [])}
 ;
explicit_constructor_invocation :
  THIS LPAREN argument_list_opt RPAREN SEMICOLON {ThisCall ($3)}
 | SUPER LPAREN argument_list_opt RPAREN SEMICOLON {SuperCall ($3)}
 | primary DOT THIS LPAREN argument_list_opt RPAREN SEMICOLON 
             {PrimThisCall ($1,$5)}
 | primary DOT SUPER LPAREN argument_list_opt RPAREN SEMICOLON
             {PrimSuperCall ($1,$5)}
 ;

// 19.9) Interfaces

// 19.9.1) Interface Declarations
interface_declaration :
  modifiers_opt INTERFACE IDENTIFIER extends_interfaces_opt 
   interface_body   {($1,SName ($3),$4,$5)}
 ;
extends_interfaces_opt :  {[]}
 | extends_interfaces  {List.rev $1} 
 ;
extends_interfaces :
  EXTENDS interface_type {[$2]}
 | extends_interfaces COMMA interface_type {$3::$1}
 ;
interface_body :
  LBRACE interface_member_declarations_opt RBRACE {$2}
 ;
interface_member_declarations_opt :   {[]}
 | interface_member_declarations  {List.rev $1}
 ;
interface_member_declarations :
  interface_member_declaration  {[$1]}
 | interface_member_declarations interface_member_declaration
           {$2::$1}
 ;
interface_member_declaration :
  constant_declaration   {$1}
 | abstract_method_declaration  {$1}
 | class_declaration    {ClassDec ($1)}
 | interface_declaration   {InterfDec ($1)}
 ;
constant_declaration :
  field_declaration    {FieldDec ($1)}
 // need to semantically check that modifiers of field declaration
 // include only PUBIC, STATIC, or FINAL.  Other modifiers are
 // disallowed.
 ;
abstract_method_declaration :
  method_header SEMICOLON   {MethodDec ($1,[])}
 ;

// 19.10) Arrays
array_initializer :
  LBRACE variable_initializers COMMA RBRACE {ArrInitC (List.rev $2)}
 | LBRACE variable_initializers RBRACE   {ArrInitNC (List.rev $2)}
 | LBRACE COMMA RBRACE       {ArrInitC ([])}
 | LBRACE RBRACE        {ArrInitNC ([])}
 ;
variable_initializers :
  variable_initializer   {[$1]}
 | variable_initializers COMMA variable_initializer {$3::$1}
 ;

// 19.11) Blocks and Statements
block : LBRACE block_statements_opt RBRACE {$2}
 ;
block_statements_opt :
       {[]}
 | block_statements {List.rev $1}
 ;
block_statements :
  block_statement      {[$1]}
 | block_statements block_statement {$2::$1}
 ;
block_statement :
  local_variable_declaration_statement {$1}
 | statement        {BSStmt ($1)}
 | class_declaration      {BSClassDec ($1)}
 | interface_declaration     {BSInterfDec ($1)}
 ;
local_variable_declaration_statement :
  local_variable_declaration SEMICOLON {BSVarDecl ($1)}
 ;
local_variable_declaration :
  type variable_declarators   {(FinalNo,$1,List.rev $2)}
 | FINAL type variable_declarators  {(FinalYes,$2,List.rev $3)}
 ;
statement : statement_without_trailing_substatement {$1}
 | labeled_statement   {LabeledStmt ($1)}
 | if_then_statement   {IfThenStmt ($1)}
 | if_then_else_statement  {IfThenElseStmt ($1)}
 | while_statement    {WhileStmt ($1)}
 | for_statement    {ForStmt ($1)}
 ;
statement_no_short_if :
  statement_without_trailing_substatement {$1}
 | labeled_statement_no_short_if   {LabeledStmt ($1)}
 | if_then_else_statement_no_short_if  {IfThenElseStmt ($1)}
 | while_statement_no_short_if    {WhileStmt ($1)}
 | for_statement_no_short_if    {ForStmt ($1)}
 ;
statement_without_trailing_substatement :
  block     {BlockStmt ($1)}
 | empty_statement   {EmptyStmt}
 | expression_statement {ExprStmt ($1)}
 | switch_statement  {SwitchStmt ($1)}
 | do_statement   {DoStmt ($1)}
 | break_statement   {BreakStmt ($1)}
 | continue_statement  {ContinueStmt ($1)}
 | return_statement  {ReturnStmt ($1)}
 | synchronized_statement {SynchStmt ($1)}
 | throw_statement   {ThrowStmt ($1)}
 | try_statement   {TryStmt ($1)}
 ;
empty_statement :
  SEMICOLON  {}
 ;
labeled_statement :
  IDENTIFIER COLON statement {(SName ($1),$3)}
 ;
labeled_statement_no_short_if :
  IDENTIFIER COLON statement_no_short_if {(SName ($1),$3)}
 ;
expression_statement :
  statement_expression SEMICOLON  {$1}
 ;
statement_expression : 
  assignment       {Assignment ($1)}
 | preincrement_expression  
   {Expression($1)}
 | predecrement_expression
   {Expression($1)}
 | postincrement_expression
   {Expression($1)}
 | postdecrement_expression
   {Expression($1)}
 | method_invocation
   {Expression(PrimaryExpr (PMethodInv ($1)))}
 | class_instance_creation_expression
   {Expression(PrimaryExpr (PClassInstanceCreat($1)))}
 ;
if_then_statement :
  IF LPAREN expression RPAREN statement {($3,$5)}
 ;
if_then_else_statement :
  IF LPAREN expression RPAREN statement_no_short_if 
   ELSE statement
     {($3,$5,$7)}
 ;
if_then_else_statement_no_short_if :
  IF LPAREN expression RPAREN statement_no_short_if
   ELSE statement_no_short_if
     {($3,$5,$7)}
 ;
switch_statement :
  SWITCH LPAREN expression RPAREN switch_block {($3,List.rev $5)}
 ;
switch_block :
  LBRACE switch_block_statement_groups switch_labels RBRACE 
              {([$3],[])::$2}
 | LBRACE switch_block_statement_groups RBRACE  {$2}
 | LBRACE switch_labels RBRACE      {[([$2],[])]}
 | LBRACE RBRACE         {[([],[])]}
 ;
switch_block_statement_groups :
  switch_block_statement_group {[$1]}
 | switch_block_statement_groups switch_block_statement_group {$2::$1}
 ;
switch_block_statement_group :
  switch_labels block_statements {(List.rev $1,List.rev $2)}
 ;
switch_labels :
  switch_label    {[$1]}
 | switch_labels switch_label {$2::$1}
 ;
switch_label :
  CASE constant_expression COLON {Case ($2)}
 | DEFAULT COLON     {Default}
 ;

while_statement :
  WHILE LPAREN expression RPAREN statement {($3,$5)}
 ;
while_statement_no_short_if :
  WHILE LPAREN expression RPAREN statement_no_short_if   {($3,$5)}
 ;
do_statement :
  DO statement WHILE LPAREN expression RPAREN SEMICOLON {($2,$5)}
 ;
for_statement :
  FOR LPAREN for_init_opt SEMICOLON expression_opt SEMICOLON
   for_update_opt RPAREN statement
    {($3,$5,$7,$9)}
 ;
for_statement_no_short_if :
  FOR LPAREN for_init_opt SEMICOLON expression_opt SEMICOLON
   for_update_opt RPAREN statement_no_short_if
    {($3,$5,$7,$9)}
 ;
for_init_opt :  {None}
 | for_init {Some ($1)}
 ;
for_init : statement_expression_list {ForInit1 (List.rev $1)}
 | local_variable_declaration  {ForInit2 ($1)}
 ;
for_update_opt : {[]}
 | for_update {$1}
 ;
for_update : statement_expression_list {List.rev $1}
 ;
statement_expression_list :
  statement_expression  {[$1]}
 | statement_expression_list COMMA statement_expression {$3::$1}
 ;

identifier_opt : {None}
 | IDENTIFIER {Some (SName ($1))}
 ;

break_statement :
  BREAK identifier_opt SEMICOLON  {$2}
 ;

continue_statement :
  CONTINUE identifier_opt SEMICOLON {$2}
 ;
return_statement :
  RETURN expression_opt SEMICOLON  {$2}
 ;
throw_statement :
  THROW expression SEMICOLON   {$2}
 ;
synchronized_statement :
  SYNCHRONIZED LPAREN expression RPAREN block {($3,$5)}
 ;
try_statement :
  TRY block catches    {($2,List.rev $3, [])}
 | TRY block catches_opt finally {($2, $3, $4)}
 ;
catches_opt :  {[]}
 | catches  {List.rev $1}
 ;
catches : catch_clause  {[$1]}
 | catches catch_clause {$2::$1} 
 ;
catch_clause :
  CATCH LPAREN formal_parameter RPAREN block {Catch ($3,$5)}
 ;
finally : FINALLY block   {$2}
 ;

// 19.12) Expressions
primary : primary_no_new_array {
//  pa_primary 0 $1;  ps "\n---------------\n";
  $1}
 | array_creation_expression {
//  pa_primary 0 $1;  ps "\n---------------\n";
 $1}
 ;
primary_no_new_array :
  literal        {$1}
 | THIS        {PThis}
 | LPAREN expression RPAREN   {PExpr ($2)}
 | class_instance_creation_expression {PClassInstanceCreat ($1) }
 | field_access      {PFieldAcc ($1)} 
 | method_invocation     {PMethodInv ($1)}
 | array_access      {PArrayAcc ($1)}
 | primitive_type DOT CLASS   {PPClass ($1)}
 | VOID DOT CLASS      {PVClass}
 | array_type DOT CLASS    {PAClass ($1)}
 | name DOT CLASS      {PNClass ($1)}
 | name DOT THIS      {PNThis ($1)}
 ;
class_instance_creation_expression :
  NEW class_type LPAREN argument_list_opt RPAREN
           {(None,$2,$4,None)}
 | NEW class_type LPAREN argument_list_opt RPAREN class_body
           {(None,$2,$4,Some($6))}
 | primary DOT NEW IDENTIFIER
   LPAREN argument_list_opt RPAREN
           {(Some($1),SName ($4),$6,None)}
 | primary DOT NEW IDENTIFIER
   LPAREN argument_list_opt RPAREN class_body
           {(Some($1),SName ($4),$6,Some($8))}
 ;
argument_list_opt :      {[]}
 | argument_list     {List.rev $1}
 ;
argument_list :
  expression      {[$1]}
 | argument_list COMMA expression {$3::$1}
 ;
array_creation_expression :
  NEW primitive_type dim_exprs dims_opt   
      {PArrayCreat1(PrimTy $2,List.rev $3,$4)}
 | NEW class_or_interface_type dim_exprs dims_opt 
      {PArrayCreat1(RefTy(ClassInterfTy $2),List.rev $3,$4)}
 | NEW primitive_type dims array_initializer  
      {PArrayCreat2(PrimTy $2,List.rev $3,$4)}
 | NEW class_or_interface_type dims array_initializer
      {PArrayCreat2(RefTy(ClassInterfTy $2),List.rev $3,$4)}
             
 ;
dim_exprs : dim_expr  {[$1]}
 | dim_exprs dim_expr {$2::$1}
 ;
dim_expr : LBRACK expression RBRACK {$2}
 ;
dims_opt :  {[]}
 | dims {List.rev $1}
 ;
dims : LBRACK RBRACK   {[LRBrack]}
 | dims LBRACK RBRACK  {LRBrack::$1}
 ;
field_access :
  primary DOT IDENTIFIER   {PrimName ($1,SName ($3))}
 | SUPER DOT IDENTIFIER   {SuperName (SName ($3))}
 | name DOT SUPER DOT IDENTIFIER {NameSuperName ($1,SName ($5))}
 ;
method_invocation :
  name LPAREN argument_list_opt RPAREN {(Name ($1),$3)}
 | primary DOT IDENTIFIER LPAREN argument_list_opt RPAREN
          {(PrimName ($1,SName ($3)),$5)}
 | SUPER DOT IDENTIFIER LPAREN argument_list_opt RPAREN
          {(SuperName (SName ($3)),$5)}
 | name DOT SUPER DOT IDENTIFIER LPAREN argument_list_opt RPAREN
          {(NameSuperName ($1,SName ($5)),$7)}
 ;
array_access :
  name LBRACK expression RBRACK {SArrayAcc ($1,$3)}
 | primary_no_new_array LBRACK expression RBRACK
          {CArrayAcc ($1,$3)} 
 ;
postfix_expression :
  primary      {PrimaryExpr ($1)}
 | name      {NameExpr ($1)}
 | postincrement_expression {$1}
 | postdecrement_expression {$1}
 ;
postincrement_expression :
  postfix_expression PLUSPLUS  {UnaryExpr (UPostIncr, $1)}
 ;
postdecrement_expression :
  postfix_expression MINUSMINUS {UnaryExpr (UPostDecr, $1)}
 ;
unary_expression :
  preincrement_expression   {$1}
 | predecrement_expression   {$1}
 | PLUS unary_expression   {UnaryExpr (UPlus,$2)}
 | MINUS unary_expression   {UnaryExpr (UMinus,$2)}
 | unary_expression_not_plus_minus {$1}
 ;
preincrement_expression :
  PLUSPLUS unary_expression  {UnaryExpr (UPreIncr,$2)}
 ;
predecrement_expression :
  MINUSMINUS unary_expression  {UnaryExpr (UPreDecr,$2)}
 ;
unary_expression_not_plus_minus :
  postfix_expression  {$1}
 | COMP unary_expression {UnaryExpr (UCompare,$2)}
 | NOT unary_expression {UnaryExpr (UNot,$2)}
 | cast_expression   {$1}
 ;
cast_expression :
  LPAREN primitive_type dims_opt RPAREN unary_expression
        {UnaryExpr (UCast1 ($2,$3),$5)}
 | LPAREN expression RPAREN unary_expression_not_plus_minus
        {UnaryExpr (UCast2 $2, $4)}
 | LPAREN name dims RPAREN unary_expression_not_plus_minus
        {UnaryExpr (UCast3 ($2,$3), $5)}
 ;
multiplicative_expression :
  unary_expression  {$1}
 | multiplicative_expression MULT unary_expression {BinExpr (BMult,$1,$3)} 
 | multiplicative_expression DIV unary_expression {BinExpr (BDiv,$1,$3)}
 | multiplicative_expression MOD unary_expression {BinExpr (BMod,$1,$3)}
 ;
additive_expression :
  multiplicative_expression {$1}
 | additive_expression PLUS multiplicative_expression 
        {BinExpr (BPlus,$1,$3)}
 | additive_expression MINUS multiplicative_expression
        {BinExpr (BMinus,$1,$3)}
 ;
shift_expression :
  additive_expression   {$1}
 | shift_expression LSHIFT additive_expression {BinExpr (BLShift,$1,$3)}
 | shift_expression RSHIFT additive_expression {BinExpr (BRShift,$1,$3)}
 | shift_expression URSHIFT additive_expression{BinExpr(BURShift,$1,$3)}
 ;
relational_expression :
  shift_expression   {$1}
 | relational_expression LT shift_expression  {BinExpr (BLt,$1,$3)}
 | relational_expression GT shift_expression  {BinExpr (BGt,$1,$3)}
 | relational_expression LTEQ shift_expression  {BinExpr (BLtEq,$1,$3)}
 | relational_expression GTEQ shift_expression  {BinExpr (BGtEq,$1,$3)}
 | relational_expression INSTANCEOF reference_type {InstanceOfExpr ($1,$3)}
 ;
equality_expression :
  relational_expression  {$1}
 | equality_expression EQEQ relational_expression {BinExpr (BEqEq,$1,$3)}
 | equality_expression NOTEQ relational_expression {BinExpr (BNotEq,$1,$3)}
 ;
and_expression :
  equality_expression   {$1}
 | and_expression AND equality_expression {BinExpr (BAnd,$1,$3)}
 ;
exclusive_or_expression :
  and_expression    {$1}
 | exclusive_or_expression XOR and_expression {BinExpr (BXOr,$1,$3)}
 ;
inclusive_or_expression :
  exclusive_or_expression  {$1} 
 | inclusive_or_expression OR exclusive_or_expression {BinExpr (BOr,$1,$3)}
 ;
conditional_and_expression :
  inclusive_or_expression  {$1}
 | conditional_and_expression ANDAND inclusive_or_expression
         {AndAndExpr ($1,$3)}
 ;
conditional_or_expression :
  conditional_and_expression {$1}
 | conditional_or_expression OROR conditional_and_expression
         {OrOrExpr ($1,$3)}
 ;
conditional_expression :
  conditional_or_expression {$1}
 | conditional_or_expression QUESTION expression 
   COLON conditional_expression
         {ConditionalExpr ($1,$3,$5)}
 ;
assignment_expression :
  conditional_expression {Expression ($1)}
 | assignment    {Assignment ($1)}
 ;
assignment : left_hand_side assignment_operator assignment_expression
      {($1,$2,$3)}
 ;
left_hand_side :
  name   {LHSName ($1)}
 | field_access {LHSFieldAcc ($1)}
 | array_access {LHSArrayAcc ($1)}
 ;
assignment_operator :
  EQ   {AEq}
 | MULTEQ  {AMultEq}
 | DIVEQ  {ADivEq}
 | MODEQ  {AModEq}
 | PLUSEQ  {APlusEq}
 | MINUSEQ  {AMinusEq}
 | LSHIFTEQ {ALShiftEq}
 | RSHIFTEQ {ARShiftEq}
 | URSHIFTEQ {AURShiftEq}
 | ANDEQ  {AAndEq}
 | XOREQ  {AXOrEq}
 | OREQ  {AOrEq}
 ;
expression_opt : {None}
 | expression {Some ($1)}
 ;
expression : assignment_expression {$1}
 ;
constant_expression :
  expression {$1}
 ;
