(*
 Sungsoon Jang, The LET Project, KAIST

 Copyright(c) 2001 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*)

(********************************************************************)
(*																	*)
(*	 	Change the ast tree appropriate to the comparison			*)
(*			 2001. 6. 15. Jang Seong Soon							*)
(*																	*)
(********************************************************************)

signature LANG_JAVA =
  sig
    type absyn
    val abstract: (string * absyn) list -> (string * Compare.exp) list
    val parsel  : string list -> (string * absyn) list
  end

structure Lang_JAVA: LANG_JAVA =
  struct

open Ast_gen
open Compare
open List
type absyn = top_dec


fun oops f None = [] | oops f Some e = [f e]
fun oopsl f None = [] | oopsl f Some l = map f l 
// 19.5) Names 
// 19.4) Types, Values and Variables
// primitive types
fun chng_array_acc p =
  case p of
    SArrayAcc(_,e)	=> chng_expr e // name[expr]
  | CArrayAcc(p,e)	=> NODE(13,[chng_primary p,chng_expr e])
			  		// primary[expr] 
					// here, primary doesn't include ArrayCreatExpr(..)

and chng_primary p =
  case p of
    PExpr e => NODE(8,[chng_expr e]) // (expression)
  | PThis	=> CON 42				// "this" expr
  | PInt _	=> CON 43
  | PLInt _ => CON 43
  | PFloat _=> CON 44
  | PBool _ => CON 45
  | PChar _ => CON 46
  | PString _ => CON 47
  | PNull 	=> CON 48			// literal 				
  | PClassInstanceCreat(p,_,l,b) 
  		=> NODE(10,oops chng_primary p @ map chng_expr l @
				   oopsl chng_class_body_dec b)
					// primary.class type name (arg list) {class body}
  | PFieldAcc _ => CON 49
  | PMethodInv (_,l) => NODE(11,CON 50::map chng_expr l)
			   				   // method_name (arg list)
  | PArrayAcc a => chng_array_acc a
  | PPClass _	=> CON 51  		// primitive_type.class
  | PVClass 	=> CON 52		// void.class
  | PAClass _   => CON 53		// array_type.class
  | PNClass _	=> CON 54		// name.class
  | PNThis _	=> CON 55		// name.this
  | PArrayCreat1(_,l,_) => NODE(12,map chng_expr l)
			// type [expr][expr]...[expr][][]...[]
  | PArrayCreat2(_,_,a)	=> chng_array_init a
		    // type []...[] = array_initializer

and chng_unary_op p =
  case p of
	UPlus	=> CON 32
  | UMinus	=> CON 33
  | UPostIncr => CON 34
  | UPostDecr => CON 35
  | UPreIncr => CON 36
  | UPreDecr => CON 37
  | UCompare => CON 38
  | UNot	=> CON 39
  | UCast1 _ => CON 40
  | UCast2 e => chng_expr e
  | UCast3 _ => CON 41

and chng_binary_op p = 
  case p of
    BOr 	=> CON 15
  | BXOr 	=> CON 16
  | BAnd	=> CON 17
  | BEqEq	=> CON 18
  | BNotEq	=> CON 19
  | BLt		=> CON 20
  | BGt		=> CON 21
  | BLtEq	=> CON 22
  | BGtEq 	=> CON 23
  | BLShift	=> CON 24
  | BRShift	=> CON 25
  | BURShift=> CON 26
  | BPlus	=> CON 27
  | BMinus	=> CON 28
  | BMult	=> CON 29
  | BDiv	=> CON 30
  | BMod	=> CON 31

and chng_cond_expr p =
  case p of
    PrimaryExpr e 		=> chng_primary e
  | NameExpr n	  		=> CON 15
  | UnaryExpr(u,e) 		=> NODE(3,[chng_unary_op u,chng_cond_expr e])
  | AndAndExpr(e,e') 	=> NODE(4,map chng_cond_expr [e,e'])
  | OrOrExpr(e,e')		=> NODE(5,map chng_cond_expr [e,e'])
  | ConditionalExpr(e,e',e'') => NODE(6,chng_cond_expr e::chng_expr e'::
  						   [chng_cond_expr e''])
  | InstanceOfExpr(e,_)	=> chng_cond_expr e
  | BinExpr(b,e,e')		=> NODE(7,chng_binary_op b :: map chng_cond_expr [e,e'])

and chng_assign_op p =
  case p of
    AEq 	=> CON 3
  | AMultEq => CON 4
  | ADivEq 	=> CON 5
  | AModEq  => CON 6
  | APlusEq => CON 7
  | AMinusEq => CON 8
  | ALShiftEq => CON 9
  | ARShiftEq => CON 10
  | AURShiftEq => CON 11
  | AAndEq	=> CON 12
  | AXOrEq	=> CON 13
  | AOrEq	=> CON 14

and chng_expr p =
  case p of
    Expression c => chng_cond_expr c
  | Assignment(_,op,e) => VACANT_NODE(5,[chng_assign_op op,chng_expr e])

and chng_block_stmt p =
  case p of
    BSVarDecl(f,_,l) => NODE(9,map chng_var_dec l)
  | BSStmt s		 => chng_stmt s
  | BSClassDec c	 => chng_class_dec c
  | BSInterfDec	i	 => chng_interf_dec i

and chng_var_init p =
  case p of
    VarInitExpr e => chng_expr e
  | VarInitArray a => chng_array_init a

and chng_array_init p =
  case p of
    ArrInitC l => VACANT_NODE(4,map chng_var_init l)
  | ArrInitNC l => VACANT_NODE(4,map chng_var_init l)

and chng_var_dec p =
  case p of
    VarId(_,_) => CON 2
  | VarInitId(_,_,i) => chng_var_init i

and chng_catch_clauses Catch (_,l) = NODE(24,map chng_block_stmt l)
and chng_for_init p =
  case p of
    ForInit1 l => NODE(25,map chng_expr l)
  | ForInit2 (_,_,l) => NODE(26,map chng_var_dec l)
and chng_stmt p =
  case p of
    BlockStmt l => NODE(13,map chng_block_stmt l)
  | EmptyStmt 	=> CON 56
  | ExprStmt e	=> NODE(14,[chng_expr e])
  | SwitchStmt(e,l) => NODE(15,chng_expr e::
  		fold_right (fn (_,o) ll => map chng_block_stmt o @ ll) l [])
  | DoStmt (s,e) 	=> NODE(16,[chng_stmt s,chng_expr e])
  | BreakStmt _		=> CON 57
  | ContinueStmt _ 	=> CON 58
  | ReturnStmt e  	=> NODE(15,oops chng_expr e)
  | SynchStmt(e,l) 	=> NODE(16,chng_expr e::map chng_block_stmt l)
  | ThrowStmt e		=> NODE(17,[chng_expr e])
  | TryStmt(l,l',l'') => NODE(18,map chng_block_stmt l @
  								 map chng_catch_clauses l' @
								 map chng_block_stmt l'')
  | LabeledStmt(_,e)=> NODE(19,[chng_stmt e])
  | IfThenStmt(e,s) => NODE(20,[chng_expr e,chng_stmt s])
  | IfThenElseStmt(e,s,s') => NODE(21,[chng_expr e,chng_stmt s,chng_stmt s'])
  | WhileStmt(e,s)	=> NODE(22,[chng_expr e,chng_stmt s])
  | ForStmt(f,e,l,s) => NODE(23,oops chng_for_init f @
  								oops chng_expr e @
								map chng_expr l @ [chng_stmt s])

and chng_class_body_dec p =
  case p of
    FieldDec (_,_,l) => NODE(0,map chng_var_dec l)
  | MethodDec(_,l)	 => NODE(1,map chng_block_stmt l)
  | ClassDec c		 => chng_class_dec c
  | InterfDec i		 => chng_interf_dec i
  | StaticInit l	 => NODE(2,map chng_block_stmt l)
  | ConstrDec(_,_,_,_,(_,l)) => NODE(27,map chng_block_stmt l)
  | BlockDec l		 => VACANT_NODE(3,map chng_block_stmt  l)
  | AbsMethodDec _	 => CON 1
and chng_class_dec (_,_,_,_,l) = VACANT_NODE(2,map chng_class_body_dec l)
and chng_interf_dec (_,_,_,l) = VACANT_NODE(1,map chng_class_body_dec l)

fun chng_typ_dec p =
  case p of
  	TNothing => CON 0
  | TClassDec c => chng_class_dec c
  | TInterfDec i => chng_interf_dec i
fun change CompileUnit(_,_,tl) = VACANT_NODE(0,map chng_typ_dec tl)

fun abstract l = map (fn (n,e) => (n, change e)) l

(************** Parsing files *********************************)
(************** and triangular n(n-1)/2 times iteration *******)
fun parse_file filename =
  let val ic = open_in_bin filename
      val lexbuf = Lexing.from_channel ic handle
            x => close_in ic; raise x
      val ast = Java_parser.goal Java_lexer.token lexbuf
  in
    close_in ic;
    ast
  end

fun parsel [] = []
  | parsel(h::l) = (h,parse_file h)::parsel l
                   handle _ =>
                        Format.eprintf "%s is an invalid file\n" h;
						flush stderr;	// flush the buffer
                        parsel l    	// remove parse error file


end
