(*
 Sungsoon Jang, The LET Project, KAIST

 Copyright(c) 2001 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*)

(********************************************************************)
(*																	*)
(*	 	Change the ast tree appropriate to the comparison			*)
(*		To include the size of the program 							*)
(*			 2001. 4. 9. Jang Seong Soon							*)
(*																	*)
(********************************************************************)

signature LANG_nML = 
  sig
    type absyn
	type exp
    val abstract: (string * absyn) list -> (string * exp) list
    val parsel  : string list -> (string * absyn) list
	val compares: (string * exp) list -> int * (string * string * real) list
  end

structure AbstractAst =
  struct

type exp =
   UnitExp 
 | IntExp
 | RealExp
 | StringExp
 | CharExp
 | VarExp
 | ConExp		   of string			// Compare constructor name
 | AppExp          of exps * exps list
 | RecordExp
 | RecordFieldExp
 | ArrayFieldExp   of exps * exps		// Array field
 | UpdateArrayExp  of exps * exps * exps 
 | SubstRecordExp
 | TupleExp        of exps list
 | ListExp         of exps list
 | ArrayExp		   of exps list
 | LetExp          of exps * exps
 | HandleExp       of exps * exps list  // + list size
 | RaiseExp        of exps
 | FnExp           of exps list			// + list size 
 | AssignExp       of exps * exps
 | RefExp          of exps
 | DeRefExp        of exps
 | SeqExp          of exps list			// sequence itself have no size
 		// Because "(a)" is interpreted as single sequence.
		// I want "(a)" have same as "a".
 | CaseExp         of exps * exps list	// + list size
 | IfExp           of exps * exps * exps
 | WhileExp        of exps * exps
 | ForExp          of exps * exps * exps * exps
 | ConstraintExp   of exps
 | ValDec       of exps list 			// + list size
 | FunDec       of exps list 			// + list size
 | SeqDec       of exps list

and exps = exp * int

fun size ((_,n):exps) = n
val size_dec = size
fun expof ((a,_):exps) = a

(*
fun print_exps (a,b) = print_exp a; print_int b; print_string " "
and print_exp p =
 case p of
   ValDec l => print_string "ValDec ";List.iter print_exps l
 | FunDec l => print_string "FunDec ";List.iter print_exps l
 | SeqDec l => print_string "SeqDec\n";List.iter print_exps l
 | AppExp(e,e')=> print_string "AppExp ( ";
 				  print_exps e; print_string " , "; List.iter print_exps e';
				  print_string " ) "
 | TupleExp l  => print_string "Tuple ( ";
 			      List.iter (fn x => print_exps x; print_string ", ") l;
				  print_string " ) "
 | ListExp l   => print_string "ListExp ( ";
 			      List.iter (fn x => print_exps x; print_string ", ") l;
				  print_string " ) "
 | LetExp(d,e) => print_string "LetExp ( ";
 				  print_exps d; print_string " , "; print_exps e;
				  print_string " ) "
 | HandleExp(e,l) => print_string "Handle "
 | RaiseExp e  => print_string "RaiseExp "
 | FnExp l	   => print_string "FnExp ( ";
 			      List.iter (fn x => print_exps x; print_string "\n") l;
				  print_string " ) "
 | AssignExp(e,e') => print_string "AssignExp ( ";
 				  print_exps e; print_string " , "; print_exps e';
				  print_string " ) "
 | RefExp e    => print_string "RefExp "; print_exps e
 | DeRefExp e  => print_string "DeRefExp "; print_exps e
 | SeqExp l    => print_string "SeqExp ( ";
				  List.iter (fn x => print_exps x; print_string "\n") l;
				  print_string " ) "
 | CaseExp(e,l)=> print_string "CaseExp "
 | IfExp(e,e1,e2) => print_string "IfExp ( "; 
 					 print_exps e;
					 print_string " ) ";
					 print_exps e1;
					 print_string " | ";
					 print_exps e2
 | WhileExp(e,e1) => print_string "WhileExp ( ";
 					 print_exps e;
					 print_string " ) ";
					 print_exps e1
 | ForExp(e,e1,e2,e3) => print_string "ForExp ( ";
 					 print_exps e;
					 print_string " , ";
					 print_exps e1;
					 print_string " , ";
					 print_exps e2;
					 print_string " ) ";
					 print_exps e3
 | ConstraintExp e => print_string "ConstraintExp "; print_exps e
 | ConExp s => print_string ("Constructor "^s)
 | ArrayExp l  => print_string "ArrayExp [| ";
				  List.iter (fn x => print_exps x; print_string ", ") l;
				  print_string " |] "

 | ArrayFieldExp(e,e1) => print_exps e;
 					 print_string ".[ ";
					 print_exps e1;
					 print_string " ] "
 | UpdateArrayExp(e,e1,e2) => print_exps e;
 					 print_string ".[ ";
 					 print_exps e1;
					 print_string " ] <- ";
					 print_exps e2
 | _ => print_string "BaseExp "

 val print_dec = print_exps
 *)
end

structure ChangeAst =
struct

open List
open StringAst
structure A = AbstractAst

(****************************************************************)
(*																*)
(*				return the stripped program						*)
(*				throw off topdec and strdec						*)
(*				2001. 3. 21  Jang Seong Soon					*)
(*																*)
(****************************************************************)

fun strip_fold f l = fold_right (fn p a => f p @ a) l []
fun strip_strexp p =
  case p of
    StrStr(d,_) => strip_strdec d
  | SigStr(e,_,_) => strip_strexp e
  | FctAppStr(_,l,_) => strip_fold strip_strexp l
  | _ => []
and strip_strbind (_,_,d,_) = strip_strexp d
and strip_strdec p =
  case p of
    SimpleDec(d,_) => [d]
  | StrDec(l,_) => strip_fold strip_strbind l
  | SeqStrDec(l,_) => strip_fold strip_strdec l

fun strip_fctdec (_,_,_,e,_) = strip_strexp e
fun strip_topdec p =
  case p of
    Str (d,_) => strip_strdec d
  | Fct (d,_) => strip_fctdec d
  | SeqTopDec (dl,_) => strip_fold strip_topdec dl
  | _ => []

fun strip p = SeqDec(strip_topdec p,Location.none)

(****************************************************************)
(*																*)
(*				return the changed program						*)
(*				throw off topdec and strdec						*)
(*				2001. 4. 10  Jang Seong Soon					*)
(*																*)
(****************************************************************)
fun sum l = fold_right (fn (_,m) n => m + n) l 0
  
fun change_rule (_,e,_) = change_exp e
and change_fnrule (_,e,_) = change_exp e
and change_exp p =
  case p of
    UnitExp _			=> (A.UnitExp,0)
  | IntExp _			=> (A.IntExp,0)
  | RealExp _			=> (A.RealExp,0)
  | StringExp _			=> (A.StringExp,0)
  | CharExp	_			=> (A.CharExp,0)
  | VarExp _			=> (A.VarExp,0)
  | ConExp ((_,(s,_),_),_)	=> (A.ConExp s,0)
  | AppExp(e,l,_)		=> let val e' = change_exp e
							   val l' = map change_exp l
						   in
						   	   (A.AppExp(e',l'),snd e' + sum l' + 1)
						   end
  | RecordExp _			=> (A.RecordExp,0)
  | RecordFieldExp _	=> (A.RecordFieldExp,0)
  | ArrayFieldExp(e,a,_)=> let val e' = change_exp e
  							   val a' = change_exp a 
						   in
						   	   (A.ArrayFieldExp(e',a'),snd e' + snd a' + 1)
						   end
  | UpdateArrayExp(e,e1,e2,_)
  						=> let val e' = change_exp e
							   val e1'= change_exp e1
							   val e2'= change_exp e2
						   in
						   		(A.UpdateArrayExp(e',e1',e2'),snd e' + snd e1' + snd e2' + 1)
						   end
  | SubstRecordExp _	=> (A.SubstRecordExp,0)
  | TupleExp(l,_)		=> let val l' = map change_exp l
  						   in 
						   	   (A.TupleExp l',sum l' + 1)
						   end
  | ListExp(l,_)		=> let val l' = map change_exp l
  						   in
						       (A.ListExp l',sum l' + 1)
						   end
  | ArrayExp(l,_)		=> let val l' = map change_exp l
  						   in
						       (A.ArrayExp l',sum l' + 1)
						   end
  | LetExp(d,e,_)		=> let val d' = change_dec d
  							   val e' = change_exp e
						   in
						       (A.LetExp(d',e'),snd d' + snd e' + 1)
						   end
  | HandleExp(e,l,_)	=> let val e' = change_exp e
  							   val l' = map change_rule l
			  			   in
							   (A.HandleExp(e',l'),snd e' + sum l' + 1)
						   end
  | RaiseExp(e,_)		=> let val e' = change_exp e
  						   in
						   	   (A.RaiseExp e',snd e' + 1)
						   end
  | FnExp(l,_)			=> let val l' = map change_fnrule l
  						   in
						   	   (A.FnExp l',sum l' + 1)
						   end
  | AssignExp(e,e1,_)	=> let val e'  = change_exp e
  							   val e1' = change_exp e1
						   in
						   	   (A.AssignExp(e',e1'),snd e' + snd e1' + 1)
						   end
  | RefExp(e,_)			=> let val e' = change_exp e
  						   in
						   	   (A.RefExp e',snd e' + 1)
						   end
  | DeRefExp(e,_)		=> let val e' = change_exp e
  						   in
						   	   (A.DeRefExp e',snd e' + 1)
						   end
  | SeqExp([e],_)		=> change_exp e
  | SeqExp(l,_)			=> let val l' = map change_exp l
  						   in
						   	   (A.SeqExp l',sum l')
						   end
  | CaseExp(e,l,_)		=> let val e' = change_exp e
  							   val l' = map change_rule l
			  			   in
							   (A.HandleExp(e',l'),snd e' + sum l' + 1)
						   end
  | IfExp(e,e1,e2,_)	=> let val e'  = change_exp e
  							   val e1' = change_exp e1
							   val e2' = change_exp e2
						   in
						   	   (A.IfExp(e',e1',e2'),snd e' + snd e1' + snd e2' + 1)
						   end
  | WhileExp(e,e1,_)	=> let val e'  = change_exp e
  							   val e1' = change_exp e1
						   in
						   	   (A.WhileExp(e',e1'),snd e' + snd e1' + 1)
						   end
  | ForExp(_,e,e1,e2,e3,_)=> let val e'  = change_exp e
  							   val e1' = change_exp e1
							   val e2' = change_exp e2
							   val e3' = change_exp e3
						   in
						   	   (A.ForExp(e',e1',e2',e3'),
							    snd e' + snd e1' + snd e2' + snd e3' + 1)
						   end
  | ConstraintExp(e,_,_)=> let val e' = change_exp e
  						   in
						   	   (A.ConstraintExp e',snd e' + 1)
						   end

and change_valbind (_,_,e,_) = change_exp e
and change_funbind (l,_) =
    map (fn (_,_,e,_) => change_exp e) l
and change_dec p =
  case p of
    ValDec(_,l,_) 	=> let val l' = map change_valbind l
					   in (A.ValDec l',sum l' + 1) end
  | RecValDec(_,l,_)=> let val l' = map change_valbind l
					   in (A.ValDec l',sum l' + 1) end
  | FunDec(_,l,_) 	=> let val l' = flatten(map change_funbind l)
					   in (A.FunDec l',sum l' + 1) end
  | SeqDec(l,_)		=> let val l' = map change_dec l
					   in (A.SeqDec l',sum l') end
  | _ => (A.SeqDec[],0)

fun change p = change_dec (strip p)

end

(****************************************************************)
(*																*)
(*				STRUCTURE Compare compare files	 				*)
(*			the ast tree changed to have each size 				*)
(*				2001. 4. 10  Jang Seong Soon					*)
(*																*)
(****************************************************************)

structure CompareAbstract =
  struct

open AbstractAst
open Hashtbl

(****************************************************************)
(*																*)
(*		explode : make all the subtree of p as a list			*)
(*				2001. 6. 7. Jang Sung Soon						*)
(*																*)
(****************************************************************)

fun sfold f l = List.fold_right (fn x n => f x @ n) l []
fun explode p =
  if size p = 0 then []
  else if size p = 1 then [p]
  else
	case expof p of
      AppExp(e,l)       => p :: sfold explode (e::l)
    | TupleExp l        => p :: sfold explode l
    | ListExp l         => p :: sfold explode l
    | LetExp(e,e')      => p :: sfold explode [e,e']
    | HandleExp(e,l)    => p :: sfold explode (e::l)
    | RaiseExp e        => p :: explode e
    | FnExp l           => p :: sfold explode l
    | AssignExp(e,e')   => p :: sfold explode [e,e']
    | RefExp e          => p :: explode e
    | DeRefExp e        => p :: explode e
    | SeqExp l          =>      sfold explode l
    | CaseExp(e,l)      => p :: sfold explode (e::l)
    | IfExp(e,e',e'')   => p :: sfold explode [e,e',e'']
    | WhileExp(e,e')    => p :: sfold explode [e,e']
    | ForExp(e,e1,e2,e3)=> p :: sfold explode [e,e1,e2,e3]
    | ConstraintExp e   => p :: explode e
    | ArrayExp l        => p :: sfold explode l
    | ArrayFieldExp(e,e')=>p :: sfold explode [e,e']
    | UpdateArrayExp(e,e',e'') => p :: sfold explode [e,e',e'']
    | ValDec l    		=> p :: sfold explode l
    | FunDec l    		=> p :: sfold explode l
    | SeqDec l    		=>      sfold explode l
	| _ => []

(****************************************************************)
(*																*)
(*					Compares Program							*)
(*				2001. 6. 6.  Jang Sung Soon						*)
(*																*)
(****************************************************************)

val sum = ref 0
val trial = ref 0

fun compares l l' =
  let
    // tbl1 contains (hash value, a tree in p')
	val tbl1 = create (List.length l')
	// tbl contains (a tree in p, trees in p')
	// val tbl  = create (List.length l + List.length l')
	// search the table and find the candidate possibly equal one...
	fun eqn x = List.exists (fn y => x = y) (find_all tbl1 (hash x))
	fun addit n x = if mem tbl1 n && eqn x then () else add tbl1 n x
    fun eqn2 x =
        let
          val l = (find_all tbl1 (hash x))
        in
          // sum += List.length l;
          // trial ++;
          List.exists (fn y => x = y) l
        end
  in
	// Make a hash table of the 2nd tree
	List.iter (fn x => addit (hash x) x) l';
	List.fold_right (fn x n => if eqn2 x then n + 1 else n) l 0
  end

(****************************************************************)
(*																*)
(*	       return the similarity of two program					*)
(*      2 x equal node size / total of program size				*)
(*				2001. 6. 6   Jang Sung Soon						*)
(****************************************************************)

fun compare' p p' =
  let
	val l = List.rev(explode p)
	val l' = List.rev(explode p')
	val eqnode1 = compares l l'
	val eqnode2 = compares l' l
  in
    // print_string "The trees of average hash: ";
    // print_float ((float_of_int(!sum):real)/ float_of_int(!trial));
    // print_string "\n";
	if size p + size p' = 0 then 0.0
	else (float_of_int(eqnode1 + eqnode2):real) / float_of_int(size p + size p')
  end

end

structure Lang_nML : LANG_nML =
  struct

type absyn = StringAst.topdec
type exp = AbstractAst.exps

fun abstract l = List.map (fn (n,e) => (n, ChangeAst.change e)) l

(************** Parsing files *********************************)
(************** and triangular n(n-1)/2 times iteration *******)
fun parse_file filename =
  let val ic = open_in_bin filename
      val ast = Parse.batch_parse (Lexing.from_channel ic) handle
            x => (close_in ic; raise x)
  in
    close_in ic;
    ast
  end

fun parsel [] = []
  | parsel(h::l) = (h,parse_file h)::parsel l
                   handle _ =>(
                        Format.eprintf "%s is an invalid file\n" h;
						flush stderr;	// flush the buffer
                        parsel l)    	// ignore parse error file

fun trfold f [] = []
  | trfold f (h::l) = List.fold_right (fn x l => f h x::l) l (trfold f l)

fun count100 n = n - 100 * (n / 100) = 0
fun print_dot n l =
    (if count100 n then (Format.eprintf "."; flush stderr);
	if n = l then (Format.eprintf "\n"; flush stderr))

fun compares astl =
    let
		val n = List.length astl
		val num = ref 1
        fun check (an,a) (bn,b) =(
			print_dot !num ((n-1)*n/2);
            num++;
			// Format.printf "%s %s %f\n" an bn (CompareAbstract.compare' a b);
            (an,bn,CompareAbstract.compare' a b))
    in
        (n, trfold check astl)
        // iter (fn (s,_) => print_string s;print_string "\n") astl
    end

end

