(*
 Sungsoon Jang, The LET Project, KAIST

 Copyright(c) 2001 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*)

(********************************************************************)
(*																	*)
(*	 	Change the ast tree appropriate to the comparison			*)
(*			 2001. 6. 15. Jang Seong Soon							*)
(*																	*)
(********************************************************************)

signature LANG_SCHEME =
  sig
    type absyn
    val abstract: (string * absyn) list -> (string * Compare.exp) list
    val parsel  : string list -> (string * absyn) list
  end

structure Lang_SCHEME: LANG_SCHEME =
  struct

open List
open Compare
type absyn = Scm.exp list

fun change l = VACANT_NODE(0,map change_exp l)
and change_exp p =
      case p of
	  Scm.Var s => CON 0
	| Scm.Lit l => NODE (1,[change_lit l])
	| Scm.Lst l => NODE (2,map change_exp l)
	| Scm.LstnPair(l,e) => NODE (3,change_exp e::map change_exp l)
	| Scm.Quote e => NODE(4,[change_exp e])
	| Scm.QQuote e => NODE(5,[change_exp e])
	| Scm.UQuote e => NODE(6,[change_exp e])
	| Scm.UQSplice e => NODE(7,[change_exp e])
	| Scm.Vector l => NODE(8,map change_exp l)
and change_lit p =
	  case p of
	  Scm.Int i => CON 1
	| Scm.Bool b => CON 2
	| Scm.Real r => CON 3
	| Scm.Char c => CON 4
	| Scm.String s => CON 5

fun abstract l = map (fn (n,e) => (n, change e)) l

(************** Parsing files *********************************)
(************** and triangular n(n-1)/2 times iteration *******)
fun parse_file filename =
  let val ic = open_in_bin filename
      val lexbuf = Lexing.from_channel ic handle
            x => close_in ic; raise x
      val ast = Scm_parser.goal Scm_lexer.token lexbuf
  in
    close_in ic;
    ast
  end

fun parsel [] = []
  | parsel(h::l) = (h,parse_file h)::parsel l
                   handle _ =>
                        Format.eprintf "%s is an invalid file\n" h;
						flush stderr;	// flush the buffer
                        parsel l    	// ignore parse error file

end
