(*
 Sungsoon Jang, The LET Project, KAIST

 Copyright(c) 2001 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*)

{
 open Scm_parser
 open String
 exception LexError
 fun create_hashtable n l = 
   let
	 val tbl = Hashtbl.create n
   in
	 List.iter (fn (k,d) => Hashtbl.add tbl k d) l;
	 tbl
   end
 val keyword_table =
 	create_hashtable 4 [
	  ("quote", QUOTE),
	  ("quasiquote",QQUOTE),
	  ("unquote",UQUOTE),
	  ("unquote-splicing",UQSPLICE) ]
 val comLevel = ref 0
} 

val blank = [' ' '\012' '\009' '\010' '\013']+
val digital = '-'?['0'-'9']
val number = digital+
val letter = ['a'-'z' 'A'-'Z']
val id_chars =
  [ '!' '@' '#' '$' '%' '^' '&' '*' '~' '`' '-' '_' '+' '=' 
    ':' ';' ''' '<' '>' ',' '.' '?' '/' '"' '\\']
val chars = letter | digital | id_chars | ['(' ')' '[' ']' '{' '}' '|']
val id 	   = (letter | digital | id_chars )+
val rnumber = digital+ '.' digital*
rule token = parse
	  blank		{ token lexbuf }
	| id		{ let val s = lowercase(Lexing.lexeme lexbuf)
				  in
				  	(Hashtbl.find keyword_table s) handle Not_found => ID s
				  end }
 	| "#t"		{ BOOL(true) }
	| "#f"		{ BOOL(false) }
	| number	{ INT (int_of_string(Lexing.lexeme lexbuf)) }
	| rnumber	{ FLOAT (float_of_string(Lexing.lexeme lexbuf)) }
	| "#\\" chars
				{ CHAR(sub (Lexing.lexeme lexbuf) 2 1) }
	| "#\\newline"	{ CHAR("\n") }
	| "#\\space"	{ CHAR(" ") }
	| '"' ([^'"' '\n' '\\'] | "\\\\" | "\\\"")* '"'
				{ let val s = Lexing.lexeme lexbuf
				  in
				  	STRING(sub s 1 (length s - 2))
				  end }
    | eof 		{ EOF }
    | "(" 		{ LP }
    | ")" 		{ RP }
	| "#("		{ SLP }
	| "'"		{ QUOTE }
	| "`"		{ QQUOTE }
	| ","		{ UQUOTE }
	| "."		{ DOT }
	| ",@"		{ UQSPLICE }
	| ";"[^'\n']* '\n'
				{ token lexbuf }
	| "#|"		{ comLevel:= 0; comment lexbuf; token lexbuf }
    | _ 		{ raise LexError}
and comment = parse
	  "#|"		{ comLevel := succ !comLevel; comment lexbuf }
	| "|#"		{ comLevel := pred !comLevel;
					if !comLevel > 0 then comment lexbuf}
	| eof		{ raise LexError}
	| _			{ comment lexbuf }
