signature C_parser = sig 

type token =
    EOF
  | COLON
  | SEMICOLON
  | LPAREN
  | RPAREN
  | LCURLY
  | RCURLY
  | LBRACE
  | RBRACE
  | DOT
  | COMMA
  | QUESTION
  | PERCENT
  | AMP
  | BAR
  | TILDE
  | DIVIDE
  | PLUS
  | MINUS
  | HAT
  | BANG
  | TIMES
  | INC
  | DEC
  | ARROW
  | ID of (string)
  | EQUALS
  | PLUSEQUALS
  | MINUSEQUALS
  | XOREQUALS
  | MODEQUALS
  | TIMESEQUALS
  | DIVEQUALS
  | OREQUALS
  | ANDEQUALS
  | LSHIFTEQUALS
  | RSHIFTEQUALS
  | LTE
  | GTE
  | LT
  | GT
  | EQ
  | NEQ
  | OR
  | AND
  | LSHIFT
  | RSHIFT
  | DECNUM of (int)
  | REALNUM of (string)
  | STRING of (string)
  | CCONST of (int)
  | EXTERN
  | AUTO
  | STATIC
  | REGISTER
  | CONST
  | VOLATILE
  | IF
  | THEN
  | ELSE
  | FOR
  | DO
  | SWITCH
  | CASE
  | DEFAULT
  | WHILE
  | RETURN
  | BREAK
  | CONTINUE
  | GOTO
  | CHAR
  | DOUBLE
  | ENUM
  | FLOAT
  | INT
  | LONG
  | SHORT
  | FRACTIONAL
  | SATURATE
  | STRUCT
  | UNION
  | UNSIGNED
  | SIGNED
  | VOID
  | SIZEOF
  | TYPEDEF
  | UNARY
  | ELIPSIS
  | TYPE_NAME of (string)
  | PARALLEL
  | SEQUENTIAL
  | COUNT
  | IN
  | REDUCE
  | WHERE
  | DEPENDENCY
  | ASSOCIATIVE
  | TO
  | KWAND

val goal :
  (Lexing.lexbuf  -> token) -> Lexing.lexbuf -> ParseTree.externalDecl list


end (* END of signature C_parser *)


structure C_parser = struct 

type token =
    EOF
  | COLON
  | SEMICOLON
  | LPAREN
  | RPAREN
  | LCURLY
  | RCURLY
  | LBRACE
  | RBRACE
  | DOT
  | COMMA
  | QUESTION
  | PERCENT
  | AMP
  | BAR
  | TILDE
  | DIVIDE
  | PLUS
  | MINUS
  | HAT
  | BANG
  | TIMES
  | INC
  | DEC
  | ARROW
  | ID of (string)
  | EQUALS
  | PLUSEQUALS
  | MINUSEQUALS
  | XOREQUALS
  | MODEQUALS
  | TIMESEQUALS
  | DIVEQUALS
  | OREQUALS
  | ANDEQUALS
  | LSHIFTEQUALS
  | RSHIFTEQUALS
  | LTE
  | GTE
  | LT
  | GT
  | EQ
  | NEQ
  | OR
  | AND
  | LSHIFT
  | RSHIFT
  | DECNUM of (int)
  | REALNUM of (string)
  | STRING of (string)
  | CCONST of (int)
  | EXTERN
  | AUTO
  | STATIC
  | REGISTER
  | CONST
  | VOLATILE
  | IF
  | THEN
  | ELSE
  | FOR
  | DO
  | SWITCH
  | CASE
  | DEFAULT
  | WHILE
  | RETURN
  | BREAK
  | CONTINUE
  | GOTO
  | CHAR
  | DOUBLE
  | ENUM
  | FLOAT
  | INT
  | LONG
  | SHORT
  | FRACTIONAL
  | SATURATE
  | STRUCT
  | UNION
  | UNSIGNED
  | SIGNED
  | VOID
  | SIZEOF
  | TYPEDEF
  | UNARY
  | ELIPSIS
  | TYPE_NAME of (string)
  | PARALLEL
  | SEQUENTIAL
  | COUNT
  | IN
  | REDUCE
  | WHERE
  | DEPENDENCY
  | ASSOCIATIVE
  | TO
  | KWAND

open Parsing
(* # 56 "c/c_parser.ny" *)
open ParseTree (* PortingHelp *)

val unknown_decltype = {storage=[],qualifiers=[],specifiers=[]}:decltype

(* this code duplicated in BuildAst in function processDeclarator *)
fun ctypeDecrToTypeName (typ as {qualifiers, specifiers},decr) =
      let fun mkTyp spc = {qualifiers=[], specifiers=[spc]}
   fun addQual q = {qualifiers=q::qualifiers, specifiers=specifiers}
       in case decr
     of VarDecr x => (typ,Some x)
      | PointerDecr x =>
  ctypeDecrToTypeName (mkTyp (Pointer typ),x)
      | ArrayDecr (x,sz) =>
  ctypeDecrToTypeName (mkTyp (Array (sz,typ)),x)
      | FuncDecr (x,lst) =>
  ctypeDecrToTypeName (mkTyp (Function{retType=typ,params=lst}),x)
      | QualDecr (q,decr) =>
  ctypeDecrToTypeName (addQual q, decr)
      | EmptyDecr => (typ, None)
      | EllipsesDecr => (mkTyp Ellipses, Some("**ellipses**"))
      | MARKdeclarator(decr) => ctypeDecrToTypeName(typ, decr)
      end

fun dclr2str dcl = 
  (case ctypeDecrToTypeName ({qualifiers=[],specifiers=[]}, dcl)
     of (_,Some s) => s
      | (_,None) => "")

fun combineDecltypes 
    ( {qualifiers=q1,storage=st1,specifiers=sp1}
      , {qualifiers=q2,storage=st2,specifiers=sp2}) =
  {qualifiers=q1@q2,storage=st1@st2,specifiers=sp1@sp2}  

fun applyPointer (PointerDecr x,rest) = PointerDecr (applyPointer (x,rest))
  | applyPointer (QualDecr (q,x),rest) = QualDecr (q, applyPointer (x,rest))
  | applyPointer (EmptyDecr, rest) = rest
  | applyPointer (_, rest) = rest  
    (* NCH/DBM[6/14/99]: this case can never occur  *)

fun decltypeAddStorage(st, {qualifiers,storage,specifiers}) =
 {qualifiers=qualifiers,storage=st::storage,specifiers=specifiers}

fun decltypeAddQualifiers(qs, {qualifiers,storage,specifiers}) =
 {qualifiers=qs@qualifiers,storage=storage,specifiers=specifiers} 

fun decltypeAddQualifier(q, {qualifiers,storage,specifiers}) =
 {qualifiers=q::qualifiers,storage=storage,specifiers=specifiers}

fun decltypeAddSpecifier(sp, {qualifiers,storage,specifiers}) =
 {qualifiers=qualifiers,storage=storage,specifiers=sp::specifiers}

val decltypeAddAll = combineDecltypes

fun loopQd (q::rst, acc) = loopQd(rst, QualDecr(q, acc))
  | loopQd (nil, acc) = acc

(*********** typedef checking : START *****************************)
(* DBM: major kludge, using TYPEDEF as storage class *)
fun insertDeclNames ({storage,...}: decltype, idl) =
    case storage
      of [TYPEDEF] =>  List.iter (fn (dcl,_) => TypeDefs.addTdef (dclr2str dcl)) idl
       | _ => List.iter (fn (dcl,_) => TypeDefs.addNoTdef (dclr2str dcl)) idl

fun insertFuncName dcl =
  let
    val name = dclr2str dcl
  in
    TypeDefs.addNoTdef name
  end

fun insertFuncParams (FuncDecr (_,params)) = 
     let
       fun getName (ct, dclr) = dclr2str dclr
       val names = List.map getName params
     in
       List.iter TypeDefs.addNoTdef names
     end
  | insertFuncParams (ArrayDecr(dcl,_)) = insertFuncParams dcl
  | insertFuncParams (PointerDecr dcl) = insertFuncParams dcl
  | insertFuncParams _ = ()  (* this is actually an error, but it will be 
                              * caught in BuildAst when processing 
			      *  a PT.FunctionDef *)

(*********** typedef checking : END *****************************)

(* datatype and functions for list-like data*)
type 'a seq = SEQ of 'a list
val emptySeq = SEQ nil
fun singletonSeq x = SEQ[x]
fun addToSeq(x, SEQ yl) = SEQ(x :: yl)  (* add to end of sequence! *)
  (* fun addListToEnd(xl, yl) = SEQ((List.rev xl) @ yl) *)
fun addOptToEnd(None, yl) = yl
  | addOptToEnd(Some x, SEQ yl) = SEQ(x :: yl)
fun seqToList(SEQ yl) = List.rev yl

(* Line 99, file c/c_parser.n *)
val yytransl_const = [|
    0 (* EOF *),
  257 (* COLON *),
  258 (* SEMICOLON *),
  259 (* LPAREN *),
  260 (* RPAREN *),
  261 (* LCURLY *),
  262 (* RCURLY *),
  263 (* LBRACE *),
  264 (* RBRACE *),
  265 (* DOT *),
  266 (* COMMA *),
  267 (* QUESTION *),
  268 (* PERCENT *),
  269 (* AMP *),
  270 (* BAR *),
  271 (* TILDE *),
  272 (* DIVIDE *),
  273 (* PLUS *),
  274 (* MINUS *),
  275 (* HAT *),
  276 (* BANG *),
  277 (* TIMES *),
  278 (* INC *),
  279 (* DEC *),
  280 (* ARROW *),
  282 (* EQUALS *),
  283 (* PLUSEQUALS *),
  284 (* MINUSEQUALS *),
  285 (* XOREQUALS *),
  286 (* MODEQUALS *),
  287 (* TIMESEQUALS *),
  288 (* DIVEQUALS *),
  289 (* OREQUALS *),
  290 (* ANDEQUALS *),
  291 (* LSHIFTEQUALS *),
  292 (* RSHIFTEQUALS *),
  293 (* LTE *),
  294 (* GTE *),
  295 (* LT *),
  296 (* GT *),
  297 (* EQ *),
  298 (* NEQ *),
  299 (* OR *),
  300 (* AND *),
  301 (* LSHIFT *),
  302 (* RSHIFT *),
  307 (* EXTERN *),
  308 (* AUTO *),
  309 (* STATIC *),
  310 (* REGISTER *),
  311 (* CONST *),
  312 (* VOLATILE *),
  313 (* IF *),
  314 (* THEN *),
  315 (* ELSE *),
  316 (* FOR *),
  317 (* DO *),
  318 (* SWITCH *),
  319 (* CASE *),
  320 (* DEFAULT *),
  321 (* WHILE *),
  322 (* RETURN *),
  323 (* BREAK *),
  324 (* CONTINUE *),
  325 (* GOTO *),
  326 (* CHAR *),
  327 (* DOUBLE *),
  328 (* ENUM *),
  329 (* FLOAT *),
  330 (* INT *),
  331 (* LONG *),
  332 (* SHORT *),
  333 (* FRACTIONAL *),
  334 (* SATURATE *),
  335 (* STRUCT *),
  336 (* UNION *),
  337 (* UNSIGNED *),
  338 (* SIGNED *),
  339 (* VOID *),
  340 (* SIZEOF *),
  341 (* TYPEDEF *),
  342 (* UNARY *),
  343 (* ELIPSIS *),
  345 (* PARALLEL *),
  346 (* SEQUENTIAL *),
  347 (* COUNT *),
  348 (* IN *),
  349 (* REDUCE *),
  350 (* WHERE *),
  351 (* DEPENDENCY *),
  352 (* ASSOCIATIVE *),
  353 (* TO *),
  354 (* KWAND *),
    0|]

val yytransl_block = [|
  281 (* ID *),
  303 (* DECNUM *),
  304 (* REALNUM *),
  305 (* STRING *),
  306 (* CCONST *),
  344 (* TYPE_NAME *),
    0|]

val yylhs = "\255\255\
\001\000\002\000\002\000\003\000\003\000\003\000\006\000\006\000\
\006\000\006\000\006\000\006\000\006\000\006\000\006\000\006\000\
\006\000\006\000\006\000\006\000\006\000\006\000\004\000\010\000\
\010\000\010\000\015\000\015\000\016\000\016\000\009\000\009\000\
\020\000\020\000\020\000\020\000\020\000\020\000\021\000\021\000\
\021\000\021\000\021\000\021\000\021\000\021\000\021\000\021\000\
\021\000\021\000\021\000\021\000\021\000\021\000\021\000\021\000\
\021\000\021\000\021\000\021\000\021\000\021\000\021\000\021\000\
\021\000\021\000\021\000\021\000\021\000\021\000\021\000\021\000\
\021\000\021\000\021\000\021\000\021\000\021\000\021\000\021\000\
\021\000\021\000\021\000\021\000\021\000\021\000\021\000\021\000\
\021\000\024\000\024\000\008\000\008\000\007\000\007\000\023\000\
\023\000\022\000\022\000\011\000\011\000\028\000\028\000\028\000\
\029\000\029\000\029\000\029\000\013\000\013\000\013\000\013\000\
\025\000\025\000\035\000\035\000\036\000\036\000\027\000\027\000\
\027\000\027\000\031\000\031\000\031\000\031\000\031\000\031\000\
\031\000\031\000\031\000\033\000\033\000\033\000\033\000\033\000\
\037\000\037\000\038\000\038\000\039\000\039\000\040\000\040\000\
\041\000\041\000\042\000\042\000\042\000\043\000\043\000\043\000\
\032\000\032\000\034\000\034\000\034\000\034\000\034\000\046\000\
\046\000\048\000\048\000\030\000\030\000\030\000\030\000\030\000\
\047\000\047\000\012\000\012\000\049\000\049\000\014\000\014\000\
\051\000\051\000\044\000\044\000\052\000\052\000\045\000\045\000\
\056\000\056\000\054\000\054\000\054\000\054\000\053\000\053\000\
\053\000\053\000\053\000\053\000\053\000\055\000\055\000\055\000\
\055\000\055\000\055\000\055\000\057\000\057\000\057\000\057\000\
\057\000\057\000\050\000\050\000\060\000\060\000\018\000\018\000\
\059\000\059\000\061\000\061\000\061\000\005\000\005\000\026\000\
\026\000\026\000\062\000\062\000\062\000\062\000\062\000\062\000\
\062\000\062\000\062\000\058\000\058\000\063\000\063\000\064\000\
\064\000\064\000\064\000\064\000\064\000\017\000\019\000\000\000"

val yylen = "\002\000\
\001\000\000\000\002\000\001\000\001\000\001\000\009\000\005\000\
\005\000\007\000\002\000\002\000\003\000\003\000\001\000\003\000\
\003\000\004\000\005\000\007\000\002\000\001\000\002\000\001\000\
\002\000\002\000\001\000\000\000\001\000\002\000\006\000\003\000\
\001\000\001\000\001\000\001\000\001\000\001\000\005\000\003\000\
\003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
\003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
\003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
\003\000\003\000\003\000\003\000\002\000\002\000\002\000\002\000\
\002\000\002\000\004\000\004\000\004\000\003\000\004\000\003\000\
\003\000\003\000\003\000\003\000\001\000\001\000\001\000\001\000\
\001\000\001\000\002\000\001\000\003\000\000\000\001\000\001\000\
\003\000\001\000\002\000\002\000\003\000\000\000\002\000\002\000\
\001\000\001\000\001\000\001\000\001\000\002\000\001\000\002\000\
\002\000\003\000\000\000\002\000\001\000\002\000\001\000\001\000\
\001\000\001\000\001\000\001\000\001\000\001\000\001\000\001\000\
\001\000\001\000\001\000\004\000\005\000\005\000\002\000\002\000\
\001\000\001\000\001\000\002\000\003\000\003\000\001\000\003\000\
\001\000\003\000\001\000\002\000\003\000\001\000\002\000\003\000\
\001\000\001\000\005\000\006\000\006\000\002\000\002\000\003\000\
\001\000\001\000\003\000\001\000\001\000\001\000\001\000\001\000\
\000\000\001\000\001\000\003\000\001\000\003\000\001\000\003\000\
\001\000\003\000\001\000\001\000\001\000\002\000\001\000\002\000\
\001\000\002\000\001\000\002\000\002\000\003\000\001\000\003\000\
\003\000\004\000\003\000\004\000\004\000\001\000\003\000\003\000\
\004\000\003\000\004\000\004\000\001\000\003\000\004\000\003\000\
\004\000\004\000\001\000\004\000\001\000\003\000\001\000\002\000\
\001\000\003\000\001\000\002\000\002\000\002\000\003\000\001\000\
\001\000\002\000\003\000\002\000\003\000\003\000\004\000\002\000\
\003\000\003\000\004\000\001\000\003\000\001\000\003\000\002\000\
\002\000\001\000\002\000\002\000\002\000\000\000\000\000\002\000"

val yydefred = "\000\000\
\002\000\000\000\248\000\000\000\005\000\000\000\000\000\198\000\
\164\000\166\000\165\000\167\000\153\000\154\000\124\000\129\000\
\000\000\128\000\126\000\127\000\125\000\137\000\138\000\131\000\
\130\000\123\000\168\000\122\000\003\000\004\000\006\000\000\000\
\000\000\000\000\102\000\109\000\119\000\111\000\120\000\121\000\
\000\000\219\000\000\000\000\000\000\000\000\000\000\000\000\000\
\189\000\000\000\000\000\000\000\023\000\000\000\191\000\000\000\
\000\000\180\000\171\000\179\000\000\000\000\000\000\000\102\000\
\110\000\112\000\000\000\175\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\215\000\222\000\000\000\000\000\
\000\000\199\000\118\000\190\000\000\000\000\000\161\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\103\000\104\000\105\000\106\000\107\000\108\000\
\000\000\115\000\000\000\000\000\139\000\000\000\000\000\202\000\
\217\000\000\000\000\000\000\000\000\000\000\000\238\000\000\000\
\200\000\033\000\037\000\035\000\036\000\038\000\034\000\000\000\
\000\000\088\000\085\000\086\000\000\000\087\000\000\000\000\000\
\000\000\089\000\022\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\029\000\
\000\000\015\000\000\000\000\000\000\000\000\000\000\000\000\000\
\216\000\223\000\000\000\000\000\000\000\000\000\000\000\192\000\
\172\000\000\000\000\000\174\000\195\000\000\000\000\000\193\000\
\000\000\176\000\178\000\000\000\000\000\143\000\000\000\000\000\
\000\000\115\000\000\000\145\000\000\000\132\000\140\000\000\000\
\000\000\000\000\000\000\205\000\243\000\240\000\000\000\241\000\
\000\000\000\000\245\000\244\000\000\000\203\000\204\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\091\000\
\000\000\000\000\000\000\000\000\000\000\201\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\069\000\070\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\011\000\012\000\000\000\021\000\000\000\032\000\
\030\000\000\000\000\000\160\000\155\000\000\000\000\000\213\000\
\000\000\196\000\197\000\194\000\000\000\141\000\000\000\000\000\
\116\000\000\000\000\000\142\000\000\000\000\000\133\000\134\000\
\232\000\000\000\000\000\228\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\218\000\237\000\239\000\084\000\000\000\
\000\000\099\000\000\000\000\000\078\000\000\000\000\000\000\000\
\080\000\082\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\081\000\083\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\016\000\000\000\000\000\000\000\000\000\000\000\
\017\000\000\000\013\000\014\000\000\000\247\000\156\000\157\000\
\000\000\000\000\144\000\000\000\146\000\000\000\227\000\233\000\
\229\000\208\000\000\000\000\000\206\000\000\000\234\000\000\000\
\230\000\000\000\000\000\000\000\079\000\000\000\077\000\000\000\
\000\000\000\000\000\000\000\000\018\000\000\000\000\000\214\000\
\212\000\209\000\210\000\207\000\235\000\231\000\000\000\000\000\
\000\000\000\000\000\000\009\000\008\000\031\000\000\000\000\000\
\000\000\020\000\000\000\010\000\000\000\007\000"

val yydgoto = "\002\000\
\003\000\004\000\029\000\077\000\031\000\152\000\009\001\153\000\
\154\000\032\000\114\000\056\000\115\000\063\000\155\000\156\000\
\157\000\081\000\143\001\136\000\158\000\211\000\063\001\138\000\
\105\000\042\001\035\000\069\000\099\000\036\000\037\000\038\000\
\039\000\040\000\184\000\107\000\041\000\108\000\109\000\181\000\
\187\000\182\000\188\000\159\000\058\000\086\000\165\000\087\000\
\059\000\172\000\068\000\060\000\061\000\062\000\044\000\200\000\
\201\000\043\001\117\000\025\001\045\000\202\000\118\000\119\000"

val yysindex = "\041\000\
\000\000\000\000\000\000\168\002\000\000\189\255\031\255\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\021\255\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\066\255\
\081\000\212\003\000\000\000\000\000\000\000\000\000\000\000\000\
\042\255\000\000\040\255\200\255\154\004\067\255\001\255\077\255\
\000\000\093\255\118\255\127\255\000\000\081\000\000\000\129\255\
\158\255\000\000\000\000\000\000\009\000\033\255\206\255\000\000\
\000\000\000\000\244\255\000\000\113\013\135\013\030\000\048\000\
\200\255\251\003\021\001\050\005\000\000\000\000\081\000\212\003\
\154\004\000\000\000\000\000\000\014\000\050\000\000\000\093\255\
\093\255\059\000\081\000\171\005\036\004\182\005\009\000\081\000\
\113\013\171\005\000\000\000\000\000\000\000\000\000\000\000\000\
\071\255\000\000\011\002\124\002\000\000\135\013\135\013\000\000\
\000\000\070\000\131\003\061\000\240\255\073\000\000\000\049\003\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\062\006\
\062\006\000\000\000\000\000\000\043\000\000\000\100\006\062\006\
\153\009\000\000\000\000\093\000\141\000\144\000\050\005\152\000\
\062\006\121\000\153\000\062\006\155\000\172\000\162\000\000\000\
\081\255\000\000\140\000\050\005\061\013\117\010\158\255\244\255\
\000\000\000\000\062\006\093\255\169\000\050\000\050\000\000\000\
\000\000\171\005\117\010\000\000\000\000\180\000\029\000\000\000\
\197\009\000\000\000\000\062\006\098\255\000\000\175\000\224\001\
\062\006\000\000\134\255\000\000\196\000\000\000\000\000\068\005\
\090\005\088\003\220\005\000\000\000\000\000\000\192\255\000\000\
\055\000\129\000\000\000\000\000\215\255\000\000\000\000\181\000\
\006\005\099\000\201\000\205\255\177\005\128\000\128\000\000\000\
\049\003\128\000\128\000\002\006\062\006\000\000\013\255\062\006\
\062\006\062\006\062\006\062\006\062\006\062\006\062\006\062\006\
\000\000\000\000\113\255\062\006\062\006\062\006\062\006\062\006\
\062\006\062\006\062\006\062\006\062\006\062\006\062\006\062\006\
\062\006\062\006\062\006\062\006\062\006\062\006\062\006\062\006\
\050\005\062\006\062\006\190\000\062\006\049\255\050\005\062\006\
\213\000\215\000\000\000\000\000\244\000\000\000\062\006\000\000\
\000\000\093\002\117\010\000\000\000\000\002\001\003\001\000\000\
\217\000\000\000\000\000\000\000\117\010\000\000\071\255\062\006\
\000\000\117\010\224\001\000\000\071\255\062\006\000\000\000\000\
\000\000\011\001\012\001\000\000\241\009\055\000\129\000\075\004\
\013\006\116\004\051\006\000\000\000\000\000\000\000\000\062\006\
\171\003\000\000\221\000\024\001\000\000\117\010\100\000\068\000\
\000\000\000\000\024\007\128\000\227\012\125\012\128\000\228\001\
\228\001\165\012\128\000\000\000\000\000\117\010\117\010\117\010\
\117\010\117\010\117\010\117\010\117\010\117\010\117\010\117\010\
\051\013\051\013\051\013\051\013\011\013\011\013\213\011\041\012\
\217\013\217\013\000\000\101\000\028\001\029\001\119\000\050\005\
\000\000\139\000\000\000\000\000\117\010\000\000\000\000\000\000\
\171\005\027\001\000\000\117\010\000\000\117\010\000\000\000\000\
\000\000\000\000\033\001\150\000\000\000\029\010\000\000\036\001\
\000\000\073\010\128\000\104\006\000\000\062\006\000\000\062\006\
\050\005\062\006\062\006\050\005\000\000\050\005\039\001\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\117\010\081\011\
\245\000\045\001\161\000\000\000\000\000\000\000\050\005\062\006\
\049\001\000\000\050\001\000\000\050\005\000\000"

val yyrindex = "\000\000\
\000\000\000\000\000\000\055\002\000\000\000\000\059\255\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\054\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\048\001\000\000\000\000\238\001\248\255\
\000\000\000\000\045\000\138\000\000\000\000\000\000\000\055\001\
\206\002\000\000\000\000\000\000\106\001\000\000\057\001\000\000\
\000\000\000\000\005\003\000\000\239\255\000\000\178\000\010\001\
\144\001\000\000\000\000\194\004\000\000\000\000\054\001\000\000\
\000\000\000\000\000\000\000\000\196\255\061\001\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\202\001\000\000\
\065\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\163\000\000\000\000\000\000\000\056\001\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\188\006\000\000\000\000\000\000\
\000\000\000\000\000\000\109\009\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\070\001\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\067\001\000\000\035\003\187\255\213\255\
\000\000\000\000\000\000\069\001\000\000\061\001\061\001\000\000\
\000\000\000\000\024\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\225\255\072\255\
\000\000\000\000\000\000\000\000\035\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\168\000\000\000\
\173\000\197\000\000\000\000\000\168\000\000\000\000\000\000\000\
\000\000\000\000\000\000\072\001\000\000\070\007\116\007\000\000\
\000\000\162\007\208\007\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\070\001\000\000\000\000\000\000\000\000\000\000\
\000\000\007\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\071\001\016\001\000\000\000\000\000\000\000\000\000\000\
\061\001\000\000\000\000\000\000\054\000\000\000\000\000\000\000\
\000\000\072\000\112\255\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\212\000\233\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\074\001\000\000\000\000\235\000\000\000\000\000\
\000\000\000\000\000\000\254\007\003\012\130\006\044\008\182\008\
\228\008\087\012\090\008\000\000\000\000\160\000\194\000\234\000\
\017\001\133\001\163\001\065\002\099\002\118\002\182\002\224\002\
\163\010\207\010\251\010\039\011\127\011\171\011\075\013\189\012\
\018\009\064\009\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\040\003\000\000\000\000\000\000\
\069\001\000\000\000\000\079\000\000\000\083\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\136\008\234\006\000\000\000\000\000\000\000\000\
\000\000\070\001\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\237\000\146\013\
\238\004\000\000\000\000\000\000\000\000\000\000\000\000\075\001\
\000\000\000\000\000\000\000\000\000\000\000\000"

val yygindex = "\000\000\
\000\000\000\000\000\000\001\000\000\000\141\255\045\255\138\255\
\004\000\000\000\025\000\000\000\033\000\000\000\076\001\000\000\
\000\000\186\001\000\000\000\000\181\255\127\001\000\000\219\001\
\146\255\175\255\177\000\033\002\094\255\159\000\194\255\151\000\
\198\255\199\255\176\001\008\000\000\000\143\000\250\255\000\000\
\000\000\225\000\000\000\232\255\010\000\019\001\110\255\201\001\
\254\255\166\255\000\000\058\002\051\002\253\255\245\255\000\000\
\171\001\188\255\176\255\000\000\000\000\075\255\000\000\162\001"

val yytablesize = 3825
val yytable = "\137\000\
\043\000\210\000\043\000\049\000\030\000\116\000\101\000\179\000\
\057\000\212\000\103\000\104\000\175\000\042\000\048\000\046\000\
\171\000\047\001\177\000\022\001\023\001\033\001\171\000\047\001\
\174\000\050\000\006\001\004\001\033\000\010\001\043\000\073\000\
\197\000\203\000\101\000\054\000\034\000\065\001\103\000\104\000\
\017\001\001\000\006\000\067\000\084\000\051\000\070\000\101\001\
\078\000\104\001\073\000\007\000\214\000\215\000\083\000\013\000\
\014\000\008\000\015\001\218\000\219\000\187\000\187\000\046\000\
\008\000\187\000\071\000\053\000\187\000\079\000\082\000\180\000\
\113\000\054\000\113\000\113\000\043\000\080\000\113\000\024\001\
\183\000\161\000\014\001\187\000\162\000\013\000\014\000\019\001\
\169\000\160\000\015\001\007\000\113\000\178\000\171\000\008\000\
\113\000\007\000\210\000\030\001\066\001\191\000\064\001\043\000\
\029\001\079\000\212\000\031\001\052\000\034\001\199\000\205\000\
\114\000\080\000\114\000\114\000\189\000\085\000\114\000\045\001\
\055\000\101\000\088\000\198\000\204\000\103\000\104\000\213\000\
\033\001\072\000\058\001\089\000\114\000\047\001\114\001\036\001\
\114\000\076\001\091\000\100\001\010\001\099\001\103\001\037\001\
\062\001\106\001\187\000\105\001\067\001\068\001\069\001\070\001\
\071\001\072\001\073\001\074\001\075\001\047\000\055\000\113\000\
\078\001\079\001\080\001\081\001\082\001\083\001\084\001\085\001\
\086\001\087\001\088\001\089\001\090\001\091\001\092\001\093\001\
\094\001\095\001\096\001\097\001\098\001\079\000\154\001\092\000\
\066\000\191\000\191\000\073\000\173\000\080\000\205\000\006\000\
\065\000\073\000\194\000\109\001\173\000\047\000\195\000\114\000\
\077\001\162\000\074\000\046\000\163\001\162\000\075\000\057\001\
\059\001\007\000\064\000\195\000\116\001\008\000\177\000\096\000\
\008\000\194\000\118\001\102\000\047\000\195\000\177\000\124\001\
\213\000\007\000\147\000\100\000\101\000\126\001\066\000\130\001\
\103\000\104\000\147\000\123\001\131\001\128\001\065\000\008\000\
\100\000\100\000\100\000\207\000\141\001\100\000\106\000\102\000\
\100\000\208\000\188\000\188\000\192\000\193\000\188\000\100\000\
\064\000\188\000\047\000\100\000\047\000\047\000\183\000\100\000\
\095\000\066\000\095\000\093\000\183\000\098\000\047\000\094\000\
\188\000\065\000\161\000\010\001\155\001\153\001\144\001\196\000\
\156\001\211\000\157\001\186\000\106\000\211\000\106\000\106\000\
\027\001\211\000\110\000\064\000\150\000\171\000\208\000\163\000\
\106\000\010\001\079\000\162\001\150\000\158\000\158\000\158\000\
\158\000\166\001\080\000\158\000\111\000\059\001\158\000\148\000\
\131\001\048\001\151\001\164\000\152\001\049\001\168\000\148\000\
\206\000\158\000\101\000\101\000\101\000\158\000\100\000\101\000\
\194\000\151\000\101\000\135\001\195\000\015\001\102\000\188\000\
\149\000\151\000\209\000\054\000\152\000\101\000\047\000\047\000\
\149\000\101\000\007\000\133\000\152\000\001\001\008\000\158\000\
\158\000\158\000\158\000\158\000\158\000\007\000\055\001\133\001\
\137\001\008\000\166\000\167\000\015\001\134\001\015\001\047\000\
\106\000\106\000\158\000\158\000\158\000\158\000\158\000\158\000\
\158\000\007\001\140\001\158\000\158\000\158\000\158\000\158\000\
\015\001\158\000\220\000\050\001\158\000\186\000\221\000\051\001\
\223\000\106\000\159\000\159\000\159\000\159\000\142\001\002\001\
\159\000\016\001\003\001\159\000\015\001\233\000\234\000\235\000\
\101\000\147\001\005\001\008\001\011\001\196\000\159\000\208\000\
\050\000\050\000\159\000\050\000\161\001\050\000\242\000\050\000\
\055\000\050\000\015\001\224\000\242\000\012\001\021\001\032\001\
\185\000\224\000\135\000\135\000\135\000\135\000\185\000\026\001\
\135\000\102\000\013\001\135\000\159\000\159\000\159\000\159\000\
\159\000\159\000\040\000\040\000\038\001\040\000\135\000\040\000\
\225\000\040\000\135\000\040\000\056\001\052\001\225\000\159\000\
\159\000\159\000\159\000\159\000\159\000\159\000\107\001\186\000\
\159\000\159\000\159\000\159\000\159\000\186\000\159\000\057\001\
\015\001\159\000\113\001\195\000\135\000\135\000\135\000\135\000\
\135\000\135\000\041\000\041\000\226\000\041\000\096\000\041\000\
\097\000\041\000\226\000\041\000\096\000\108\001\097\000\135\000\
\135\000\135\000\135\000\135\000\135\000\135\000\102\001\115\001\
\135\000\135\000\135\000\135\000\135\000\117\001\135\000\111\001\
\112\001\135\000\136\000\136\000\136\000\136\000\119\001\120\001\
\136\000\045\000\045\000\136\000\045\000\163\000\045\000\120\000\
\045\000\163\000\045\000\132\001\121\000\138\001\136\000\139\001\
\145\001\122\000\136\000\123\000\146\001\124\000\125\000\149\001\
\126\000\127\000\128\000\129\000\158\001\130\000\160\001\159\001\
\183\000\183\000\164\001\183\000\183\000\165\001\001\000\024\000\
\025\000\183\000\026\000\236\000\136\000\136\000\136\000\136\000\
\136\000\136\000\169\000\131\000\132\000\133\000\134\000\094\000\
\027\000\183\000\170\000\098\000\028\000\224\000\094\000\136\000\
\136\000\136\000\136\000\136\000\136\000\136\000\018\001\060\001\
\136\000\136\000\136\000\136\000\136\000\110\001\136\000\216\000\
\097\000\136\000\183\000\183\000\183\000\183\000\183\000\183\000\
\135\000\035\001\181\000\181\000\020\001\181\000\181\000\090\000\
\095\000\046\001\054\001\181\000\000\000\183\000\183\000\183\000\
\183\000\183\000\183\000\183\000\000\000\000\000\183\000\183\000\
\183\000\183\000\183\000\181\000\183\000\044\000\044\000\183\000\
\044\000\000\000\044\000\000\000\044\000\000\000\044\000\000\000\
\184\000\184\000\000\000\184\000\184\000\000\000\000\000\000\000\
\000\000\184\000\000\000\000\000\181\000\181\000\181\000\181\000\
\181\000\181\000\000\000\042\000\042\000\000\000\042\000\000\000\
\042\000\184\000\042\000\000\000\042\000\000\000\000\000\181\000\
\181\000\181\000\181\000\181\000\181\000\181\000\000\000\000\000\
\181\000\181\000\181\000\181\000\181\000\000\000\181\000\000\000\
\000\000\181\000\184\000\184\000\184\000\184\000\184\000\184\000\
\000\000\000\000\182\000\182\000\000\000\182\000\182\000\000\000\
\000\000\000\000\000\000\182\000\000\000\184\000\184\000\184\000\
\184\000\184\000\184\000\184\000\000\000\000\000\184\000\184\000\
\184\000\184\000\184\000\182\000\184\000\000\000\220\000\184\000\
\000\000\000\000\221\000\000\000\223\000\000\000\117\000\225\000\
\117\000\117\000\000\000\228\000\117\000\000\000\000\000\117\000\
\232\000\233\000\234\000\235\000\182\000\182\000\182\000\182\000\
\182\000\182\000\117\000\000\000\000\000\000\000\117\000\000\000\
\000\000\000\000\000\000\185\000\000\000\006\000\000\000\182\000\
\182\000\182\000\182\000\182\000\182\000\182\000\013\000\014\000\
\182\000\182\000\182\000\182\000\182\000\000\000\182\000\007\000\
\000\000\182\000\000\000\008\000\000\000\015\000\016\000\017\000\
\018\000\019\000\020\000\021\000\000\000\000\000\022\000\023\000\
\024\000\025\000\026\000\117\000\117\000\117\000\117\000\117\000\
\117\000\117\000\000\000\000\000\117\000\117\000\117\000\117\000\
\117\000\043\000\043\000\000\000\043\000\117\000\043\000\000\000\
\043\000\000\000\043\000\000\000\000\000\000\000\000\000\000\000\
\015\000\016\000\017\000\018\000\019\000\020\000\021\000\000\000\
\000\000\022\000\023\000\024\000\025\000\026\000\139\000\120\000\
\000\000\076\000\028\000\046\000\046\000\000\000\046\000\000\000\
\046\000\122\000\046\000\123\000\046\000\124\000\125\000\000\000\
\126\000\127\000\128\000\129\000\000\000\140\000\047\000\047\000\
\000\000\047\000\000\000\047\000\000\000\047\000\000\000\047\000\
\000\000\190\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\131\000\132\000\133\000\134\000\009\000\
\010\000\011\000\012\000\013\000\014\000\141\000\000\000\000\000\
\142\000\143\000\144\000\145\000\146\000\147\000\148\000\149\000\
\150\000\151\000\015\000\016\000\017\000\018\000\019\000\020\000\
\021\000\005\000\006\000\022\000\023\000\024\000\025\000\026\000\
\135\000\027\000\013\000\014\000\028\000\000\000\048\000\048\000\
\000\000\048\000\000\000\048\000\007\000\048\000\000\000\048\000\
\008\000\015\000\016\000\017\000\018\000\019\000\020\000\021\000\
\000\000\000\000\022\000\023\000\024\000\025\000\026\000\173\000\
\000\000\000\000\220\000\028\000\000\000\000\000\000\000\173\000\
\000\000\000\000\009\000\010\000\011\000\012\000\013\000\014\000\
\049\000\049\000\000\000\049\000\000\000\049\000\000\000\049\000\
\000\000\049\000\000\000\000\000\000\000\015\000\016\000\017\000\
\018\000\019\000\020\000\021\000\000\000\000\000\022\000\023\000\
\024\000\025\000\026\000\000\000\027\000\000\000\000\000\028\000\
\220\000\220\000\220\000\220\000\220\000\220\000\177\000\000\000\
\000\000\221\000\000\000\000\000\000\000\000\000\177\000\000\000\
\000\000\000\000\000\000\220\000\220\000\220\000\220\000\220\000\
\220\000\220\000\000\000\000\000\220\000\220\000\220\000\220\000\
\220\000\000\000\220\000\092\000\092\000\220\000\092\000\000\000\
\093\000\093\000\092\000\093\000\092\000\000\000\000\000\093\000\
\000\000\093\000\000\000\120\000\000\000\000\000\000\000\221\000\
\221\000\221\000\221\000\221\000\221\000\122\000\000\000\123\000\
\000\000\124\000\125\000\000\000\126\000\127\000\128\000\129\000\
\000\000\130\000\221\000\221\000\221\000\221\000\221\000\221\000\
\221\000\000\000\000\000\221\000\221\000\221\000\221\000\221\000\
\000\000\221\000\194\000\041\001\221\000\000\000\195\000\131\000\
\132\000\133\000\134\000\000\000\000\000\000\000\000\000\013\000\
\014\000\000\000\000\000\000\000\007\000\000\000\000\000\000\000\
\008\000\000\000\000\000\000\000\000\000\000\000\015\000\016\000\
\017\000\018\000\019\000\020\000\021\000\000\000\000\000\022\000\
\023\000\024\000\025\000\026\000\135\000\194\000\000\000\000\000\
\028\000\195\000\009\000\010\000\011\000\012\000\013\000\014\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\007\000\
\000\000\000\000\000\000\008\000\000\000\015\000\016\000\017\000\
\018\000\019\000\020\000\021\000\000\000\000\000\022\000\023\000\
\024\000\025\000\026\000\000\000\027\000\057\001\041\001\028\000\
\000\000\195\000\000\000\000\000\000\000\009\000\010\000\011\000\
\012\000\013\000\014\000\000\000\000\000\000\000\000\000\007\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\015\000\016\000\017\000\018\000\019\000\020\000\021\000\000\000\
\000\000\022\000\023\000\024\000\025\000\026\000\006\000\027\000\
\000\000\000\000\028\000\000\000\000\000\009\000\010\000\011\000\
\012\000\013\000\014\000\000\000\000\000\000\000\000\000\000\000\
\007\000\000\000\000\000\000\000\008\000\000\000\000\000\000\000\
\015\000\016\000\017\000\018\000\019\000\020\000\021\000\000\000\
\000\000\022\000\023\000\024\000\025\000\026\000\112\000\027\000\
\000\000\000\000\028\000\000\000\000\000\000\000\009\000\010\000\
\011\000\012\000\013\000\014\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\113\000\000\000\000\000\000\000\000\000\
\000\000\015\000\016\000\017\000\018\000\019\000\020\000\021\000\
\000\000\000\000\022\000\023\000\024\000\025\000\026\000\173\000\
\027\000\000\000\000\000\028\000\000\000\009\000\010\000\011\000\
\012\000\013\000\014\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\113\000\000\000\000\000\000\000\
\015\000\016\000\017\000\018\000\019\000\020\000\021\000\000\000\
\000\000\022\000\023\000\024\000\025\000\026\000\122\001\027\000\
\000\000\000\000\028\000\000\000\000\000\000\000\009\000\010\000\
\011\000\012\000\013\000\014\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\113\000\000\000\000\000\000\000\000\000\
\000\000\015\000\016\000\017\000\018\000\019\000\020\000\021\000\
\000\000\000\000\022\000\023\000\024\000\025\000\026\000\127\001\
\027\000\000\000\000\000\028\000\000\000\009\000\010\000\011\000\
\012\000\013\000\014\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\015\000\016\000\017\000\018\000\019\000\020\000\021\000\000\000\
\000\000\022\000\023\000\024\000\025\000\026\000\076\000\027\000\
\000\000\000\000\028\000\000\000\000\000\000\000\009\000\010\000\
\011\000\012\000\013\000\014\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\015\000\016\000\017\000\018\000\019\000\020\000\021\000\
\000\000\000\000\022\000\023\000\024\000\025\000\026\000\028\000\
\027\000\000\000\000\000\028\000\009\000\010\000\011\000\012\000\
\013\000\014\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\015\000\
\016\000\017\000\018\000\019\000\020\000\021\000\000\000\000\000\
\022\000\023\000\024\000\025\000\026\000\000\000\027\000\019\000\
\019\000\028\000\019\000\019\000\246\000\246\000\246\000\246\000\
\246\000\246\000\019\000\000\000\019\000\000\000\019\000\019\000\
\000\000\019\000\019\000\019\000\019\000\000\000\019\000\246\000\
\246\000\246\000\246\000\246\000\246\000\246\000\000\000\000\000\
\246\000\246\000\246\000\246\000\246\000\000\000\246\000\000\000\
\000\000\246\000\000\000\000\000\019\000\019\000\019\000\019\000\
\000\000\000\000\000\000\000\000\000\000\000\000\019\000\000\000\
\000\000\019\000\019\000\019\000\019\000\019\000\019\000\019\000\
\019\000\019\000\019\000\139\000\120\000\000\000\076\000\000\000\
\009\000\010\000\011\000\012\000\013\000\014\000\122\000\000\000\
\123\000\019\000\124\000\125\000\000\000\126\000\127\000\128\000\
\129\000\039\001\140\000\015\000\016\000\017\000\018\000\019\000\
\020\000\021\000\000\000\000\000\022\000\023\000\024\000\025\000\
\026\000\000\000\027\000\000\000\053\001\028\000\000\000\040\001\
\131\000\132\000\133\000\134\000\000\000\000\000\000\000\000\000\
\000\000\000\000\141\000\000\000\000\000\142\000\143\000\144\000\
\145\000\146\000\147\000\148\000\149\000\150\000\151\000\000\000\
\000\000\000\000\013\000\014\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\135\000\000\000\000\000\
\000\000\015\000\016\000\017\000\018\000\019\000\020\000\021\000\
\013\000\014\000\022\000\023\000\024\000\025\000\026\000\000\000\
\000\000\000\000\000\000\028\000\000\000\000\000\000\000\015\000\
\016\000\017\000\018\000\019\000\020\000\021\000\000\000\000\000\
\022\000\023\000\024\000\025\000\026\000\120\000\000\000\170\000\
\000\000\028\000\000\000\000\000\000\000\000\000\000\000\122\000\
\120\000\123\000\000\000\124\000\125\000\176\000\126\000\127\000\
\128\000\129\000\122\000\130\000\123\000\000\000\124\000\125\000\
\000\000\126\000\127\000\128\000\129\000\000\000\130\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\131\000\132\000\133\000\134\000\000\000\120\000\000\000\
\000\000\000\000\000\000\044\001\131\000\132\000\133\000\134\000\
\122\000\000\000\123\000\000\000\124\000\125\000\000\000\126\000\
\127\000\128\000\129\000\000\000\130\000\000\000\015\000\016\000\
\017\000\018\000\019\000\020\000\021\000\000\000\135\000\022\000\
\023\000\024\000\025\000\026\000\120\000\061\001\000\000\000\000\
\028\000\135\000\131\000\132\000\133\000\134\000\122\000\120\000\
\123\000\000\000\124\000\125\000\125\001\126\000\127\000\128\000\
\129\000\122\000\130\000\123\000\000\000\124\000\125\000\000\000\
\126\000\127\000\128\000\129\000\000\000\130\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\135\000\
\131\000\132\000\133\000\134\000\000\000\120\000\000\000\000\000\
\000\000\000\000\129\001\131\000\132\000\133\000\134\000\122\000\
\120\000\123\000\000\000\124\000\125\000\000\000\126\000\127\000\
\128\000\129\000\122\000\130\000\123\000\000\000\124\000\125\000\
\000\000\126\000\127\000\128\000\129\000\135\000\130\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\135\000\131\000\132\000\133\000\134\000\000\000\217\000\000\000\
\000\000\000\000\120\000\000\000\131\000\132\000\133\000\134\000\
\122\000\000\000\123\000\000\000\124\000\125\000\123\000\126\000\
\127\000\128\000\129\000\126\000\130\000\128\000\129\000\000\000\
\130\000\000\000\053\000\053\000\000\000\053\000\135\000\053\000\
\000\000\053\000\000\000\053\000\053\000\000\000\000\000\053\000\
\000\000\135\000\131\000\132\000\133\000\134\000\131\000\132\000\
\133\000\134\000\000\000\053\000\053\000\053\000\053\000\053\000\
\053\000\053\000\053\000\053\000\053\000\053\000\000\000\000\000\
\000\000\000\000\000\000\000\000\053\000\053\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\135\000\
\000\000\000\000\000\000\135\000\090\000\090\000\090\000\090\000\
\000\000\090\000\090\000\090\000\090\000\090\000\090\000\090\000\
\090\000\090\000\000\000\090\000\090\000\090\000\090\000\000\000\
\090\000\090\000\090\000\090\000\000\000\090\000\090\000\090\000\
\090\000\090\000\090\000\090\000\090\000\090\000\090\000\090\000\
\090\000\090\000\090\000\090\000\090\000\090\000\090\000\090\000\
\090\000\090\000\076\000\076\000\000\000\076\000\000\000\076\000\
\076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
\000\000\076\000\076\000\076\000\076\000\000\000\076\000\000\000\
\000\000\076\000\000\000\076\000\076\000\076\000\076\000\076\000\
\076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
\076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
\136\001\000\000\220\000\000\000\000\000\000\000\221\000\000\000\
\223\000\000\000\224\000\225\000\226\000\227\000\000\000\228\000\
\229\000\230\000\231\000\000\000\232\000\233\000\234\000\235\000\
\000\000\236\000\237\000\238\000\239\000\240\000\241\000\242\000\
\243\000\244\000\245\000\246\000\247\000\248\000\249\000\250\000\
\251\000\252\000\253\000\254\000\255\000\000\001\071\000\071\000\
\000\000\071\000\000\000\071\000\000\000\071\000\000\000\071\000\
\071\000\071\000\071\000\071\000\000\000\071\000\071\000\071\000\
\071\000\000\000\071\000\000\000\000\000\000\000\000\000\071\000\
\071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
\071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
\071\000\071\000\071\000\071\000\072\000\072\000\000\000\072\000\
\000\000\072\000\000\000\072\000\000\000\072\000\072\000\072\000\
\072\000\072\000\000\000\072\000\072\000\072\000\072\000\000\000\
\072\000\000\000\000\000\000\000\000\000\072\000\072\000\072\000\
\072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
\072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
\072\000\072\000\074\000\074\000\000\000\074\000\000\000\074\000\
\000\000\074\000\000\000\074\000\074\000\074\000\074\000\074\000\
\000\000\074\000\074\000\074\000\074\000\000\000\074\000\000\000\
\000\000\000\000\000\000\074\000\074\000\074\000\074\000\074\000\
\074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
\074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
\073\000\073\000\000\000\073\000\000\000\073\000\000\000\073\000\
\000\000\073\000\073\000\073\000\073\000\073\000\000\000\073\000\
\073\000\073\000\073\000\000\000\073\000\000\000\000\000\000\000\
\000\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
\073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
\073\000\073\000\073\000\073\000\073\000\073\000\068\000\068\000\
\000\000\068\000\000\000\068\000\000\000\068\000\000\000\068\000\
\068\000\068\000\068\000\068\000\000\000\068\000\068\000\068\000\
\068\000\000\000\068\000\000\000\000\000\000\000\000\000\068\000\
\068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
\068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
\068\000\068\000\068\000\068\000\067\000\067\000\000\000\067\000\
\000\000\067\000\000\000\067\000\000\000\067\000\067\000\067\000\
\067\000\067\000\000\000\067\000\067\000\067\000\067\000\000\000\
\067\000\000\000\000\000\000\000\000\000\067\000\067\000\067\000\
\067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
\067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
\067\000\067\000\066\000\066\000\000\000\066\000\000\000\066\000\
\000\000\066\000\000\000\066\000\066\000\066\000\066\000\066\000\
\000\000\066\000\066\000\066\000\066\000\000\000\066\000\000\000\
\000\000\000\000\000\000\066\000\066\000\066\000\066\000\066\000\
\066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
\066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
\075\000\075\000\000\000\075\000\000\000\075\000\000\000\075\000\
\000\000\075\000\075\000\075\000\075\000\075\000\000\000\075\000\
\075\000\075\000\075\000\000\000\075\000\000\000\000\000\000\000\
\000\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
\075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
\075\000\075\000\075\000\075\000\075\000\075\000\064\000\064\000\
\000\000\064\000\000\000\064\000\000\000\064\000\000\000\064\000\
\064\000\000\000\064\000\064\000\000\000\000\000\064\000\064\000\
\064\000\000\000\000\000\000\000\000\000\000\000\000\000\064\000\
\064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\
\064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\
\064\000\064\000\064\000\064\000\065\000\065\000\000\000\065\000\
\000\000\065\000\000\000\065\000\000\000\065\000\065\000\000\000\
\065\000\065\000\000\000\000\000\065\000\065\000\065\000\000\000\
\000\000\000\000\000\000\000\000\000\000\065\000\065\000\065\000\
\065\000\065\000\065\000\065\000\065\000\065\000\065\000\065\000\
\065\000\065\000\065\000\065\000\065\000\065\000\065\000\065\000\
\065\000\065\000\062\000\062\000\000\000\062\000\000\000\062\000\
\000\000\062\000\000\000\062\000\062\000\000\000\062\000\062\000\
\000\000\000\000\000\000\000\000\062\000\000\000\000\000\000\000\
\000\000\000\000\000\000\062\000\062\000\062\000\062\000\062\000\
\062\000\062\000\062\000\062\000\062\000\062\000\062\000\062\000\
\062\000\062\000\062\000\062\000\062\000\062\000\062\000\062\000\
\063\000\063\000\000\000\063\000\000\000\063\000\000\000\063\000\
\000\000\063\000\063\000\000\000\063\000\063\000\000\000\000\000\
\000\000\000\000\063\000\000\000\000\000\000\000\000\000\000\000\
\000\000\063\000\063\000\063\000\063\000\063\000\063\000\063\000\
\063\000\063\000\063\000\063\000\063\000\063\000\063\000\063\000\
\063\000\063\000\063\000\063\000\063\000\063\000\088\000\088\000\
\000\000\000\000\000\000\088\000\000\000\088\000\088\000\088\000\
\088\000\088\000\088\000\000\000\088\000\088\000\088\000\088\000\
\000\000\088\000\088\000\088\000\088\000\000\000\088\000\088\000\
\088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
\088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
\088\000\088\000\088\000\220\000\000\000\000\000\000\000\221\000\
\222\000\223\000\000\000\224\000\225\000\226\000\227\000\000\000\
\228\000\229\000\230\000\231\000\000\000\232\000\233\000\234\000\
\235\000\000\000\236\000\237\000\238\000\239\000\240\000\241\000\
\242\000\243\000\244\000\245\000\246\000\247\000\248\000\249\000\
\250\000\251\000\252\000\253\000\254\000\255\000\000\001\220\000\
\000\000\000\000\000\000\221\000\028\001\223\000\000\000\224\000\
\225\000\226\000\227\000\000\000\228\000\229\000\230\000\231\000\
\000\000\232\000\233\000\234\000\235\000\000\000\236\000\237\000\
\238\000\239\000\240\000\241\000\242\000\243\000\244\000\245\000\
\246\000\247\000\248\000\249\000\250\000\251\000\252\000\253\000\
\254\000\255\000\000\001\220\000\000\000\000\000\000\000\221\000\
\121\001\223\000\000\000\224\000\225\000\226\000\227\000\000\000\
\228\000\229\000\230\000\231\000\000\000\232\000\233\000\234\000\
\235\000\000\000\236\000\237\000\238\000\239\000\240\000\241\000\
\242\000\243\000\244\000\245\000\246\000\247\000\248\000\249\000\
\250\000\251\000\252\000\253\000\254\000\255\000\000\001\220\000\
\000\000\000\000\000\000\221\000\148\001\223\000\000\000\224\000\
\225\000\226\000\227\000\000\000\228\000\229\000\230\000\231\000\
\000\000\232\000\233\000\234\000\235\000\000\000\236\000\237\000\
\238\000\239\000\240\000\241\000\242\000\243\000\244\000\245\000\
\246\000\247\000\248\000\249\000\250\000\251\000\252\000\253\000\
\254\000\255\000\000\001\220\000\000\000\000\000\000\000\221\000\
\150\001\223\000\000\000\224\000\225\000\226\000\227\000\000\000\
\228\000\229\000\230\000\231\000\000\000\232\000\233\000\234\000\
\235\000\000\000\236\000\237\000\238\000\239\000\240\000\241\000\
\242\000\243\000\244\000\245\000\246\000\247\000\248\000\249\000\
\250\000\251\000\252\000\253\000\254\000\255\000\000\001\220\000\
\000\000\000\000\000\000\221\000\000\000\223\000\000\000\224\000\
\225\000\226\000\227\000\000\000\228\000\229\000\230\000\231\000\
\000\000\232\000\233\000\234\000\235\000\000\000\236\000\237\000\
\238\000\239\000\240\000\241\000\242\000\243\000\244\000\245\000\
\246\000\247\000\248\000\249\000\250\000\251\000\252\000\253\000\
\254\000\255\000\000\001\060\000\060\000\000\000\060\000\000\000\
\060\000\000\000\060\000\000\000\060\000\060\000\000\000\060\000\
\060\000\000\000\000\000\000\000\000\000\060\000\000\000\000\000\
\000\000\000\000\000\000\000\000\060\000\060\000\060\000\060\000\
\060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
\060\000\060\000\060\000\060\000\060\000\060\000\060\000\061\000\
\061\000\000\000\061\000\000\000\061\000\000\000\061\000\000\000\
\061\000\061\000\000\000\061\000\061\000\000\000\000\000\000\000\
\000\000\061\000\000\000\000\000\000\000\000\000\000\000\000\000\
\061\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\
\061\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\
\061\000\061\000\061\000\058\000\058\000\000\000\058\000\000\000\
\058\000\000\000\058\000\000\000\058\000\058\000\000\000\058\000\
\058\000\000\000\000\000\000\000\000\000\058\000\000\000\000\000\
\000\000\000\000\000\000\000\000\058\000\058\000\058\000\058\000\
\058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\
\058\000\058\000\058\000\058\000\058\000\058\000\058\000\059\000\
\059\000\000\000\059\000\000\000\059\000\000\000\059\000\000\000\
\059\000\059\000\000\000\059\000\059\000\000\000\000\000\000\000\
\000\000\059\000\000\000\000\000\000\000\000\000\000\000\000\000\
\059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
\059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
\059\000\059\000\059\000\220\000\000\000\000\000\000\000\221\000\
\000\000\223\000\000\000\224\000\225\000\226\000\227\000\000\000\
\228\000\229\000\230\000\231\000\000\000\232\000\233\000\234\000\
\235\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\247\000\248\000\249\000\
\250\000\251\000\252\000\253\000\254\000\255\000\000\001\056\000\
\056\000\000\000\056\000\000\000\056\000\000\000\056\000\000\000\
\056\000\056\000\000\000\056\000\056\000\000\000\000\000\000\000\
\000\000\056\000\000\000\000\000\000\000\000\000\000\000\000\000\
\056\000\056\000\056\000\056\000\056\000\056\000\056\000\056\000\
\056\000\056\000\056\000\000\000\000\000\000\000\000\000\056\000\
\056\000\056\000\056\000\057\000\057\000\000\000\057\000\000\000\
\057\000\000\000\057\000\000\000\057\000\057\000\000\000\057\000\
\057\000\000\000\000\000\000\000\000\000\057\000\000\000\000\000\
\000\000\000\000\000\000\000\000\057\000\057\000\057\000\057\000\
\057\000\057\000\057\000\057\000\057\000\057\000\057\000\000\000\
\000\000\000\000\000\000\057\000\057\000\057\000\057\000\220\000\
\000\000\000\000\000\000\221\000\000\000\223\000\000\000\000\000\
\225\000\226\000\227\000\000\000\228\000\229\000\230\000\231\000\
\000\000\232\000\233\000\234\000\235\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\247\000\248\000\249\000\250\000\251\000\252\000\000\000\
\254\000\255\000\000\001\055\000\055\000\000\000\055\000\000\000\
\055\000\000\000\055\000\000\000\055\000\055\000\000\000\055\000\
\055\000\000\000\000\000\000\000\000\000\055\000\000\000\000\000\
\000\000\000\000\000\000\000\000\055\000\055\000\055\000\055\000\
\055\000\055\000\055\000\055\000\055\000\055\000\055\000\000\000\
\000\000\000\000\000\000\220\000\000\000\055\000\055\000\221\000\
\000\000\223\000\000\000\000\000\225\000\226\000\227\000\000\000\
\228\000\229\000\230\000\231\000\000\000\232\000\233\000\234\000\
\235\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\247\000\248\000\249\000\
\250\000\251\000\252\000\000\000\000\000\255\000\000\001\054\000\
\054\000\000\000\054\000\000\000\054\000\000\000\054\000\000\000\
\054\000\054\000\000\000\000\000\054\000\000\000\000\000\000\000\
\000\000\054\000\000\000\000\000\000\000\000\000\000\000\000\000\
\054\000\054\000\054\000\054\000\054\000\054\000\054\000\054\000\
\054\000\054\000\054\000\000\000\000\000\000\000\000\000\220\000\
\000\000\054\000\054\000\221\000\000\000\223\000\000\000\000\000\
\225\000\226\000\000\000\000\000\228\000\229\000\230\000\231\000\
\000\000\232\000\233\000\234\000\235\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\247\000\248\000\249\000\250\000\251\000\252\000\220\000\
\000\000\255\000\000\001\221\000\000\000\223\000\000\000\000\000\
\225\000\226\000\000\000\000\000\228\000\229\000\230\000\000\000\
\000\000\232\000\233\000\234\000\235\000\052\000\052\000\000\000\
\052\000\000\000\052\000\000\000\052\000\000\000\052\000\052\000\
\000\000\247\000\248\000\249\000\250\000\251\000\252\000\000\000\
\000\000\255\000\000\001\000\000\000\000\000\000\052\000\052\000\
\052\000\052\000\052\000\052\000\052\000\052\000\052\000\052\000\
\052\000\000\000\000\000\000\000\000\000\220\000\000\000\052\000\
\052\000\221\000\000\000\223\000\000\000\000\000\225\000\000\000\
\000\000\000\000\228\000\229\000\230\000\000\000\000\000\232\000\
\233\000\234\000\235\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\247\000\
\248\000\249\000\250\000\251\000\252\000\220\000\000\000\255\000\
\000\001\221\000\000\000\223\000\000\000\000\000\225\000\000\000\
\000\000\000\000\228\000\229\000\230\000\000\000\000\000\232\000\
\233\000\234\000\235\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\247\000\
\248\000\249\000\250\000\000\000\000\000\220\000\000\000\255\000\
\000\001\221\000\000\000\223\000\000\000\000\000\225\000\000\000\
\000\000\000\000\228\000\229\000\230\000\000\000\000\000\232\000\
\233\000\234\000\235\000\051\000\051\000\000\000\051\000\000\000\
\051\000\000\000\051\000\000\000\051\000\051\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\255\000\
\000\001\000\000\000\000\000\000\051\000\051\000\051\000\051\000\
\051\000\051\000\051\000\051\000\051\000\051\000\051\000\009\000\
\010\000\011\000\012\000\013\000\014\000\051\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\015\000\016\000\017\000\018\000\019\000\020\000\
\021\000\000\000\000\000\022\000\023\000\024\000\025\000\026\000\
\000\000\027\000\039\000\039\000\028\000\039\000\000\000\039\000\
\000\000\039\000\000\000\039\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\009\000\010\000\011\000\012\000\013\000\
\014\000\000\000\000\000\039\000\039\000\039\000\039\000\039\000\
\039\000\039\000\039\000\039\000\039\000\039\000\015\000\016\000\
\017\000\018\000\019\000\020\000\021\000\013\000\014\000\022\000\
\023\000\024\000\025\000\026\000\000\000\027\000\000\000\000\000\
\000\000\000\000\000\000\000\000\015\000\016\000\017\000\018\000\
\019\000\020\000\021\000\000\000\000\000\022\000\023\000\024\000\
\025\000\026\000\000\000\220\000\000\000\000\000\028\000\221\000\
\000\000\223\000\000\000\000\000\225\000\000\000\000\000\000\000\
\228\000\229\000\230\000\000\000\000\000\232\000\233\000\234\000\
\235\000"

val yycheck = "\075\000\
\004\000\120\000\006\000\007\000\004\000\074\000\069\000\098\000\
\033\000\120\000\069\000\069\000\093\000\004\000\007\000\006\000\
\092\000\199\000\094\000\166\000\167\000\184\000\098\000\205\000\
\093\000\005\001\145\000\143\000\004\000\148\000\034\000\043\000\
\114\000\115\000\097\000\003\001\004\000\025\001\097\000\097\000\
\156\000\001\000\003\001\034\000\048\000\025\001\005\001\003\001\
\045\000\001\001\062\000\021\001\128\000\129\000\047\000\055\001\
\056\001\025\001\010\001\135\000\136\000\003\001\004\001\054\000\
\025\001\007\001\025\001\002\001\010\001\045\000\004\001\001\001\
\001\001\003\001\003\001\004\001\080\000\045\000\007\001\170\000\
\105\000\081\000\002\001\025\001\081\000\055\001\056\001\163\000\
\091\000\080\000\010\001\021\001\021\001\096\000\170\000\025\001\
\025\001\021\001\217\000\002\001\088\001\108\000\221\000\107\000\
\180\000\081\000\217\000\010\001\088\001\185\000\114\000\115\000\
\001\001\081\000\003\001\004\001\107\000\025\001\007\001\195\000\
\088\001\184\000\005\001\114\000\115\000\184\000\184\000\120\000\
\035\001\088\001\212\000\005\001\021\001\059\001\025\001\002\001\
\025\001\025\001\010\001\002\001\003\001\001\001\005\001\010\001\
\220\000\008\001\088\001\007\001\224\000\225\000\226\000\227\000\
\228\000\229\000\230\000\231\000\232\000\007\000\088\001\088\001\
\236\000\237\000\238\000\239\000\240\000\241\000\242\000\243\000\
\244\000\245\000\246\000\247\000\248\000\249\000\250\000\251\000\
\252\000\253\000\254\000\255\000\000\001\157\000\138\001\026\001\
\034\000\192\000\193\000\199\000\002\001\157\000\194\000\003\001\
\034\000\205\000\003\001\015\001\010\001\047\000\007\001\088\001\
\088\001\006\001\003\001\194\000\160\001\010\001\007\001\003\001\
\212\000\021\001\034\000\007\001\032\001\025\001\002\001\010\001\
\025\001\003\001\038\001\069\000\070\000\007\001\010\001\048\001\
\217\000\021\001\002\001\069\000\035\001\049\001\080\000\051\001\
\035\001\035\001\010\001\048\001\056\001\050\001\080\000\025\001\
\002\001\003\001\004\001\004\001\104\001\007\001\070\000\097\000\
\010\001\010\001\003\001\004\001\110\000\111\000\007\001\097\000\
\080\000\010\001\108\000\021\001\110\000\111\000\031\001\025\001\
\002\001\115\000\004\001\003\001\037\001\026\001\120\000\007\001\
\025\001\115\000\018\001\138\001\139\001\137\001\113\001\088\001\
\140\001\002\001\142\001\107\000\108\000\006\001\110\000\111\000\
\004\001\010\001\005\001\115\000\002\001\113\001\010\001\026\001\
\120\000\160\001\018\001\159\001\010\001\001\001\002\001\003\001\
\004\001\165\001\018\001\007\001\005\001\057\001\010\001\002\001\
\132\001\003\001\134\001\010\001\136\001\007\001\004\001\010\001\
\004\001\021\001\002\001\003\001\004\001\025\001\088\001\007\001\
\003\001\002\001\010\001\008\001\007\001\010\001\184\000\088\001\
\002\001\010\001\010\001\003\001\002\001\021\001\192\000\193\000\
\010\001\025\001\021\001\049\001\010\001\001\001\025\001\051\001\
\052\001\053\001\054\001\055\001\056\001\021\001\004\001\004\001\
\004\001\025\001\088\000\089\000\010\001\010\001\010\001\217\000\
\192\000\193\000\070\001\071\001\072\001\073\001\074\001\075\001\
\076\001\001\001\004\001\079\001\080\001\081\001\082\001\083\001\
\010\001\085\001\003\001\003\001\088\001\213\000\007\001\007\001\
\009\001\217\000\001\001\002\001\003\001\004\001\004\001\003\001\
\007\001\006\001\003\001\010\001\010\001\022\001\023\001\024\001\
\088\001\004\001\003\001\003\001\002\001\088\001\021\001\010\001\
\001\001\002\001\025\001\004\001\004\001\006\001\004\001\008\001\
\088\001\010\001\010\001\004\001\010\001\002\001\006\001\001\001\
\004\001\010\001\001\001\002\001\003\001\004\001\010\001\004\001\
\007\001\035\001\025\001\010\001\051\001\052\001\053\001\054\001\
\055\001\056\001\001\001\002\001\001\001\004\001\021\001\006\001\
\004\001\008\001\025\001\010\001\004\001\025\001\010\001\070\001\
\071\001\072\001\073\001\074\001\075\001\076\001\002\001\004\001\
\079\001\080\001\081\001\082\001\083\001\010\001\085\001\003\001\
\010\001\088\001\010\001\007\001\051\001\052\001\053\001\054\001\
\055\001\056\001\001\001\002\001\004\001\004\001\004\001\006\001\
\004\001\008\001\010\001\010\001\010\001\002\001\010\001\070\001\
\071\001\072\001\073\001\074\001\075\001\076\001\065\001\031\001\
\079\001\080\001\081\001\082\001\083\001\037\001\085\001\006\001\
\006\001\088\001\001\001\002\001\003\001\004\001\004\001\004\001\
\007\001\001\001\002\001\010\001\004\001\006\001\006\001\003\001\
\008\001\010\001\010\001\004\001\008\001\002\001\021\001\003\001\
\006\001\013\001\025\001\015\001\004\001\017\001\018\001\004\001\
\020\001\021\001\022\001\023\001\006\001\025\001\002\001\059\001\
\001\001\002\001\002\001\004\001\005\001\004\001\000\000\002\001\
\002\001\010\001\002\001\004\001\051\001\052\001\053\001\054\001\
\055\001\056\001\006\001\047\001\048\001\049\001\050\001\002\001\
\006\001\026\001\006\001\004\001\006\001\004\001\004\001\070\001\
\071\001\072\001\073\001\074\001\075\001\076\001\157\000\217\000\
\079\001\080\001\081\001\082\001\083\001\018\001\085\001\133\000\
\064\000\088\001\051\001\052\001\053\001\054\001\055\001\056\001\
\084\001\186\000\001\001\002\001\164\000\004\001\005\001\054\000\
\062\000\199\000\209\000\010\001\255\255\070\001\071\001\072\001\
\073\001\074\001\075\001\076\001\255\255\255\255\079\001\080\001\
\081\001\082\001\083\001\026\001\085\001\001\001\002\001\088\001\
\004\001\255\255\006\001\255\255\008\001\255\255\010\001\255\255\
\001\001\002\001\255\255\004\001\005\001\255\255\255\255\255\255\
\255\255\010\001\255\255\255\255\051\001\052\001\053\001\054\001\
\055\001\056\001\255\255\001\001\002\001\255\255\004\001\255\255\
\006\001\026\001\008\001\255\255\010\001\255\255\255\255\070\001\
\071\001\072\001\073\001\074\001\075\001\076\001\255\255\255\255\
\079\001\080\001\081\001\082\001\083\001\255\255\085\001\255\255\
\255\255\088\001\051\001\052\001\053\001\054\001\055\001\056\001\
\255\255\255\255\001\001\002\001\255\255\004\001\005\001\255\255\
\255\255\255\255\255\255\010\001\255\255\070\001\071\001\072\001\
\073\001\074\001\075\001\076\001\255\255\255\255\079\001\080\001\
\081\001\082\001\083\001\026\001\085\001\255\255\003\001\088\001\
\255\255\255\255\007\001\255\255\009\001\255\255\001\001\012\001\
\003\001\004\001\255\255\016\001\007\001\255\255\255\255\010\001\
\021\001\022\001\023\001\024\001\051\001\052\001\053\001\054\001\
\055\001\056\001\021\001\255\255\255\255\255\255\025\001\255\255\
\255\255\255\255\255\255\001\001\255\255\003\001\255\255\070\001\
\071\001\072\001\073\001\074\001\075\001\076\001\055\001\056\001\
\079\001\080\001\081\001\082\001\083\001\255\255\085\001\021\001\
\255\255\088\001\255\255\025\001\255\255\070\001\071\001\072\001\
\073\001\074\001\075\001\076\001\255\255\255\255\079\001\080\001\
\081\001\082\001\083\001\070\001\071\001\072\001\073\001\074\001\
\075\001\076\001\255\255\255\255\079\001\080\001\081\001\082\001\
\083\001\001\001\002\001\255\255\004\001\088\001\006\001\255\255\
\008\001\255\255\010\001\255\255\255\255\255\255\255\255\255\255\
\070\001\071\001\072\001\073\001\074\001\075\001\076\001\255\255\
\255\255\079\001\080\001\081\001\082\001\083\001\002\001\003\001\
\255\255\005\001\088\001\001\001\002\001\255\255\004\001\255\255\
\006\001\013\001\008\001\015\001\010\001\017\001\018\001\255\255\
\020\001\021\001\022\001\023\001\255\255\025\001\001\001\002\001\
\255\255\004\001\255\255\006\001\255\255\008\001\255\255\010\001\
\255\255\006\001\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\255\255\255\255\047\001\048\001\049\001\050\001\051\001\
\052\001\053\001\054\001\055\001\056\001\057\001\255\255\255\255\
\060\001\061\001\062\001\063\001\064\001\065\001\066\001\067\001\
\068\001\069\001\070\001\071\001\072\001\073\001\074\001\075\001\
\076\001\002\001\003\001\079\001\080\001\081\001\082\001\083\001\
\084\001\085\001\055\001\056\001\088\001\255\255\001\001\002\001\
\255\255\004\001\255\255\006\001\021\001\008\001\255\255\010\001\
\025\001\070\001\071\001\072\001\073\001\074\001\075\001\076\001\
\255\255\255\255\079\001\080\001\081\001\082\001\083\001\002\001\
\255\255\255\255\005\001\088\001\255\255\255\255\255\255\010\001\
\255\255\255\255\051\001\052\001\053\001\054\001\055\001\056\001\
\001\001\002\001\255\255\004\001\255\255\006\001\255\255\008\001\
\255\255\010\001\255\255\255\255\255\255\070\001\071\001\072\001\
\073\001\074\001\075\001\076\001\255\255\255\255\079\001\080\001\
\081\001\082\001\083\001\255\255\085\001\255\255\255\255\088\001\
\051\001\052\001\053\001\054\001\055\001\056\001\002\001\255\255\
\255\255\005\001\255\255\255\255\255\255\255\255\010\001\255\255\
\255\255\255\255\255\255\070\001\071\001\072\001\073\001\074\001\
\075\001\076\001\255\255\255\255\079\001\080\001\081\001\082\001\
\083\001\255\255\085\001\001\001\002\001\088\001\004\001\255\255\
\001\001\002\001\008\001\004\001\010\001\255\255\255\255\008\001\
\255\255\010\001\255\255\003\001\255\255\255\255\255\255\051\001\
\052\001\053\001\054\001\055\001\056\001\013\001\255\255\015\001\
\255\255\017\001\018\001\255\255\020\001\021\001\022\001\023\001\
\255\255\025\001\070\001\071\001\072\001\073\001\074\001\075\001\
\076\001\255\255\255\255\079\001\080\001\081\001\082\001\083\001\
\255\255\085\001\003\001\004\001\088\001\255\255\007\001\047\001\
\048\001\049\001\050\001\255\255\255\255\255\255\255\255\055\001\
\056\001\255\255\255\255\255\255\021\001\255\255\255\255\255\255\
\025\001\255\255\255\255\255\255\255\255\255\255\070\001\071\001\
\072\001\073\001\074\001\075\001\076\001\255\255\255\255\079\001\
\080\001\081\001\082\001\083\001\084\001\003\001\255\255\255\255\
\088\001\007\001\051\001\052\001\053\001\054\001\055\001\056\001\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\021\001\
\255\255\255\255\255\255\025\001\255\255\070\001\071\001\072\001\
\073\001\074\001\075\001\076\001\255\255\255\255\079\001\080\001\
\081\001\082\001\083\001\255\255\085\001\003\001\004\001\088\001\
\255\255\007\001\255\255\255\255\255\255\051\001\052\001\053\001\
\054\001\055\001\056\001\255\255\255\255\255\255\255\255\021\001\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
\070\001\071\001\072\001\073\001\074\001\075\001\076\001\255\255\
\255\255\079\001\080\001\081\001\082\001\083\001\003\001\085\001\
\255\255\255\255\088\001\255\255\255\255\051\001\052\001\053\001\
\054\001\055\001\056\001\255\255\255\255\255\255\255\255\255\255\
\021\001\255\255\255\255\255\255\025\001\255\255\255\255\255\255\
\070\001\071\001\072\001\073\001\074\001\075\001\076\001\255\255\
\255\255\079\001\080\001\081\001\082\001\083\001\004\001\085\001\
\255\255\255\255\088\001\255\255\255\255\255\255\051\001\052\001\
\053\001\054\001\055\001\056\001\255\255\255\255\255\255\255\255\
\255\255\255\255\255\255\025\001\255\255\255\255\255\255\255\255\
\255\255\070\001\071\001\072\001\073\001\074\001\075\001\076\001\
\255\255\255\255\079\001\080\001\081\001\082\001\083\001\004\001\
\085\001\255\255\255\255\088\001\255\255\051\001\052\001\053\001\
\054\001\055\001\056\001\255\255\255\255\255\255\255\255\255\255\
\255\255\255\255\255\255\255\255\025\001\255\255\255\255\255\255\
\070\001\071\001\072\001\073\001\074\001\075\001\076\001\255\255\
\255\255\079\001\080\001\081\001\082\001\083\001\004\001\085\001\
\255\255\255\255\088\001\255\255\255\255\255\255\051\001\052\001\
\053\001\054\001\055\001\056\001\255\255\255\255\255\255\255\255\
\255\255\255\255\255\255\025\001\255\255\255\255\255\255\255\255\
\255\255\070\001\071\001\072\001\073\001\074\001\075\001\076\001\
\255\255\255\255\079\001\080\001\081\001\082\001\083\001\004\001\
\085\001\255\255\255\255\088\001\255\255\051\001\052\001\053\001\
\054\001\055\001\056\001\255\255\255\255\255\255\255\255\255\255\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
\070\001\071\001\072\001\073\001\074\001\075\001\076\001\255\255\
\255\255\079\001\080\001\081\001\082\001\083\001\005\001\085\001\
\255\255\255\255\088\001\255\255\255\255\255\255\051\001\052\001\
\053\001\054\001\055\001\056\001\255\255\255\255\255\255\255\255\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\070\001\071\001\072\001\073\001\074\001\075\001\076\001\
\255\255\255\255\079\001\080\001\081\001\082\001\083\001\006\001\
\085\001\255\255\255\255\088\001\051\001\052\001\053\001\054\001\
\055\001\056\001\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\070\001\
\071\001\072\001\073\001\074\001\075\001\076\001\255\255\255\255\
\079\001\080\001\081\001\082\001\083\001\255\255\085\001\002\001\
\003\001\088\001\005\001\006\001\051\001\052\001\053\001\054\001\
\055\001\056\001\013\001\255\255\015\001\255\255\017\001\018\001\
\255\255\020\001\021\001\022\001\023\001\255\255\025\001\070\001\
\071\001\072\001\073\001\074\001\075\001\076\001\255\255\255\255\
\079\001\080\001\081\001\082\001\083\001\255\255\085\001\255\255\
\255\255\088\001\255\255\255\255\047\001\048\001\049\001\050\001\
\255\255\255\255\255\255\255\255\255\255\255\255\057\001\255\255\
\255\255\060\001\061\001\062\001\063\001\064\001\065\001\066\001\
\067\001\068\001\069\001\002\001\003\001\255\255\005\001\255\255\
\051\001\052\001\053\001\054\001\055\001\056\001\013\001\255\255\
\015\001\084\001\017\001\018\001\255\255\020\001\021\001\022\001\
\023\001\006\001\025\001\070\001\071\001\072\001\073\001\074\001\
\075\001\076\001\255\255\255\255\079\001\080\001\081\001\082\001\
\083\001\255\255\085\001\255\255\087\001\088\001\255\255\006\001\
\047\001\048\001\049\001\050\001\255\255\255\255\255\255\255\255\
\255\255\255\255\057\001\255\255\255\255\060\001\061\001\062\001\
\063\001\064\001\065\001\066\001\067\001\068\001\069\001\255\255\
\255\255\255\255\055\001\056\001\255\255\255\255\255\255\255\255\
\255\255\255\255\255\255\255\255\255\255\084\001\255\255\255\255\
\255\255\070\001\071\001\072\001\073\001\074\001\075\001\076\001\
\055\001\056\001\079\001\080\001\081\001\082\001\083\001\255\255\
\255\255\255\255\255\255\088\001\255\255\255\255\255\255\070\001\
\071\001\072\001\073\001\074\001\075\001\076\001\255\255\255\255\
\079\001\080\001\081\001\082\001\083\001\003\001\255\255\005\001\
\255\255\088\001\255\255\255\255\255\255\255\255\255\255\013\001\
\003\001\015\001\255\255\017\001\018\001\008\001\020\001\021\001\
\022\001\023\001\013\001\025\001\015\001\255\255\017\001\018\001\
\255\255\020\001\021\001\022\001\023\001\255\255\025\001\255\255\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\047\001\048\001\049\001\050\001\255\255\003\001\255\255\
\255\255\255\255\255\255\008\001\047\001\048\001\049\001\050\001\
\013\001\255\255\015\001\255\255\017\001\018\001\255\255\020\001\
\021\001\022\001\023\001\255\255\025\001\255\255\070\001\071\001\
\072\001\073\001\074\001\075\001\076\001\255\255\084\001\079\001\
\080\001\081\001\082\001\083\001\003\001\004\001\255\255\255\255\
\088\001\084\001\047\001\048\001\049\001\050\001\013\001\003\001\
\015\001\255\255\017\001\018\001\008\001\020\001\021\001\022\001\
\023\001\013\001\025\001\015\001\255\255\017\001\018\001\255\255\
\020\001\021\001\022\001\023\001\255\255\025\001\255\255\255\255\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\084\001\
\047\001\048\001\049\001\050\001\255\255\003\001\255\255\255\255\
\255\255\255\255\008\001\047\001\048\001\049\001\050\001\013\001\
\003\001\015\001\255\255\017\001\018\001\255\255\020\001\021\001\
\022\001\023\001\013\001\025\001\015\001\255\255\017\001\018\001\
\255\255\020\001\021\001\022\001\023\001\084\001\025\001\255\255\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
\084\001\047\001\048\001\049\001\050\001\255\255\003\001\255\255\
\255\255\255\255\003\001\255\255\047\001\048\001\049\001\050\001\
\013\001\255\255\015\001\255\255\017\001\018\001\015\001\020\001\
\021\001\022\001\023\001\020\001\025\001\022\001\023\001\255\255\
\025\001\255\255\001\001\002\001\255\255\004\001\084\001\006\001\
\255\255\008\001\255\255\010\001\011\001\255\255\255\255\014\001\
\255\255\084\001\047\001\048\001\049\001\050\001\047\001\048\001\
\049\001\050\001\255\255\026\001\027\001\028\001\029\001\030\001\
\031\001\032\001\033\001\034\001\035\001\036\001\255\255\255\255\
\255\255\255\255\255\255\255\255\043\001\044\001\255\255\255\255\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\084\001\
\255\255\255\255\255\255\084\001\001\001\002\001\003\001\004\001\
\255\255\006\001\007\001\008\001\009\001\010\001\011\001\012\001\
\013\001\014\001\255\255\016\001\017\001\018\001\019\001\255\255\
\021\001\022\001\023\001\024\001\255\255\026\001\027\001\028\001\
\029\001\030\001\031\001\032\001\033\001\034\001\035\001\036\001\
\037\001\038\001\039\001\040\001\041\001\042\001\043\001\044\001\
\045\001\046\001\001\001\002\001\255\255\004\001\255\255\006\001\
\007\001\008\001\009\001\010\001\011\001\012\001\013\001\014\001\
\255\255\016\001\017\001\018\001\019\001\255\255\021\001\255\255\
\255\255\024\001\255\255\026\001\027\001\028\001\029\001\030\001\
\031\001\032\001\033\001\034\001\035\001\036\001\037\001\038\001\
\039\001\040\001\041\001\042\001\043\001\044\001\045\001\046\001\
\001\001\255\255\003\001\255\255\255\255\255\255\007\001\255\255\
\009\001\255\255\011\001\012\001\013\001\014\001\255\255\016\001\
\017\001\018\001\019\001\255\255\021\001\022\001\023\001\024\001\
\255\255\026\001\027\001\028\001\029\001\030\001\031\001\032\001\
\033\001\034\001\035\001\036\001\037\001\038\001\039\001\040\001\
\041\001\042\001\043\001\044\001\045\001\046\001\001\001\002\001\
\255\255\004\001\255\255\006\001\255\255\008\001\255\255\010\001\
\011\001\012\001\013\001\014\001\255\255\016\001\017\001\018\001\
\019\001\255\255\021\001\255\255\255\255\255\255\255\255\026\001\
\027\001\028\001\029\001\030\001\031\001\032\001\033\001\034\001\
\035\001\036\001\037\001\038\001\039\001\040\001\041\001\042\001\
\043\001\044\001\045\001\046\001\001\001\002\001\255\255\004\001\
\255\255\006\001\255\255\008\001\255\255\010\001\011\001\012\001\
\013\001\014\001\255\255\016\001\017\001\018\001\019\001\255\255\
\021\001\255\255\255\255\255\255\255\255\026\001\027\001\028\001\
\029\001\030\001\031\001\032\001\033\001\034\001\035\001\036\001\
\037\001\038\001\039\001\040\001\041\001\042\001\043\001\044\001\
\045\001\046\001\001\001\002\001\255\255\004\001\255\255\006\001\
\255\255\008\001\255\255\010\001\011\001\012\001\013\001\014\001\
\255\255\016\001\017\001\018\001\019\001\255\255\021\001\255\255\
\255\255\255\255\255\255\026\001\027\001\028\001\029\001\030\001\
\031\001\032\001\033\001\034\001\035\001\036\001\037\001\038\001\
\039\001\040\001\041\001\042\001\043\001\044\001\045\001\046\001\
\001\001\002\001\255\255\004\001\255\255\006\001\255\255\008\001\
\255\255\010\001\011\001\012\001\013\001\014\001\255\255\016\001\
\017\001\018\001\019\001\255\255\021\001\255\255\255\255\255\255\
\255\255\026\001\027\001\028\001\029\001\030\001\031\001\032\001\
\033\001\034\001\035\001\036\001\037\001\038\001\039\001\040\001\
\041\001\042\001\043\001\044\001\045\001\046\001\001\001\002\001\
\255\255\004\001\255\255\006\001\255\255\008\001\255\255\010\001\
\011\001\012\001\013\001\014\001\255\255\016\001\017\001\018\001\
\019\001\255\255\021\001\255\255\255\255\255\255\255\255\026\001\
\027\001\028\001\029\001\030\001\031\001\032\001\033\001\034\001\
\035\001\036\001\037\001\038\001\039\001\040\001\041\001\042\001\
\043\001\044\001\045\001\046\001\001\001\002\001\255\255\004\001\
\255\255\006\001\255\255\008\001\255\255\010\001\011\001\012\001\
\013\001\014\001\255\255\016\001\017\001\018\001\019\001\255\255\
\021\001\255\255\255\255\255\255\255\255\026\001\027\001\028\001\
\029\001\030\001\031\001\032\001\033\001\034\001\035\001\036\001\
\037\001\038\001\039\001\040\001\041\001\042\001\043\001\044\001\
\045\001\046\001\001\001\002\001\255\255\004\001\255\255\006\001\
\255\255\008\001\255\255\010\001\011\001\012\001\013\001\014\001\
\255\255\016\001\017\001\018\001\019\001\255\255\021\001\255\255\
\255\255\255\255\255\255\026\001\027\001\028\001\029\001\030\001\
\031\001\032\001\033\001\034\001\035\001\036\001\037\001\038\001\
\039\001\040\001\041\001\042\001\043\001\044\001\045\001\046\001\
\001\001\002\001\255\255\004\001\255\255\006\001\255\255\008\001\
\255\255\010\001\011\001\012\001\013\001\014\001\255\255\016\001\
\017\001\018\001\019\001\255\255\021\001\255\255\255\255\255\255\
\255\255\026\001\027\001\028\001\029\001\030\001\031\001\032\001\
\033\001\034\001\035\001\036\001\037\001\038\001\039\001\040\001\
\041\001\042\001\043\001\044\001\045\001\046\001\001\001\002\001\
\255\255\004\001\255\255\006\001\255\255\008\001\255\255\010\001\
\011\001\255\255\013\001\014\001\255\255\255\255\017\001\018\001\
\019\001\255\255\255\255\255\255\255\255\255\255\255\255\026\001\
\027\001\028\001\029\001\030\001\031\001\032\001\033\001\034\001\
\035\001\036\001\037\001\038\001\039\001\040\001\041\001\042\001\
\043\001\044\001\045\001\046\001\001\001\002\001\255\255\004\001\
\255\255\006\001\255\255\008\001\255\255\010\001\011\001\255\255\
\013\001\014\001\255\255\255\255\017\001\018\001\019\001\255\255\
\255\255\255\255\255\255\255\255\255\255\026\001\027\001\028\001\
\029\001\030\001\031\001\032\001\033\001\034\001\035\001\036\001\
\037\001\038\001\039\001\040\001\041\001\042\001\043\001\044\001\
\045\001\046\001\001\001\002\001\255\255\004\001\255\255\006\001\
\255\255\008\001\255\255\010\001\011\001\255\255\013\001\014\001\
\255\255\255\255\255\255\255\255\019\001\255\255\255\255\255\255\
\255\255\255\255\255\255\026\001\027\001\028\001\029\001\030\001\
\031\001\032\001\033\001\034\001\035\001\036\001\037\001\038\001\
\039\001\040\001\041\001\042\001\043\001\044\001\045\001\046\001\
\001\001\002\001\255\255\004\001\255\255\006\001\255\255\008\001\
\255\255\010\001\011\001\255\255\013\001\014\001\255\255\255\255\
\255\255\255\255\019\001\255\255\255\255\255\255\255\255\255\255\
\255\255\026\001\027\001\028\001\029\001\030\001\031\001\032\001\
\033\001\034\001\035\001\036\001\037\001\038\001\039\001\040\001\
\041\001\042\001\043\001\044\001\045\001\046\001\002\001\003\001\
\255\255\255\255\255\255\007\001\255\255\009\001\010\001\011\001\
\012\001\013\001\014\001\255\255\016\001\017\001\018\001\019\001\
\255\255\021\001\022\001\023\001\024\001\255\255\026\001\027\001\
\028\001\029\001\030\001\031\001\032\001\033\001\034\001\035\001\
\036\001\037\001\038\001\039\001\040\001\041\001\042\001\043\001\
\044\001\045\001\046\001\003\001\255\255\255\255\255\255\007\001\
\008\001\009\001\255\255\011\001\012\001\013\001\014\001\255\255\
\016\001\017\001\018\001\019\001\255\255\021\001\022\001\023\001\
\024\001\255\255\026\001\027\001\028\001\029\001\030\001\031\001\
\032\001\033\001\034\001\035\001\036\001\037\001\038\001\039\001\
\040\001\041\001\042\001\043\001\044\001\045\001\046\001\003\001\
\255\255\255\255\255\255\007\001\008\001\009\001\255\255\011\001\
\012\001\013\001\014\001\255\255\016\001\017\001\018\001\019\001\
\255\255\021\001\022\001\023\001\024\001\255\255\026\001\027\001\
\028\001\029\001\030\001\031\001\032\001\033\001\034\001\035\001\
\036\001\037\001\038\001\039\001\040\001\041\001\042\001\043\001\
\044\001\045\001\046\001\003\001\255\255\255\255\255\255\007\001\
\008\001\009\001\255\255\011\001\012\001\013\001\014\001\255\255\
\016\001\017\001\018\001\019\001\255\255\021\001\022\001\023\001\
\024\001\255\255\026\001\027\001\028\001\029\001\030\001\031\001\
\032\001\033\001\034\001\035\001\036\001\037\001\038\001\039\001\
\040\001\041\001\042\001\043\001\044\001\045\001\046\001\003\001\
\255\255\255\255\255\255\007\001\008\001\009\001\255\255\011\001\
\012\001\013\001\014\001\255\255\016\001\017\001\018\001\019\001\
\255\255\021\001\022\001\023\001\024\001\255\255\026\001\027\001\
\028\001\029\001\030\001\031\001\032\001\033\001\034\001\035\001\
\036\001\037\001\038\001\039\001\040\001\041\001\042\001\043\001\
\044\001\045\001\046\001\003\001\255\255\255\255\255\255\007\001\
\008\001\009\001\255\255\011\001\012\001\013\001\014\001\255\255\
\016\001\017\001\018\001\019\001\255\255\021\001\022\001\023\001\
\024\001\255\255\026\001\027\001\028\001\029\001\030\001\031\001\
\032\001\033\001\034\001\035\001\036\001\037\001\038\001\039\001\
\040\001\041\001\042\001\043\001\044\001\045\001\046\001\003\001\
\255\255\255\255\255\255\007\001\255\255\009\001\255\255\011\001\
\012\001\013\001\014\001\255\255\016\001\017\001\018\001\019\001\
\255\255\021\001\022\001\023\001\024\001\255\255\026\001\027\001\
\028\001\029\001\030\001\031\001\032\001\033\001\034\001\035\001\
\036\001\037\001\038\001\039\001\040\001\041\001\042\001\043\001\
\044\001\045\001\046\001\001\001\002\001\255\255\004\001\255\255\
\006\001\255\255\008\001\255\255\010\001\011\001\255\255\013\001\
\014\001\255\255\255\255\255\255\255\255\019\001\255\255\255\255\
\255\255\255\255\255\255\255\255\026\001\027\001\028\001\029\001\
\030\001\031\001\032\001\033\001\034\001\035\001\036\001\037\001\
\038\001\039\001\040\001\041\001\042\001\043\001\044\001\001\001\
\002\001\255\255\004\001\255\255\006\001\255\255\008\001\255\255\
\010\001\011\001\255\255\013\001\014\001\255\255\255\255\255\255\
\255\255\019\001\255\255\255\255\255\255\255\255\255\255\255\255\
\026\001\027\001\028\001\029\001\030\001\031\001\032\001\033\001\
\034\001\035\001\036\001\037\001\038\001\039\001\040\001\041\001\
\042\001\043\001\044\001\001\001\002\001\255\255\004\001\255\255\
\006\001\255\255\008\001\255\255\010\001\011\001\255\255\013\001\
\014\001\255\255\255\255\255\255\255\255\019\001\255\255\255\255\
\255\255\255\255\255\255\255\255\026\001\027\001\028\001\029\001\
\030\001\031\001\032\001\033\001\034\001\035\001\036\001\037\001\
\038\001\039\001\040\001\041\001\042\001\043\001\044\001\001\001\
\002\001\255\255\004\001\255\255\006\001\255\255\008\001\255\255\
\010\001\011\001\255\255\013\001\014\001\255\255\255\255\255\255\
\255\255\019\001\255\255\255\255\255\255\255\255\255\255\255\255\
\026\001\027\001\028\001\029\001\030\001\031\001\032\001\033\001\
\034\001\035\001\036\001\037\001\038\001\039\001\040\001\041\001\
\042\001\043\001\044\001\003\001\255\255\255\255\255\255\007\001\
\255\255\009\001\255\255\011\001\012\001\013\001\014\001\255\255\
\016\001\017\001\018\001\019\001\255\255\021\001\022\001\023\001\
\024\001\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\255\255\255\255\255\255\255\255\037\001\038\001\039\001\
\040\001\041\001\042\001\043\001\044\001\045\001\046\001\001\001\
\002\001\255\255\004\001\255\255\006\001\255\255\008\001\255\255\
\010\001\011\001\255\255\013\001\014\001\255\255\255\255\255\255\
\255\255\019\001\255\255\255\255\255\255\255\255\255\255\255\255\
\026\001\027\001\028\001\029\001\030\001\031\001\032\001\033\001\
\034\001\035\001\036\001\255\255\255\255\255\255\255\255\041\001\
\042\001\043\001\044\001\001\001\002\001\255\255\004\001\255\255\
\006\001\255\255\008\001\255\255\010\001\011\001\255\255\013\001\
\014\001\255\255\255\255\255\255\255\255\019\001\255\255\255\255\
\255\255\255\255\255\255\255\255\026\001\027\001\028\001\029\001\
\030\001\031\001\032\001\033\001\034\001\035\001\036\001\255\255\
\255\255\255\255\255\255\041\001\042\001\043\001\044\001\003\001\
\255\255\255\255\255\255\007\001\255\255\009\001\255\255\255\255\
\012\001\013\001\014\001\255\255\016\001\017\001\018\001\019\001\
\255\255\021\001\022\001\023\001\024\001\255\255\255\255\255\255\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\037\001\038\001\039\001\040\001\041\001\042\001\255\255\
\044\001\045\001\046\001\001\001\002\001\255\255\004\001\255\255\
\006\001\255\255\008\001\255\255\010\001\011\001\255\255\013\001\
\014\001\255\255\255\255\255\255\255\255\019\001\255\255\255\255\
\255\255\255\255\255\255\255\255\026\001\027\001\028\001\029\001\
\030\001\031\001\032\001\033\001\034\001\035\001\036\001\255\255\
\255\255\255\255\255\255\003\001\255\255\043\001\044\001\007\001\
\255\255\009\001\255\255\255\255\012\001\013\001\014\001\255\255\
\016\001\017\001\018\001\019\001\255\255\021\001\022\001\023\001\
\024\001\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\255\255\255\255\255\255\255\255\037\001\038\001\039\001\
\040\001\041\001\042\001\255\255\255\255\045\001\046\001\001\001\
\002\001\255\255\004\001\255\255\006\001\255\255\008\001\255\255\
\010\001\011\001\255\255\255\255\014\001\255\255\255\255\255\255\
\255\255\019\001\255\255\255\255\255\255\255\255\255\255\255\255\
\026\001\027\001\028\001\029\001\030\001\031\001\032\001\033\001\
\034\001\035\001\036\001\255\255\255\255\255\255\255\255\003\001\
\255\255\043\001\044\001\007\001\255\255\009\001\255\255\255\255\
\012\001\013\001\255\255\255\255\016\001\017\001\018\001\019\001\
\255\255\021\001\022\001\023\001\024\001\255\255\255\255\255\255\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\037\001\038\001\039\001\040\001\041\001\042\001\003\001\
\255\255\045\001\046\001\007\001\255\255\009\001\255\255\255\255\
\012\001\013\001\255\255\255\255\016\001\017\001\018\001\255\255\
\255\255\021\001\022\001\023\001\024\001\001\001\002\001\255\255\
\004\001\255\255\006\001\255\255\008\001\255\255\010\001\011\001\
\255\255\037\001\038\001\039\001\040\001\041\001\042\001\255\255\
\255\255\045\001\046\001\255\255\255\255\255\255\026\001\027\001\
\028\001\029\001\030\001\031\001\032\001\033\001\034\001\035\001\
\036\001\255\255\255\255\255\255\255\255\003\001\255\255\043\001\
\044\001\007\001\255\255\009\001\255\255\255\255\012\001\255\255\
\255\255\255\255\016\001\017\001\018\001\255\255\255\255\021\001\
\022\001\023\001\024\001\255\255\255\255\255\255\255\255\255\255\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\037\001\
\038\001\039\001\040\001\041\001\042\001\003\001\255\255\045\001\
\046\001\007\001\255\255\009\001\255\255\255\255\012\001\255\255\
\255\255\255\255\016\001\017\001\018\001\255\255\255\255\021\001\
\022\001\023\001\024\001\255\255\255\255\255\255\255\255\255\255\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\037\001\
\038\001\039\001\040\001\255\255\255\255\003\001\255\255\045\001\
\046\001\007\001\255\255\009\001\255\255\255\255\012\001\255\255\
\255\255\255\255\016\001\017\001\018\001\255\255\255\255\021\001\
\022\001\023\001\024\001\001\001\002\001\255\255\004\001\255\255\
\006\001\255\255\008\001\255\255\010\001\011\001\255\255\255\255\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\045\001\
\046\001\255\255\255\255\255\255\026\001\027\001\028\001\029\001\
\030\001\031\001\032\001\033\001\034\001\035\001\036\001\051\001\
\052\001\053\001\054\001\055\001\056\001\043\001\255\255\255\255\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\255\255\070\001\071\001\072\001\073\001\074\001\075\001\
\076\001\255\255\255\255\079\001\080\001\081\001\082\001\083\001\
\255\255\085\001\001\001\002\001\088\001\004\001\255\255\006\001\
\255\255\008\001\255\255\010\001\255\255\255\255\255\255\255\255\
\255\255\255\255\255\255\051\001\052\001\053\001\054\001\055\001\
\056\001\255\255\255\255\026\001\027\001\028\001\029\001\030\001\
\031\001\032\001\033\001\034\001\035\001\036\001\070\001\071\001\
\072\001\073\001\074\001\075\001\076\001\055\001\056\001\079\001\
\080\001\081\001\082\001\083\001\255\255\085\001\255\255\255\255\
\255\255\255\255\255\255\255\255\070\001\071\001\072\001\073\001\
\074\001\075\001\076\001\255\255\255\255\079\001\080\001\081\001\
\082\001\083\001\255\255\003\001\255\255\255\255\088\001\007\001\
\255\255\009\001\255\255\255\255\012\001\255\255\255\255\255\255\
\016\001\017\001\018\001\255\255\255\255\021\001\022\001\023\001\
\024\001"

val yynames_const = "\
  EOF\000\
  COLON\000\
  SEMICOLON\000\
  LPAREN\000\
  RPAREN\000\
  LCURLY\000\
  RCURLY\000\
  LBRACE\000\
  RBRACE\000\
  DOT\000\
  COMMA\000\
  QUESTION\000\
  PERCENT\000\
  AMP\000\
  BAR\000\
  TILDE\000\
  DIVIDE\000\
  PLUS\000\
  MINUS\000\
  HAT\000\
  BANG\000\
  TIMES\000\
  INC\000\
  DEC\000\
  ARROW\000\
  EQUALS\000\
  PLUSEQUALS\000\
  MINUSEQUALS\000\
  XOREQUALS\000\
  MODEQUALS\000\
  TIMESEQUALS\000\
  DIVEQUALS\000\
  OREQUALS\000\
  ANDEQUALS\000\
  LSHIFTEQUALS\000\
  RSHIFTEQUALS\000\
  LTE\000\
  GTE\000\
  LT\000\
  GT\000\
  EQ\000\
  NEQ\000\
  OR\000\
  AND\000\
  LSHIFT\000\
  RSHIFT\000\
  EXTERN\000\
  AUTO\000\
  STATIC\000\
  REGISTER\000\
  CONST\000\
  VOLATILE\000\
  IF\000\
  THEN\000\
  ELSE\000\
  FOR\000\
  DO\000\
  SWITCH\000\
  CASE\000\
  DEFAULT\000\
  WHILE\000\
  RETURN\000\
  BREAK\000\
  CONTINUE\000\
  GOTO\000\
  CHAR\000\
  DOUBLE\000\
  ENUM\000\
  FLOAT\000\
  INT\000\
  LONG\000\
  SHORT\000\
  FRACTIONAL\000\
  SATURATE\000\
  STRUCT\000\
  UNION\000\
  UNSIGNED\000\
  SIGNED\000\
  VOID\000\
  SIZEOF\000\
  TYPEDEF\000\
  UNARY\000\
  ELIPSIS\000\
  PARALLEL\000\
  SEQUENTIAL\000\
  COUNT\000\
  IN\000\
  REDUCE\000\
  WHERE\000\
  DEPENDENCY\000\
  ASSOCIATIVE\000\
  TO\000\
  KWAND\000\
  "

val yynames_block = "\
  ID\000\
  DECNUM\000\
  REALNUM\000\
  STRING\000\
  CCONST\000\
  TYPE_NAME\000\
  "

val yyact = [|
  (fn _ => failwith "parser")
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 223 "c/c_parser.ny" *)
       seqToList a_1) : ParseTree.externalDecl list)  end )
, (fn parser_env =>
	Obj.repr((
(* # 226 "c/c_parser.ny" *)
                           emptySeq)))
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 227 "c/c_parser.ny" *)
                           addOptToEnd(a_2, a_1))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 230 "c/c_parser.ny" *)
                        Some(MARKexternalDecl (ExternalDecl a_1)))) end )
, (fn parser_env =>
	Obj.repr((
(* # 231 "c/c_parser.ny" *)
                        None)))
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 232 "c/c_parser.ny" *)
                        Some (MARKexternalDecl a_1))) end )
, (fn parser_env =>
	let val a_3 = peek_val parser_env 6 in
	let val a_5 = peek_val parser_env 4 in
	let val a_7 = peek_val parser_env 2 in
	let val a_9 = peek_val parser_env 0 in
	Obj.repr((
(* # 236 "c/c_parser.ny" *)
                            MARKstatement (For(a_3,a_5,a_7,a_9)))) end  end  end  end )
, (fn parser_env =>
	let val a_3 = peek_val parser_env 2 in
	let val a_5 = peek_val parser_env 0 in
	Obj.repr((
(* # 238 "c/c_parser.ny" *)
                            MARKstatement (While(a_3,a_5)))) end  end )
, (fn parser_env =>
	let val a_3 = peek_val parser_env 2 in
	let val a_5 = peek_val parser_env 0 in
	Obj.repr((
(* # 240 "c/c_parser.ny" *)
                            MARKstatement (Switch(a_3,a_5)))) end  end )
, (fn parser_env =>
	let val a_2 = peek_val parser_env 5 in
	let val a_5 = peek_val parser_env 2 in
	Obj.repr((
(* # 242 "c/c_parser.ny" *)
                            MARKstatement (Do(a_5,a_2)))) end  end )
, (fn parser_env =>
	Obj.repr((
(* # 243 "c/c_parser.ny" *)
                            MARKstatement Break)))
, (fn parser_env =>
	Obj.repr((
(* # 244 "c/c_parser.ny" *)
                            MARKstatement Continue)))
, (fn parser_env =>
	let val a_2 = peek_val parser_env 1 in
	Obj.repr((
(* # 245 "c/c_parser.ny" *)
                            MARKstatement (Return a_2))) end )
, (fn parser_env =>
	let val a_2 = (peek_val parser_env 1 : string) in
	Obj.repr((
(* # 246 "c/c_parser.ny" *)
                            MARKstatement (Goto a_2))) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 247 "c/c_parser.ny" *)
                            a_1)) end )
, (fn parser_env =>
	let val a_1 = (peek_val parser_env 2 : string) in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 248 "c/c_parser.ny" *)
                            MARKstatement (Labeled (a_1,a_3)))) end  end )
, (fn parser_env =>
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 249 "c/c_parser.ny" *)
                            MARKstatement (DefaultLabel(a_3)))) end )
, (fn parser_env =>
	let val a_2 = peek_val parser_env 2 in
	let val a_4 = peek_val parser_env 0 in
	Obj.repr((
(* # 251 "c/c_parser.ny" *)
                            MARKstatement (CaseLabel(a_2,a_4)))) end  end )
, (fn parser_env =>
	let val a_3 = peek_val parser_env 2 in
	let val a_5 = peek_val parser_env 0 in
	Obj.repr((
(* # 253 "c/c_parser.ny" *)
                            MARKstatement (IfThen(a_3,a_5)))) end  end )
, (fn parser_env =>
	let val a_3 = peek_val parser_env 4 in
	let val a_5 = peek_val parser_env 2 in
	let val a_7 = peek_val parser_env 0 in
	Obj.repr((
(* # 255 "c/c_parser.ny" *)
                            MARKstatement (IfThenElse(a_3,a_5,a_7)))) end  end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	Obj.repr((
(* # 256 "c/c_parser.ny" *)
                            MARKstatement (Expr(a_1)))) end )
, (fn parser_env =>
	Obj.repr((
(* # 257 "c/c_parser.ny" *)
                            MARKstatement (Expr(EmptyExpr)))))
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	Obj.repr((
(* # 262 "c/c_parser.ny" *)
                                 a_1)) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 267 "c/c_parser.ny" *)
      insertDeclNames (a_1, []); // see-next-time 
      MARKdeclaration (Declaration (a_1,[])))) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 271 "c/c_parser.ny" *)
      let val decl = (a_1, seqToList a_2)
      in insertDeclNames decl;	// see-next-time
         MARKdeclaration (Declaration decl) end)) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 276 "c/c_parser.ny" *)
      let val decl = (a_1, seqToList a_2)
      in insertDeclNames decl;
         MARKdeclaration (Declaration decl) end)) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 281 "c/c_parser.ny" *)
                   seqToList a_1)) end )
, (fn parser_env =>
	Obj.repr((
(* # 282 "c/c_parser.ny" *)
                   [])))
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 285 "c/c_parser.ny" *)
                singletonSeq a_1)) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 286 "c/c_parser.ny" *)
                            addToSeq(a_2, a_1))) end  end )
, (fn parser_env =>
	let val a_2 = peek_val parser_env 4 in
	let val a_3 = peek_val parser_env 3 in
	let val a_4 = peek_val parser_env 2 in
	let val a_5 = peek_val parser_env 1 in
	Obj.repr((
(* # 292 "c/c_parser.ny" *)
      Compound ((List.map (fn x => Decl x) (seqToList a_3)) @ a_4))) end  end  end  end )
, (fn parser_env =>
	let val a_2 = peek_val parser_env 1 in
	Obj.repr((
(* # 293 "c/c_parser.ny" *)
                                 Compound a_2)) end )
, (fn parser_env =>
	Obj.repr((
(* # 300 "c/c_parser.ny" *)
          AddrOf)))
, (fn parser_env =>
	Obj.repr((
(* # 301 "c/c_parser.ny" *)
          Star)))
, (fn parser_env =>
	Obj.repr((
(* # 302 "c/c_parser.ny" *)
          Uplus)))
, (fn parser_env =>
	Obj.repr((
(* # 303 "c/c_parser.ny" *)
          Negate)))
, (fn parser_env =>
	Obj.repr((
(* # 304 "c/c_parser.ny" *)
          BitNot)))
, (fn parser_env =>
	Obj.repr((
(* # 305 "c/c_parser.ny" *)
          Not)))
, (fn parser_env =>
	let val a_1 = peek_val parser_env 4 in
	let val a_3 = peek_val parser_env 2 in
	let val a_5 = peek_val parser_env 0 in
	Obj.repr((
(* # 309 "c/c_parser.ny" *)
      MARKexpression (QuestionColon(a_1,a_3,a_5)))) end  end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 310 "c/c_parser.ny" *)
                           MARKexpression (Binop(PlusAssign,a_1,a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 311 "c/c_parser.ny" *)
                           MARKexpression (Binop(MinusAssign,a_1,a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 312 "c/c_parser.ny" *)
                           MARKexpression (Binop(TimesAssign,a_1,a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 313 "c/c_parser.ny" *)
                           MARKexpression (Binop(DivAssign,a_1,a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 314 "c/c_parser.ny" *)
                           MARKexpression (Binop(ModAssign,a_1,a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 315 "c/c_parser.ny" *)
                           MARKexpression (Binop(XorAssign,a_1,a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 316 "c/c_parser.ny" *)
                           MARKexpression (Binop(OrAssign,a_1,a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 317 "c/c_parser.ny" *)
                           MARKexpression (Binop(AndAssign,a_1,a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 318 "c/c_parser.ny" *)
                           MARKexpression (Binop(LshiftAssign,a_1,a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 319 "c/c_parser.ny" *)
                           MARKexpression (Binop(RshiftAssign,a_1,a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 320 "c/c_parser.ny" *)
                           MARKexpression (Binop(Assign,a_1,a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 321 "c/c_parser.ny" *)
                           MARKexpression (Binop(Or,a_1,a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 322 "c/c_parser.ny" *)
                           MARKexpression (Binop(And,a_1,a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 323 "c/c_parser.ny" *)
                           MARKexpression (Binop(BitOr,a_1,a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 324 "c/c_parser.ny" *)
                           MARKexpression (Binop(BitXor,a_1,a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 325 "c/c_parser.ny" *)
                           MARKexpression (Binop(BitAnd,a_1,a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 326 "c/c_parser.ny" *)
                           MARKexpression (Binop(Eq,a_1,a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 327 "c/c_parser.ny" *)
                           MARKexpression (Binop(Neq,a_1,a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 328 "c/c_parser.ny" *)
                           MARKexpression (Binop(Lt,a_1,a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 329 "c/c_parser.ny" *)
                           MARKexpression (Binop(Gt,a_1,a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 330 "c/c_parser.ny" *)
                           MARKexpression (Binop(Lte,a_1,a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 331 "c/c_parser.ny" *)
                           MARKexpression (Binop(Gte,a_1,a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 332 "c/c_parser.ny" *)
                           MARKexpression (Binop(Lshift,a_1,a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 333 "c/c_parser.ny" *)
                           MARKexpression (Binop(Rshift,a_1,a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 334 "c/c_parser.ny" *)
                           MARKexpression (Binop(Plus,a_1,a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 335 "c/c_parser.ny" *)
                           MARKexpression (Binop(Minus,a_1,a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 336 "c/c_parser.ny" *)
                           MARKexpression (Binop(Times,a_1,a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 337 "c/c_parser.ny" *)
                           MARKexpression (Binop(Divide,a_1,a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 338 "c/c_parser.ny" *)
                           MARKexpression (Binop(Mod,a_1,a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	Obj.repr((
(* # 339 "c/c_parser.ny" *)
                           MARKexpression (Unop(PostInc,a_1)))) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	Obj.repr((
(* # 340 "c/c_parser.ny" *)
                           MARKexpression (Unop(PostDec,a_1)))) end )
, (fn parser_env =>
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 341 "c/c_parser.ny" *)
                           MARKexpression (Unop(PreInc,a_2)))) end )
, (fn parser_env =>
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 342 "c/c_parser.ny" *)
                           MARKexpression (Unop(PreDec,a_2)))) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 343 "c/c_parser.ny" *)
                                    MARKexpression (Unop(a_1,a_2)))) end  end )
, (fn parser_env =>
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 344 "c/c_parser.ny" *)
                           MARKexpression (Unop(Sizeof,a_2)))) end )
, (fn parser_env =>
	let val a_2 = peek_val parser_env 2 in
	let val a_4 = peek_val parser_env 0 in
	Obj.repr((
(* # 345 "c/c_parser.ny" *)
                                          MARKexpression (Cast (a_2,a_4)))) end  end )
, (fn parser_env =>
	let val a_3 = peek_val parser_env 1 in
	Obj.repr((
(* # 347 "c/c_parser.ny" *)
                           MARKexpression (Unop(SizeofType a_3,EmptyExpr)))) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 3 in
	let val a_3 = peek_val parser_env 1 in
	Obj.repr((
(* # 348 "c/c_parser.ny" *)
                                  MARKexpression (Binop(Sub,a_1,a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	Obj.repr((
(* # 351 "c/c_parser.ny" *)
                        MARKexpression (Call(a_1,[])))) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 3 in
	let val a_3 = peek_val parser_env 1 in
	Obj.repr((
(* # 353 "c/c_parser.ny" *)
                        MARKexpression (Call(a_1, seqToList a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = (peek_val parser_env 0 : string) in
	Obj.repr((
(* # 354 "c/c_parser.ny" *)
                        MARKexpression (Binop(Dot,a_1,Id(a_3))))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = (peek_val parser_env 0 : string) in
	Obj.repr((
(* # 355 "c/c_parser.ny" *)
                        MARKexpression (Binop(Arrow,a_1,Id(a_3))))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = (peek_val parser_env 0 : string) in
	Obj.repr((
(* # 356 "c/c_parser.ny" *)
                        MARKexpression (Binop(Dot,a_1,Id(a_3))))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = (peek_val parser_env 0 : string) in
	Obj.repr((
(* # 357 "c/c_parser.ny" *)
                          MARKexpression (Binop(Arrow,a_1,Id(a_3))))) end  end )
, (fn parser_env =>
	let val a_2 = peek_val parser_env 1 in
	Obj.repr((
(* # 358 "c/c_parser.ny" *)
                             MARKexpression (a_2))) end )
, (fn parser_env =>
	let val a_1 = (peek_val parser_env 0 : int) in
	Obj.repr((
(* # 359 "c/c_parser.ny" *)
                        MARKexpression (IntConst a_1))) end )
, (fn parser_env =>
	let val a_1 = (peek_val parser_env 0 : string) in
	Obj.repr((
(* # 360 "c/c_parser.ny" *)
                        MARKexpression (RealConst a_1))) end )
, (fn parser_env =>
	let val a_1 = (peek_val parser_env 0 : int) in
	Obj.repr((
(* # 361 "c/c_parser.ny" *)
                        MARKexpression (IntConst a_1))) end )
, (fn parser_env =>
	let val a_1 = (peek_val parser_env 0 : string) in
	Obj.repr((
(* # 362 "c/c_parser.ny" *)
                        MARKexpression (Id(a_1)))) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 363 "c/c_parser.ny" *)
                        MARKexpression (String(a_1)))) end )
, (fn parser_env =>
	let val a_1 = (peek_val parser_env 0 : string) in
	Obj.repr((
(* # 366 "c/c_parser.ny" *)
                        a_1)) end )
, (fn parser_env =>
	let val a_1 = (peek_val parser_env 1 : string) in
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 367 "c/c_parser.ny" *)
                        a_1 ^ a_2)) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 372 "c/c_parser.ny" *)
                           a_1)) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 373 "c/c_parser.ny" *)
                           MARKexpression (Binop(Comma,a_1,a_3)))) end  end )
, (fn parser_env =>
	Obj.repr((
(* # 375 "c/c_parser.ny" *)
                 EmptyExpr)))
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 376 "c/c_parser.ny" *)
                 a_1)) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 380 "c/c_parser.ny" *)
                                  singletonSeq a_1)) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 381 "c/c_parser.ny" *)
                                  addToSeq(a_3, a_1))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 384 "c/c_parser.ny" *)
                             a_1)) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 386 "c/c_parser.ny" *)
      let val (ty,decl) = ctypeDecrToTypeName (a_1, a_2)
      in ty end)) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 391 "c/c_parser.ny" *)
      decltypeAddSpecifier (a_1, a_2))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_2 = peek_val parser_env 1 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 393 "c/c_parser.ny" *)
      decltypeAddAll (a_1, decltypeAddSpecifier (a_2,a_3)))) end  end  end )
, (fn parser_env =>
	Obj.repr((
(* # 396 "c/c_parser.ny" *)
                                         unknown_decltype)))
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 398 "c/c_parser.ny" *)
      let val {qualifiers,specifiers} = a_2
          val decltype = 
              {qualifiers=qualifiers,specifiers=specifiers,storage=[]}
      in decltypeAddAll (decltype, a_1) end)) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 403 "c/c_parser.ny" *)
      decltypeAddStorage (a_2,a_1))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 406 "c/c_parser.ny" *)
                                {qualifiers=[],specifiers=[a_1]})) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 407 "c/c_parser.ny" *)
                                {qualifiers=[a_1],specifiers=[]})) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 408 "c/c_parser.ny" *)
                                {qualifiers=[],specifiers=[a_1]})) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 409 "c/c_parser.ny" *)
                                {qualifiers=[],specifiers=[a_1]})) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 413 "c/c_parser.ny" *)
      {storage = [a_1], qualifiers = [], specifiers = []})) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 415 "c/c_parser.ny" *)
      decltypeAddStorage(a_2,a_1))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 417 "c/c_parser.ny" *)
      {specifiers = [], storage = [], qualifiers = [a_1]})) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 419 "c/c_parser.ny" *)
      decltypeAddQualifier(a_2, a_1))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 423 "c/c_parser.ny" *)
      let val {specifiers, qualifiers} = a_2
      in {specifiers=a_1::specifiers,qualifiers=qualifiers} end)) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_2 = peek_val parser_env 1 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 427 "c/c_parser.ny" *)
      let val {specifiers, qualifiers} = a_3
      in {specifiers=a_2::specifiers,qualifiers=a_1@qualifiers} end)) end  end  end )
, (fn parser_env =>
	Obj.repr((
(* # 431 "c/c_parser.ny" *)
      {qualifiers=[],specifiers=[]})))
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 433 "c/c_parser.ny" *)
      let val {specifiers=s1, qualifiers=q1} = a_1
          val {specifiers=s2, qualifiers=q2} = a_2
      in {specifiers=s1@s2, qualifiers=q1@q2} end)) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 438 "c/c_parser.ny" *)
                                      [a_1])) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 439 "c/c_parser.ny" *)
                                      a_1::a_2)) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 442 "c/c_parser.ny" *)
                                a_1)) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 443 "c/c_parser.ny" *)
                                a_1)) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 444 "c/c_parser.ny" *)
                                a_1)) end )
, (fn parser_env =>
	let val a_1 = (peek_val parser_env 0 : string) in
	Obj.repr((
(* # 445 "c/c_parser.ny" *)
                                TypedefName a_1)) end )
, (fn parser_env =>
	Obj.repr((
(* # 448 "c/c_parser.ny" *)
              Void)))
, (fn parser_env =>
	Obj.repr((
(* # 449 "c/c_parser.ny" *)
              Char)))
, (fn parser_env =>
	Obj.repr((
(* # 450 "c/c_parser.ny" *)
              Short)))
, (fn parser_env =>
	Obj.repr((
(* # 451 "c/c_parser.ny" *)
              Int)))
, (fn parser_env =>
	Obj.repr((
(* # 452 "c/c_parser.ny" *)
              Long)))
, (fn parser_env =>
	Obj.repr((
(* # 453 "c/c_parser.ny" *)
              Float)))
, (fn parser_env =>
	Obj.repr((
(* # 454 "c/c_parser.ny" *)
              Double)))
, (fn parser_env =>
	Obj.repr((
(* # 455 "c/c_parser.ny" *)
              Signed)))
, (fn parser_env =>
	Obj.repr((
(* # 456 "c/c_parser.ny" *)
              Unsigned)))
, (fn parser_env =>
	let val a_1 = peek_val parser_env 3 in
	let val a_3 = peek_val parser_env 1 in
	Obj.repr((
(* # 461 "c/c_parser.ny" *)
      Struct{isStruct=a_1, tagOpt=None, members=seqToList a_3})) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 4 in
	let val a_2 = (peek_val parser_env 3 : string) in
	let val a_4 = peek_val parser_env 1 in
	Obj.repr((
(* # 463 "c/c_parser.ny" *)
      Struct{isStruct=a_1, tagOpt=Some a_2, members=seqToList a_4})) end  end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 4 in
	let val a_2 = (peek_val parser_env 3 : string) in
	let val a_4 = peek_val parser_env 1 in
	Obj.repr((
(* # 465 "c/c_parser.ny" *)
      Struct{isStruct=a_1, tagOpt=Some a_2, members=seqToList a_4})) end  end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	let val a_2 = (peek_val parser_env 0 : string) in
	Obj.repr((
(* # 466 "c/c_parser.ny" *)
                       StructTag {isStruct=a_1, name=a_2})) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	let val a_2 = (peek_val parser_env 0 : string) in
	Obj.repr((
(* # 467 "c/c_parser.ny" *)
                             StructTag {isStruct=a_1, name=a_2})) end  end )
, (fn parser_env =>
	Obj.repr((
(* # 471 "c/c_parser.ny" *)
              true)))
, (fn parser_env =>
	Obj.repr((
(* # 472 "c/c_parser.ny" *)
             false)))
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 475 "c/c_parser.ny" *)
                                            singletonSeq a_1)) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 476 "c/c_parser.ny" *)
                                            addToSeq(a_2, a_1))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_2 = peek_val parser_env 1 in
	Obj.repr((
(* # 480 "c/c_parser.ny" *)
      (a_1, seqToList a_2))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_2 = peek_val parser_env 1 in
	Obj.repr((
(* # 482 "c/c_parser.ny" *)
      let val ct = {qualifiers=a_1, specifiers=[]}
      in (ct, seqToList a_2) end)) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 486 "c/c_parser.ny" *)
                                                singletonSeq a_1)) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 487 "c/c_parser.ny" *)
                                                addToSeq(a_3, a_1))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 490 "c/c_parser.ny" *)
                                                      singletonSeq a_1)) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 491 "c/c_parser.ny" *)
                                                      addToSeq(a_3, a_1))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 494 "c/c_parser.ny" *)
                           (a_1, EmptyExpr))) end )
, (fn parser_env =>
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 495 "c/c_parser.ny" *)
                           (EmptyDecr, a_2))) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 496 "c/c_parser.ny" *)
                           (a_1, a_3))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 499 "c/c_parser.ny" *)
                                (a_1, EmptyExpr))) end )
, (fn parser_env =>
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 500 "c/c_parser.ny" *)
                                (EmptyDecr, a_2))) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 501 "c/c_parser.ny" *)
                                (a_1, a_3))) end  end )
, (fn parser_env =>
	Obj.repr((
(* # 504 "c/c_parser.ny" *)
               CONST)))
, (fn parser_env =>
	Obj.repr((
(* # 505 "c/c_parser.ny" *)
               VOLATILE)))
, (fn parser_env =>
	let val a_3 = peek_val parser_env 2 in
	let val a_4 = peek_val parser_env 1 in
	Obj.repr((
(* # 509 "c/c_parser.ny" *)
      Enum{tagOpt=None, enumerators=seqToList a_3, trailingComma=a_4})) end  end )
, (fn parser_env =>
	let val a_2 = (peek_val parser_env 4 : string) in
	let val a_4 = peek_val parser_env 2 in
	let val a_5 = peek_val parser_env 1 in
	Obj.repr((
(* # 511 "c/c_parser.ny" *)
      Enum{tagOpt=Some(a_2), enumerators=seqToList a_4, trailingComma=a_5})) end  end  end )
, (fn parser_env =>
	let val a_2 = (peek_val parser_env 4 : string) in
	let val a_4 = peek_val parser_env 2 in
	let val a_5 = peek_val parser_env 1 in
	Obj.repr((
(* # 513 "c/c_parser.ny" *)
      Enum{tagOpt=Some(a_2), enumerators=seqToList a_4, trailingComma=a_5})) end  end  end )
, (fn parser_env =>
	let val a_2 = (peek_val parser_env 0 : string) in
	Obj.repr((
(* # 514 "c/c_parser.ny" *)
              EnumTag(a_2))) end )
, (fn parser_env =>
	let val a_2 = (peek_val parser_env 0 : string) in
	Obj.repr((
(* # 515 "c/c_parser.ny" *)
                    EnumTag(a_2))) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 519 "c/c_parser.ny" *)
      addToSeq(a_3, a_1))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 521 "c/c_parser.ny" *)
      let val (id,e) = a_1 
      in TypeDefs.addNoTdef(id); singletonSeq a_1
      end)) end )
, (fn parser_env =>
	let val a_1 = (peek_val parser_env 0 : string) in
	Obj.repr((
(* # 526 "c/c_parser.ny" *)
                    (a_1,EmptyExpr))) end )
, (fn parser_env =>
	let val a_1 = (peek_val parser_env 2 : string) in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 527 "c/c_parser.ny" *)
                    (a_1,a_3))) end  end )
, (fn parser_env =>
	Obj.repr((
(* # 530 "c/c_parser.ny" *)
              EXTERN)))
, (fn parser_env =>
	Obj.repr((
(* # 531 "c/c_parser.ny" *)
              STATIC)))
, (fn parser_env =>
	Obj.repr((
(* # 532 "c/c_parser.ny" *)
              AUTO)))
, (fn parser_env =>
	Obj.repr((
(* # 533 "c/c_parser.ny" *)
              REGISTER)))
, (fn parser_env =>
	Obj.repr((
(* # 534 "c/c_parser.ny" *)
              TYPEDEF)))
, (fn parser_env =>
	Obj.repr((
(* # 536 "c/c_parser.ny" *)
                   false)))
, (fn parser_env =>
	Obj.repr((
(* # 537 "c/c_parser.ny" *)
             true)))
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 540 "c/c_parser.ny" *)
                                            singletonSeq a_1)) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 541 "c/c_parser.ny" *)
                                            addToSeq(a_3, a_1))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 544 "c/c_parser.ny" *)
                                  (a_1,EmptyExpr))) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 545 "c/c_parser.ny" *)
                                  (a_1,a_3))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 548 "c/c_parser.ny" *)
                                                  singletonSeq a_1)) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 549 "c/c_parser.ny" *)
                                                  addToSeq(a_3, a_1))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 552 "c/c_parser.ny" *)
                                          (a_1,EmptyExpr))) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 553 "c/c_parser.ny" *)
                                          (a_1,a_3))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 556 "c/c_parser.ny" *)
                                 a_1)) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 557 "c/c_parser.ny" *)
                                 a_1)) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 560 "c/c_parser.ny" *)
                                      a_1)) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 561 "c/c_parser.ny" *)
                                      applyPointer(a_1,a_2))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 564 "c/c_parser.ny" *)
                                   a_1)) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 565 "c/c_parser.ny" *)
                                   applyPointer(a_1, a_2))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 568 "c/c_parser.ny" *)
                                 a_1)) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 569 "c/c_parser.ny" *)
                                 applyPointer(a_1, a_2))) end  end )
, (fn parser_env =>
	Obj.repr((
(* # 572 "c/c_parser.ny" *)
                                    PointerDecr(EmptyDecr))))
, (fn parser_env =>
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 573 "c/c_parser.ny" *)
                                    PointerDecr(loopQd(a_2,EmptyDecr)))) end )
, (fn parser_env =>
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 574 "c/c_parser.ny" *)
                                    PointerDecr(a_2))) end )
, (fn parser_env =>
	let val a_2 = peek_val parser_env 1 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 575 "c/c_parser.ny" *)
                                    PointerDecr(loopQd(a_2,a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = (peek_val parser_env 0 : string) in
	Obj.repr((
(* # 578 "c/c_parser.ny" *)
                                          VarDecr a_1)) end )
, (fn parser_env =>
	let val a_2 = peek_val parser_env 1 in
	Obj.repr((
(* # 579 "c/c_parser.ny" *)
                                          a_2)) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	Obj.repr((
(* # 581 "c/c_parser.ny" *)
      ArrayDecr (a_1,EmptyExpr))) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 3 in
	let val a_3 = peek_val parser_env 1 in
	Obj.repr((
(* # 583 "c/c_parser.ny" *)
      ArrayDecr (a_1,a_3))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	Obj.repr((
(* # 585 "c/c_parser.ny" *)
      FuncDecr (a_1,nil))) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 3 in
	let val a_3 = peek_val parser_env 1 in
	Obj.repr((
(* # 587 "c/c_parser.ny" *)
      FuncDecr (a_1,a_3))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 3 in
	let val a_3 = peek_val parser_env 1 in
	Obj.repr((
(* # 589 "c/c_parser.ny" *)
      FuncDecr (a_1, 
         List.map (fn x => (unknown_decltype,VarDecr x)) (seqToList a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = (peek_val parser_env 0 : string) in
	Obj.repr((
(* # 593 "c/c_parser.ny" *)
          VarDecr a_1)) end )
, (fn parser_env =>
	let val a_2 = peek_val parser_env 1 in
	Obj.repr((
(* # 594 "c/c_parser.ny" *)
                                    a_2)) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	Obj.repr((
(* # 596 "c/c_parser.ny" *)
      ArrayDecr (a_1,EmptyExpr))) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 3 in
	let val a_3 = peek_val parser_env 1 in
	Obj.repr((
(* # 598 "c/c_parser.ny" *)
      ArrayDecr (a_1,a_3))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	Obj.repr((
(* # 600 "c/c_parser.ny" *)
      FuncDecr (a_1,nil))) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 3 in
	let val a_3 = peek_val parser_env 1 in
	Obj.repr((
(* # 602 "c/c_parser.ny" *)
      FuncDecr (a_1,a_3))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 3 in
	let val a_3 = peek_val parser_env 1 in
	Obj.repr((
(* # 604 "c/c_parser.ny" *)
      FuncDecr (a_1,
         List.map (fn x => (unknown_decltype,VarDecr x)) (seqToList a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = (peek_val parser_env 0 : string) in
	Obj.repr((
(* # 608 "c/c_parser.ny" *)
                VarDecr a_1)) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	Obj.repr((
(* # 610 "c/c_parser.ny" *)
      ArrayDecr (a_1,EmptyExpr))) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 3 in
	let val a_3 = peek_val parser_env 1 in
	Obj.repr((
(* # 612 "c/c_parser.ny" *)
      ArrayDecr (a_1,a_3))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	Obj.repr((
(* # 614 "c/c_parser.ny" *)
      FuncDecr (a_1,nil))) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 3 in
	let val a_3 = peek_val parser_env 1 in
	Obj.repr((
(* # 616 "c/c_parser.ny" *)
      FuncDecr (a_1,a_3))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 3 in
	let val a_3 = peek_val parser_env 1 in
	Obj.repr((
(* # 618 "c/c_parser.ny" *)
      FuncDecr (a_1,
          List.map (fn x => (unknown_decltype,VarDecr x)) (seqToList a_3)))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 622 "c/c_parser.ny" *)
            a_1)) end )
, (fn parser_env =>
	let val a_2 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 1 in
	Obj.repr((
(* # 624 "c/c_parser.ny" *)
      MARKexpression (InitList(seqToList a_2)))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 627 "c/c_parser.ny" *)
                  singletonSeq a_1)) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 628 "c/c_parser.ny" *)
                                      addToSeq(a_3, a_1))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 632 "c/c_parser.ny" *)
      singletonSeq(MARKdeclaration a_1))) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 634 "c/c_parser.ny" *)
      addToSeq(MARKdeclaration a_2, a_1))) end  end )
, (fn parser_env =>
	let val a_1 = (peek_val parser_env 0 : string) in
	Obj.repr((
(* # 637 "c/c_parser.ny" *)
                       singletonSeq a_1)) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = (peek_val parser_env 0 : string) in
	Obj.repr((
(* # 638 "c/c_parser.ny" *)
                       addToSeq(a_3, a_1))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 643 "c/c_parser.ny" *)
      insertFuncName(a_1);
      TypeDefs.pushScope();
      insertFuncParams(a_1);
      (unknown_decltype, a_1))) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 648 "c/c_parser.ny" *)
      insertFuncName(a_2);
      TypeDefs.pushScope();
      insertFuncParams(a_2);
      (a_1, a_2))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 653 "c/c_parser.ny" *)
      insertFuncName(a_2);
      TypeDefs.pushScope();
      insertFuncParams(a_2);
      (a_1, a_2))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 660 "c/c_parser.ny" *)
      TypeDefs.popScope(); // see-next-time
      let val (ret,fnname) = a_1
      in 
         FunctionDef
            {retType = ret,
             funDecr = fnname,
             krParams = [],
             body = a_2} end)) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_2 = peek_val parser_env 1 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 670 "c/c_parser.ny" *)
      TypeDefs.popScope();
      let val (ret, fnname) = a_1
      in
         FunctionDef 
            {retType = ret,
             funDecr = fnname,
             krParams = seqToList a_2,
             body = a_3} end)) end  end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 680 "c/c_parser.ny" *)
                                     applyPointer (a_1, EmptyDecr))) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 681 "c/c_parser.ny" *)
                                     a_1)) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 682 "c/c_parser.ny" *)
                                     applyPointer(a_1, a_2))) end  end )
, (fn parser_env =>
	let val a_2 = peek_val parser_env 1 in
	Obj.repr((
(* # 685 "c/c_parser.ny" *)
                                                a_2)) end )
, (fn parser_env =>
	Obj.repr((
(* # 686 "c/c_parser.ny" *)
                                                ArrayDecr(EmptyDecr, EmptyExpr))))
, (fn parser_env =>
	let val a_2 = peek_val parser_env 1 in
	Obj.repr((
(* # 687 "c/c_parser.ny" *)
                                                ArrayDecr(EmptyDecr, a_2))) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	Obj.repr((
(* # 688 "c/c_parser.ny" *)
                                                ArrayDecr (a_1,EmptyExpr))) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 3 in
	let val a_3 = peek_val parser_env 1 in
	Obj.repr((
(* # 689 "c/c_parser.ny" *)
                                                ArrayDecr (a_1,a_3))) end  end )
, (fn parser_env =>
	Obj.repr((
(* # 690 "c/c_parser.ny" *)
                                                FuncDecr (EmptyDecr ,nil))))
, (fn parser_env =>
	let val a_2 = peek_val parser_env 1 in
	Obj.repr((
(* # 691 "c/c_parser.ny" *)
                                                FuncDecr (EmptyDecr, a_2))) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	Obj.repr((
(* # 692 "c/c_parser.ny" *)
                                                FuncDecr (a_1,nil))) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 3 in
	let val a_3 = peek_val parser_env 1 in
	Obj.repr((
(* # 694 "c/c_parser.ny" *)
                                                FuncDecr (a_1, a_3))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 697 "c/c_parser.ny" *)
                                          seqToList a_1)) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	Obj.repr((
(* # 699 "c/c_parser.ny" *)
     let val decltype = {specifiers=[Ellipses],qualifiers=[],storage=[]}
     in (seqToList a_1) @ [(decltype, EllipsesDecr)] end)) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 704 "c/c_parser.ny" *)
      let val (ty,decl) = a_1
      in singletonSeq(ty, MARKdeclarator decl) end)) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 2 in
	let val a_3 = peek_val parser_env 0 in
	Obj.repr((
(* # 707 "c/c_parser.ny" *)
      let val (ty,decl) = a_3
      in addToSeq((ty, MARKdeclarator decl), a_1) end)) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 718 "c/c_parser.ny" *)
                                              (a_1, a_2))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 719 "c/c_parser.ny" *)
                                              (a_1, a_2))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 0 in
	Obj.repr((
(* # 721 "c/c_parser.ny" *)
      (a_1, EmptyDecr))) end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 722 "c/c_parser.ny" *)
                                              (a_1, a_2))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 723 "c/c_parser.ny" *)
                                              (a_1, a_2))) end  end )
, (fn parser_env =>
	let val a_1 = peek_val parser_env 1 in
	let val a_2 = peek_val parser_env 0 in
	Obj.repr((
(* # 724 "c/c_parser.ny" *)
                                              (a_1, a_2))) end  end )
, (fn parser_env =>
	Obj.repr((
(* # 726 "c/c_parser.ny" *)
               TypeDefs.pushScope())))
, (fn parser_env =>
	Obj.repr((
(* # 728 "c/c_parser.ny" *)
               TypeDefs.popScope())))
(* Entry goal *)
, (fn parser_env => raise (YYexit (peek_val parser_env 0)))
|]
val yytables =
  { actions=yyact,
    transl_const=yytransl_const,
    transl_block=yytransl_block,
    lhs=yylhs,
    len=yylen,
    defred=yydefred,
    dgoto=yydgoto,
    sindex=yysindex,
    rindex=yyrindex,
    gindex=yygindex,
    tablesize=yytablesize,
    table=yytable,
    check=yycheck,
    error_function=parse_error,
    names_const=yynames_const,
    names_block=yynames_block }
fun goal (lexfun : Lexing.lexbuf -> token) (lexbuf : Lexing.lexbuf) =
   (yyparse yytables 1 lexfun lexbuf : ParseTree.externalDecl list)


end (* END of structure C_parser *)
